/*****************************************************************************
*  acd2tdt -- convert acd's .rdf format to .tdt format
*
*  Written by Dave Weininger, dinked by Jeremy Yang so it's
*  a toolkit program without internal dy_ calls.
*
*  3 Feb 2000
*----------------------------------------------------------------------------
*  TODO:
*   [x] deal with continuation lines +
*   [x] deal with chirality
*   [x] deal with double bond orientation
*   [ ] deal with singlet, triplet, non-ring radicals?
*   [x] test for leaks
*   [x] check against AMW
*   [x] deal with $NAM's?
*   [?] sort names?
*   [x] deal with casnos
*   [x] sort prices?
*   [x] add $CNO
*   [ ] deal with V2000
*   [ ] add cmw as $SMI/AMW
*----------------------------------------------------------------------------
*
*   $SMI<>
*      $ACD<number>
*         PN<preferred name>
*         ISM<>
*         2D<>
*         CAT<catno;supplier;name;grade;dpg;price;remark;updated>
*      $ACD<number>
*         PN<preferred name>
*         ISM<>
*         2D<>
*         CAT<catno;supplier;name;grade;dpg;price;remark;updated>
*      $CAS<>
*         ECN<erroneous cas number>
*         ACD<>
*      $NAM<>
*
*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <memory.h>
#include <dt_smiles.h>
#include <dt_depict.h>
#include <du_utils.h>
#include "acd2tdt.h"

#define NELEMENT    103
#define MX_SLOP     0.1
#define AVE_HYDROGEN_MASS 1.007947

#define TRUE 1
#define FALSE 0

/*** Handy macro converts NULL to empty string. ***/

#define SELF(s) ((NULL == (s)) ? "" : (s))

/*** Name item struct. ***/

typedef struct _name {
   char         *name;
   char         *std;
   struct _name *next;
} NAMEITEM;

/*** Catalog entry struct. ***/

typedef struct _cent {
   char         *catno;
   char         *supplier;
   char         *catnam;
   char         *currency;
   char         *price;
   char         *amount;
   char         *unitname;
   char         *usdpg;
   char         *purity;
   char         *grade;
   char         *upcode;
   char         *remarks;
   struct _cent *next;
} CENT;

/*** Unifying struct. ***/

typedef struct _a {
   FILE      *fpin, *fpout1, *fpout2;
   dt_Handle  mol, dep;                   /* molecule and depiction */
   char      *acd, *cas, *pn, *amw;       /* cas not used */
   CENT      *cents;                      /* catalog entry list */
   NAMEITEM  *names;                      /* names */
   NAMEITEM  *casnos;                     /* cas numbers */
   void      *units;                      /* unit conversion list */
   int       nids, ntrees;                /* total IDs, trees output */
} A;

/*** Reporting data. ***/

static char g_mireg[80];
static int  g_lineno;
static int  g_msec;

/*============================================================================
 *  proginit() -- initialize program
 */

static A *proginit(int *pargc, char **argv)
{
   A *a;
   int i;
   char path[256]="acd",file[256];

   for(i=1; i<*pargc; i++) {
     if(0 == strcmp(argv[i],"-o"))
       if(++i >= *pargc)
	 printf("%s: missing argument to -o option, continuing\n");
       else
	 strcpy(path,argv[i]);
   }

   /*** Initialize unifying program struct. ***/

   a = (A *) malloc(sizeof(A));
   if (NULL == a) { fprintf(stderr, "ouch, can't initialize\n"); exit(1); }

   /*** Initialize global reporting data. ***/

  *g_mireg  = '\0';
   g_lineno = 0;
   g_msec   = du_millisec(0);

   /*** Input and output streams. ***/

   a->fpin   = stdin;
   fprintf(stderr, "Writing output file \"%s.raw.tdt\"...\n",path);
   sprintf(file,"%s%s",path,".raw.tdt");
   a->fpout1 = fopen(file, "a");
   fprintf(stderr, "Writing output file \"%s.asd.tdt\"...\n",path);
   sprintf(file,"%s%s",path,".asd.tdt");
   a->fpout2 = fopen(file, "a");
   
   if (NULL == a->fpout1 || NULL == a->fpout2) {
      fprintf(stderr, "can't open output file(s): %p/%p\n",
	      a->fpout1, a->fpout2);
      exit(1);
   }

   /*** Counts. ***/

   a->nids   = 0;
   a->ntrees = 0;

   /*** Various. ***/

   a->mol    = NULL_OB;
   a->dep    = NULL_OB;
   a->acd    = NULL;
   a->cas    = NULL;
   a->pn     = NULL;
   a->amw    = NULL;
   a->cents  = NULL;
   a->names  = NULL;
   a->casnos = NULL;
   a->units  = NULL;

   /*** Return initialize program struct. ***/

   return a;
}


/*============================================================================
 *  printerrors() -- print errors
 */

void printerrors(FILE *fp, int level)
{
   dt_Handle  errors, sob;
   char      *str;
   int        lens;
 
   if (dt_errorworst() >= level && NULL_OB != (errors = dt_errors(level))) {
      dt_reset(errors);
      while (NULL_OB != (sob = dt_next(errors))) {
         str = dt_stringvalue(&lens, sob);
         fprintf(fp, "%.*s\n", lens, str);
         dt_dealloc(sob);
      }
      dt_dealloc(errors);
   }
   dt_errorclear();
}

/*============================================================================
 *  oops() -- say error
 */
 
void oops(char *msg, char *line)
{
   fprintf(stderr, "oops: %s (%s line %d)\n",
           (msg ? msg : ""), g_mireg, g_lineno);
   if (line) fprintf(stderr, "%s\n", line);
}

/*============================================================================
 *  progress() -- show progress
 */

void progress(int nin)
{
   float sec = du_millisec(g_msec) / 1000.0;
   float nps = (float) nin / sec;
   fprintf(stderr, "%d in %.1f sec (%.1f/s) %s\n", nin, sec, nps, g_mireg);
}

/*============================================================================
 *  scratchpad() -- a more efficient approach to strcat()
 *
 *  scratchpad(NULL) .... reset scratch pad, return NULL
 *  scratchpad(str) ..... append str to scratchpad, return it
 */
 
char *scratchpad(char *str)
{
   static char *buf   = NULL;
   static int   room  = 0;
   static int   lenb  = 0;
   static int   armed = 0;
   int          lens;
 
   /*** If called with NULL once, clear; if twice, reset everything. ***/
 
   if (NULL == str) {
      if (armed) { du_free(buf); room = lenb = armed = 0; }
      else       { lenb = 0; armed = 1; }
      return NULL;
   }
   armed = 0;
 
   /*** Else make room if needed. ***/
 
   lens = strlen(str);
   if (lenb + lens + 1 > room) {
      room = MAX(1024, 2 * room + 2 * lens + 1);
      if (NULL == buf) buf  = (char *) malloc (     room * sizeof(char));
      else             buf  = (char *) realloc(buf, room * sizeof(char));
      if (NULL == buf) { oops("out of memory", str); exit(2); }
   }
 
   /*** Append string to end of buf, zero-terminate, and return buf. ***/
 
   if (0 < lens) memcpy(buf + lenb, str, lens);
   lenb      += lens;
   buf[lenb]  = '\0';
   return buf;
}

/*============================================================================
 *  fprint2d
 */

static void fprint2d(FILE *fp, dt_Handle dep)
{
   dt_Handle  mol, atoms, atom;
   dt_String  usmi;
   int        first;
   float      x, y, z;

   mol   = dt_base(dep);
   atoms = dt_e_canatom_stream(mol, FALSE, FALSE);

   if (NULL_OB != atoms) {
      fprintf(fp, "2D<");
      for (first = TRUE; NULL != (atom = dt_next(atoms)); first = FALSE) {
         if (!first) fprintf(fp, ",");
         dt_getcoord(dep, atom, &x, &y, &z);
         fprintf(fp, "%.2f,%.2f", x, y);
      }
      fprintf(fp, ">\n");
      dt_dealloc(atoms);
   }
}

/*============================================================================
 *  getline() -- if line is 82 characters and ' +'  continuations lines
 */

static char *getline(FILE *fp)
{
   static int   size = 0;
   static char *buf  = NULL;
   char        *p;

   /*** Reset if called with NULL. ***/

   if (NULL == fp) { du_free(buf); buf = NULL; size = 0; }

   /*** Allocate as needed. ***/

   if (NULL == buf) { size = 512; buf = (char *) malloc(size); *buf = '\0'; }

   /*** Read input, adding continuations and reallocating as needed. ***/

   for (p = buf; NULL != du_fgetz(p, 120, fp); p += 80) {
      g_lineno++;
      if (81 != strlen(p) || '+' != p[80]) return buf;
   }

   /*** Return NULL at EOF. ***/

   return NULL;
}

/*============================================================================
 *  my_nopunct() -- like du_nopunct(), but also removes []'s
 */
 
static char *my_nopunct(char *s)
{
  char *psrc, *pdest;
 
  if (s) {
    for (psrc = s, pdest = s; *psrc; psrc++)
      switch (*psrc) {
      case '!': case '(': case ')': case '-':
      case ',': case '.': case '?': case '\'':
      case '"': case '`': case ':': case ';':
      case '>': case '<': case '|': case '$':
      case '[': case ']':
      case '\001': case '\002': case '\003': case '\004': case '\005':
        break;
      default:
        *(pdest++) = *psrc; break;
      }
    *pdest = '\0';
  }
 
  return s;
}

/*============================================================================
 *  quotedup() -- like du_strdup(), but quote if needed
 */

static char *quotedup(char *str)
{
   char *s = du_possiblyquote(strlen(str), str);
   return (s ? s : du_strdup(str));
}

/*============================================================================
 *  savecas()
 */

static void savecas(A *a, char *in)
{
   char     *name, *std, *tmp, *cas;
   NAMEITEM *nit, *new;
   int       lens, leno;

   /*** Reject empties. ***/

   if (NULL == in || '\0' == *in) return;

   /*** Extract normal and standardized cas numbers. ***/

   tmp  = du_strdup(in);
   name = du_strdup(du_white0(tmp));
   leno = strlen(name);
   lens = leno;
   du_cas_standardize(&lens, name, &cas);

   /*** Save standard CAS number, name only if mismatch. ***/

   if (leno == lens && 0 == strncmp(cas, name, lens)) {
      std  = name;
      name = NULL;
   } else {
      std = du_strndup(cas, lens);
      oops("cas mismatch, input", name);
      oops("cas mismatch, calc.", std);
   }

   /*** Loop over extant casnos in name list. ***/

   for (nit = a->casnos; NULL != nit; nit = nit->next) {

      /*** If found in extant list, replace only if needed. ***/

      if (0 == strcmp(std, nit->std)) {
         if (NULL != name && NULL == nit->name) nit->name = name;
         else                                   du_free(name);
         du_free(std);
         break;
      }
   }

   /*** If not found in extant list, create new entry. ***/

   if (NULL == nit) {
      new       = (NAMEITEM *) malloc(sizeof(NAMEITEM));
      new->name = name;
      new->std  = std;
      new->next = a->casnos;
      a->casnos = new;
   }

   /*** Free temporary string and return. ***/

   du_free(tmp);
}

/*============================================================================
 *  savename()
 */
 
static void savename(A *a, char *in)
{
   char     *name, *std, *tmp;
   NAMEITEM *nit, *new;
 
   /*** Reject empties. ***/
 
   if (NULL == in || '\0' == *in) return;
 
   /*** Extract normal and standardized names. ***/
 
   tmp  = du_strdup(in);
   name = du_strdup(du_white1(du_trim(tmp)));
   std  = du_strdup(du_upcase(my_nopunct(du_white0(tmp))));
 
   /*** Loop over extant names in name list. ***/
 
   for (nit = a->names; NULL != nit; nit = nit->next) {
 
      /*** If found in extant list, replace only if longer. ***/
 
      if (0 == strcmp(std, nit->std)) {
         if ((int) strlen(name) > (int) strlen(nit->name)) {
            du_free(nit->name);
            nit->name = name;
            du_free(std);
         } else {
            du_free(name);
            du_free(std);
         }
         break;
      }
   }
 
   /*** If not found in extant list, create new entry. ***/
 
   if (NULL == nit) {
      new       = (NAMEITEM *) malloc(sizeof(NAMEITEM));
      new->name = name;
      new->std  = std;
      new->next = a->names;
      a->names  = new;
   }
 
   /*** Free temporary string and return. ***/
 
   du_free(tmp);
}

/*============================================================================
 *  cmpstd
 */

cmpstd(NAMEITEM *n1, NAMEITEM *n2)
{
   char   *s1, *s2, *x;

   /*** NULL pointers are greater than anything else. ***/

   if (NULL == n1 || NULL == n2) {
      if (NULL == n1 && NULL == n2) return 0;
      if (NULL == n1) return 1;
      return -1;
   }

   /*** Missing values are greater than non-missing values. ***/

   s1 = n1->std;
   s2 = n2->std;
   if (NULL == s1 || NULL == s2) {
      if (NULL == s1 && NULL == s2) return 0;
      if (NULL == s1) return 1;
      return -1;
   }

   /*** Else return normal string value comparison. ***/

   return strcmp(s1, s2);
}

/*============================================================================
 *  new_cent(), zap_cent() -- support for catalog entry struct
 */

static CENT *new_cent(void)
{
   CENT *cent = (CENT *) malloc(sizeof(CENT));

   if (cent) {
      cent->catno    = NULL;
      cent->supplier = NULL;
      cent->catnam   = NULL;
      cent->currency = NULL;
      cent->price    = NULL;
      cent->amount   = NULL;
      cent->unitname = NULL;
      cent->usdpg    = NULL;
      cent->purity   = NULL;
      cent->grade    = NULL;
      cent->upcode   = NULL;
      cent->remarks  = NULL;
      cent->next     = NULL;
   }
   return cent;
}

static void zap_cent(CENT *cent)
{
   if (cent) {
      du_free(cent->catno);
      du_free(cent->supplier);
      du_free(cent->catnam);
      du_free(cent->currency);
      du_free(cent->price);
      du_free(cent->amount);
      du_free(cent->unitname);
      du_free(cent->usdpg);
      du_free(cent->purity);
      du_free(cent->grade);
      du_free(cent->upcode);
      du_free(cent->remarks);
      du_free(cent);
   }
}

/*============================================================================
 *  deal_with_supplier() -- add catalog price entry (CENT) structs to a->cents
 */

static int deal_with_supplier(A *a)
{
   int   state = 0;
   int   lens;
   char  *line, *sp, *str;
   CENT  *cent;
   char  *supplier     = NULL;
   char  *catnam       = NULL;
   char  *upcode       = NULL;
   char  *remarks      = NULL;

   /*** Read (possibly continued) input line. ***/

   while (1) {
      if (NULL == (line = getline(a->fpin))) return FALSE;

      /*** If it starts "$DTYPE MOL:SUPPLIER(", get next line. ***/

      if (0 == strncmp("$DTYPE MOL:SUPPLIER(", line, 20))
         if (NULL == (line = getline(a->fpin))) return FALSE;

      /*** If it doesn't start "$DATUM ", something is wrong. ***/

      if (0 != strncmp(line, "$DATUM ", 7)) {
         oops("unexpected supplier line", line);
         return FALSE;
      }

      /*** Return entry list if line starts ".........." and able. ***/

      if (0 == strncmp(line + 7, "..........", 10)) {
         if (supplier) { du_free(supplier); supplier = NULL; }
         if (catnam)   { du_free(catnam);   catnam   = NULL; }
         if (upcode)   { du_free(upcode);   upcode   = NULL; }
         if (remarks)  { scratchpad(NULL);  remarks  = NULL; }
         return TRUE;

      /*** In state 0, record supplier and update code, advance. ***/

      } else if (0 == state) {
         upcode   = du_strdup(du_trim(line + 27)); line[26] = '\0';
         supplier = du_strdup(du_trim(line +  7));
         state    = 1;

      /*** In state 1, save and xref chemical name or advance. ***/

      } else if (1 == state) {
         if ('.' == line[7]) {
            state = 3;
         } else {
            catnam = du_strdup(du_trim(line + 7));
	    savename(a, catnam);
            state = 2;
         }

      /*** In state 2, xref cas number or advance. ***/

      } else if (2 == state) {
         if ('.' == line[7]) {
            state = 3;
         } else {
            savecas(a, du_trim(line + 7));
            state = 3;
         }

      /*** In state 3, record remark or advance state on '.'. ***/

      } else if (3 == state && '_' != line[7]) {
         if ('.' != line[7]) {
            if (NULL != remarks) scratchpad("  ");
            str     = du_trim(line + 7);
            lens    = strlen(str);
            remarks = scratchpad(str);
            if (lens && '.' != str[lens - 1]) remarks = scratchpad(".");
         }

      /*** Check grade. ***/

      } else if ('F' != line[8] && 'B' != line[8]) {
         oops("expected price information", line);

      /*** Lines starting "_" contain price information. ***/

      } else {
         cent       = new_cent();
         cent->next = a->cents;
         a->cents   = cent;

         /*** Extract trimmed and possibly quoted string values. ***/

         cent->supplier = (supplier && *supplier ? quotedup(supplier) : NULL);
         cent->catnam   = (catnam   && *catnam   ? quotedup(catnam  ) : NULL);
         cent->upcode   = (upcode   && *upcode   ? quotedup(upcode  ) : NULL);
         cent->remarks  = (remarks  && *remarks  ? quotedup(remarks ) : NULL);


         cent->purity   = quotedup(du_trim(line +  63));  line[62] = '\0';
         cent->price    = quotedup(du_trim(line +  51));  line[50] = '\0';
         cent->currency = quotedup(du_trim(line +  47));  line[46] = '\0';
         cent->unitname = quotedup(du_trim(line +  36));  line[35] = '\0';
         cent->amount   = quotedup(du_trim(line +  26));  line[25] = '\0';
         cent->catno    = quotedup(du_trim(line +  12));  line[ 9] = '\0';
         cent->grade    = quotedup(du_trim(line +   7));
         cent->usdpg    = NULL;

         savename(a, cent->catno);

         /*** Translate grade. ***/

         if (NULL != cent->grade && '_' == cent->grade[0]) {
            if ('B' == cent->grade[1]) {
               du_free(cent->grade);
               cent->grade = du_strdup("Bulk");
            } else if ('F' == cent->grade[1]) {
               du_free(cent->grade);
               cent->grade = du_strdup("Fine");
            }
         }

         /*** Treat "POA" price (price on application) as missing. ***/

         if (NULL != cent->price && 0 == strcmp("POA", cent->price)) {
            du_free(cent->price);
            cent->price = NULL;
         }
      }
   }
}

/*============================================================================
 *  cmpcent
 */

static int cmpcent(CENT *c1, CENT *c2)
{
   char   *p1, *p2, *x;
   float   d1,  d2;
   int     ok1, ok2;
   double  strtod(const char *nptr, char **endptr);

   /*** NULL pointers are greater than anything else. ***/

   if (NULL == c1 || NULL == c2) {
      if (NULL == c1 && NULL == c2) return 0;
      if (NULL == c1) return 1;
      return -1;
   }

   /*** Missing values are greater than non-missing values. ***/

   p1 = c1->usdpg;
   p2 = c2->usdpg;
   if (NULL == p1 || NULL == p2) {
      if (NULL == p1 && NULL == p2) return 0;
      if (NULL == p1) return 1;
      return -1;
   }

   /*** Uninterpretable values are greater than interpretable values. ***/

   d1 = strtod(p1, &x);  ok1 = (x != p1);
   d2 = strtod(p2, &x);  ok2 = (x != p2);
   if (!ok1 || !ok2) {
      if (!ok1 && !ok2) return 0;
      if (!ok1) return 1;
      return -1;
   }

   /*** Else return value comparison. ***/

   if (d1 < d2) return -1;
   if (d1 > d2) return  1;
   return 0;
}

/*============================================================================
 *  zap_entry() -- deallocate everything in a current entry.
 */

static void zap_entry(A *a)
{
   CENT     *cent, *nextcent;
   NAMEITEM *nit,  *nextnit;

   /*** No-op if called with NULL. ***/

   if (NULL == a) return;

   /*** Deallocate molecule (and dependent depiction). ***/

   dt_dealloc(a->mol);
   a->mol = NULL_OB;
   a->dep = NULL_OB;

   /*** Deallocate strings. ***/

   du_free(a->acd);  a->acd = NULL;
   du_free(a->cas);  a->cas = NULL;
   du_free(a->pn);   a->pn  = NULL;
   du_free(a->amw);  a->amw = NULL;

   /*** Deallocate catalog entries. ***/

   for (cent = a->cents; cent; cent = nextcent) {
      nextcent = cent->next;
      zap_cent(cent);
   }
   a->cents = NULL;

   /*** Deallocate cas number (namelist). ***/

   for (nit = a->casnos; nit; nit = nextnit) {
      nextnit = nit->next;
      du_free(nit->name);
      du_free(nit->std);
      du_free(nit);
   }
   a->casnos = NULL;

   /*** Deallocate other names (namelist). ***/

   for (nit = a->names; nit; nit = nextnit) {
      nextnit = nit->next;
      du_free(nit->name);
      du_free(nit->std);
      du_free(nit);
   }
   a->names = NULL;
}


/*============================================================================
 *  write_stdnames() -- output sorted names in list, return number written
 */

static int write_stdnames(FILE *fpout, NAMEITEM  *names, char *tag, char *dat)
{
   int        nnames, nout = 0;
   NAMEITEM **anames, **pnames, *nit;

   /*** Return 0 if called with NULL. ***/

   if (NULL == names || NULL == tag || '\0' == *tag) return 0;

   /*** Make pointer array for sorting. ***/

   for (nnames = 0, nit = names; nit; nit = nit->next)
      nnames++;

   anames = (NAMEITEM **) malloc(nnames * sizeof(NAMEITEM *));

   for (pnames = anames, nit = names; nit; nit = nit->next)
      *pnames++ = nit;

   /*** Sort 'em. ***/

   du_mergesort_ptr(nnames, (void**)anames, NULL, (FUNCTION)cmpstd);

   /*** Add to tree in sorted order. (non-IDs if no molecule) ***/

   for (pnames = anames; nnames--; pnames++) {
      nit = *pnames;
      if (NULL != nit->std) {
         fprintf(fpout, "%s<%s>\n", tag, nit->std);
         if (dat && *dat) fprintf(fpout, "%s\n", dat);
	 nout++;
      }
   }

   /*** Clean up and return number of items written. ***/

   du_free(anames);
   return nout;
}

/*============================================================================
 *  writetdt() -- write acd tdt (doesn't dealloc anything except local stuff)
 *
 *   $SMI<>
 *      $ACD<number>
 *         PN<preferred name>
 *         ISM<>
 *         2D<>
 *         CAT<catno;supplier;name;grade;price;dpg;remark;updated>
 *         CAT<catno;supplier;name;grade;price;dpg;remark;updated>
 *      $CAS<>
 *         ECN<erroneous cas number>
 *         ACD<>
 *      $NAM<>
 *   |
 *
 *   $ACD<number>
 *      PN<preferred name>
 *      ISM<>
 *      2D<>
 *      CAT<catno;supplier;name;grade;purity;dpg;price;remark;updated>
 *   |
 *   $CAS<>
 *      ECN<erroneous cas number>
 *      ACD<>
 *   |
 *   $NAM<>
 *      ACD<>
 */

static void writetdt(A *a)
{
   int     hasmol = (NULL_OB != a->mol);
   int     eot    = FALSE;
   char   *smi, *usmi, *ism, buf80[80], amwoop[80];
   float   dpg;
   double  amw, cmw, dmw;
   CENT   *cent, **scent, **ascent;
   int     i, lens, nonsense, nnames, ncats;
   FILE   *fpout;

   /*** Write to fpout1 (has molecule) or fpout2 (no molecule). ***/

   fpout = (hasmol ? a->fpout1 : a->fpout2);

   /*** If just-completed entry has molecule, write $SMI<> root. ***/

   ism     = NULL;
   *amwoop = '\0';

   if (hasmol) {

      /*** Check average molecular weight. ***/

      if (a->amw) {
         amw = atof(a->amw);
	 cmw = rd_amw(a->mol);
         dmw = amw - cmw;
	 if (MX_SLOP < ABS(dmw)) {
            smi = dt_cansmiles(&lens, a->mol, 1);
	    sprintf(buf80, "%s: amw %s != cmw %.3f (dmw: %.3f)",
	       a->acd, a->amw, cmw, dmw);
	    oops("molecular weight mismatch", buf80);
	    oops(smi, NULL);
	    strcpy(amwoop, buf80);
	 }
      }

      /*** Write unique SMILES root. ***/

      usmi = dt_cansmiles(&lens, a->mol, 0);
      if (NULL == usmi) {
	 hasmol = FALSE;
	 oops("can't make smiles", NULL);

      } else {
	 fprintf(fpout, "$SMI<%.*s>\n", lens, usmi);
	 a->nids++;
	 eot = TRUE;

         /*** Make isomeric SMILES if needed. ***/

         if (NULL != (smi = dt_cansmiles(&lens, a->mol, 1)))
            ism = ((0 == strcmp(smi, usmi)) ? NULL : du_strndup(smi, lens));
      }
   }

   /*** Write acd number, preferred name, amw, ism, and 2D as needed. ***/

   if (a->acd && *a->acd) {
      fprintf(fpout, "$ACD<%s>\n", a->acd);
      eot = TRUE;
      if (a->pn)  fprintf(fpout, "PN<%s>\n",  a->pn );
      if (a->amw) fprintf(fpout, "AMW<%s>\n", a->amw);
      if (*amwoop) fprintf(fpout, "WARN<%s>\n", amwoop);
      if (ism)    fprintf(fpout, "ISM<%s>\n", ism   );
      fprint2d(fpout, a->dep);

      /*** If no molecule, $ACD forms a complete tree. ***/

      a->nids++;
      if (!hasmol) { eot = FALSE; a->ntrees++; fprintf(fpout, "|\n"); }
   }
   
   /*** Output price info. ***/

   if (NULL != a->cents) {

      /*** Compute USD per gram if able. ***/

      for (cent = a->cents; cent; cent = cent->next) {
         dpg = acd_convert2dpg(a->units, cent->price, cent->currency,
                               cent->amount, cent->unitname, a->amw);
         if (0.0 >= dpg)
            cent->usdpg = NULL;
         else if (0 != sprintf(buf80, "%.4g", dpg))
            cent->usdpg = du_strdup(du_trim(buf80));
         else
            cent->usdpg = du_strdup("?");
      }

      /*** Copy list of CENTs to pointer array. ***/

      for (nonsense = 0, cent = a->cents; cent; cent = cent->next) nonsense++;
      ascent = (CENT **) malloc(nonsense * sizeof(CENT *));
      for (scent = ascent, cent = a->cents; cent; cent = cent->next)
         *scent++ = cent;

      /*** Sort by price. ***/

      du_mergesort_ptr(nonsense, (void**)ascent, NULL, (FUNCTION)cmpcent);

      /*** Add to tree in sorted order. ***/

      for (scent = ascent; nonsense--; scent++) {
         cent = *scent;
         if (cent->catno && *cent->catno) {

            /*** If no molecule, add $CAT id. ***/

	    if (!hasmol)
               fprintf(fpout, "$CAT<%s>\n", cent->catno);

            /*** Summarize price as single string, write price & usdpg. ***/

            if (cent->price && cent->currency && cent->amount && cent->unitname)
	       sprintf(buf80, "%s %s per %s %s",
		  cent->price,  cent->currency, cent->amount, cent->unitname);
            else
	       *buf80 = '\0';

            /*** Write catalog number as non-ID. ***/

            fprintf(fpout, "CAT<%s", cent->catno         );
            fprintf(fpout, ";%s",    SELF(cent->supplier));
	    fprintf(fpout, ";%s",    SELF(cent->catnam  ));
	    fprintf(fpout, ";%s",    SELF(cent->grade   ));
	    fprintf(fpout, ";%s",    SELF(cent->purity  ));
	    fprintf(fpout, ";%s",    buf80               );
	    fprintf(fpout, ";%s",    SELF(cent->usdpg   ));
	    fprintf(fpout, ";%s",    SELF(cent->remarks ));
	    fprintf(fpout, ";%s",    SELF(cent->upcode  ));
	    fprintf(fpout, ">\n");

         }

	 /*** If no molecule, this is a tree. ***/

	 if (!hasmol) {
            if (a->acd && *a->acd)
               fprintf(fpout, "ACD<%s>\n", a->acd);
            if (NULL != a->casnos)
	       write_stdnames(fpout, a->casnos, "CAS", NULL);
            fprintf(fpout, "|\n");
	    a->ntrees++;
	    eot = FALSE;
	 }
      }
      du_free(ascent);
   }

   /*** Add identifiers xrefs if molecule available. ***/

   if (hasmol) {
      if (a->acd && *a->acd)
         sprintf(buf80, "ACD<%s>", a->acd);
      else
         *buf80 = '\0';

      /*** Output cas numbers in sorted order. ***/

      if (NULL != a->casnos)
	 a->nids += write_stdnames(fpout, a->casnos, "$CAS", buf80);

      /*** Output all other names in sorted order. ***/

      if (NULL != a->names)
	 a->nids += write_stdnames(fpout, a->names, "$NAM", buf80);

      eot = TRUE;

   /*** Add pseudo-xrefs if no molecule available. ***/

   } else if (a->acd && *a->acd) {
      sprintf(buf80, "ACD<%s>\n|", a->acd);

      /*** Output cas numbers in sorted order. ***/

      if (NULL != a->casnos)
	 a->ntrees += write_stdnames(fpout, a->casnos, "$CAS", buf80);

      /*** Output all other names in sorted order. ***/

      if (NULL != a->names)
	 a->ntrees += write_stdnames(fpout, a->names, "$NAM", buf80);

      eot = FALSE;
   }

   /*** Add end-of-tree delimiter if anything written out. ***/

   if (eot) { fprintf(fpout, "|\n");  a->ntrees++; }

   /*** Clean up. ***/

   du_free(ism);
}

/*============================================================================
 *  acd2tdt main program
 */

main(int argc, char **argv)
{
   A          *a;
   char       *line, *p, buf80[80];
   char       *bf    = NULL;
   int        i, nin = 0;

   /*** Initialize program. ***/

   a = proginit(&argc, argv);

   /*** Read unit data. ***/
   fprintf(stderr, "Opening file \"units.dat\"...\n");
   a->units = acd_read_units("units.dat");
   if (NULL == a->units) { fprintf(stderr, "can't read units\n"); exit(3); }

   /*** Read rdf data. ***/

   while (NULL != (line = getline(a->fpin))) {
      printerrors(stderr, DX_ERR_ERROR);
      if (0 == strncmp("$MFMT", line, 5)) {

         /*** Note progress every 1000. ***/

         if (nin && 0 == (nin % 1000)) progress(nin);
         nin++;

         /*** Save MIREG and reset line number. ***/

         strcpy(g_mireg, du_trim(line + 7));
         g_lineno = 0;

         /** Write previous entry. ***/

         if (0 < nin) writetdt(a);
         zap_entry(a);

         /*** Skip 3 header records (should be 2!). ***/

         getline(a->fpin); getline(a->fpin); getline(a->fpin);

         /*** Interpret molecule. ***/

         a->dep  = ctab2dep(a->fpin);
         a->mol  = dt_base(a->dep);

      } else if (0 == strncmp("$DTYPE MOL:SUPPLIER(", line, 20)) {
         deal_with_supplier(a);

      /*** Ignore SYMBOLs (dealt with as catalog names). ***/

      } else if (0 == strncmp("$DTYPE MOL:SYMBOL(", line, 18)) {
         continue;

      /*** Save MOLNAME as "Preferred name". ***/

      } else if (0 == strncmp("$DTYPE MOL:MOLNAME", line, 18)) {
         a->pn = quotedup(du_white1(du_trim(getline(a->fpin) + 7)));
	 savename(a, a->pn);

      /*** Save CAS number as entry property. ***/

      } else if (0 == strncmp("$DTYPE MOL:CAS(", line, 15)) {
         for (p = getline(a->fpin) + 7; '0' == *p; p++)
            ;
         savecas(a, du_trim(p));

      /*** Save cleaned-up MDLNUMBER as ACD number. ***/

      } else if (0 == strncmp("$DTYPE MOL:MDLNUMBER", line, 20)) {
         line = getline(a->fpin);
	 if (0 != strncmp("MFCD", line + 7, 4)) {
            a->acd = quotedup(getline(a->fpin) + 7);
	 } else {
	    for (p = line + 11; '0' == *p; p++)
	       ;
	    sprintf(buf80, "%s", p);
            a->acd = quotedup(buf80);
	 }
         /* if (a->acd && '\0' != a->acd) savename(a, a->acd); redundant */

      /*** Save MOLECULAR WEIGHT as molecular property. ***/

      } else if (0 == strncmp("$DTYPE MOL:MOLECULAR.WEIGHT", line, 27)) {
         a->amw = quotedup(getline(a->fpin) + 7);

      /*** Save BULK.FINE, but why? ***/

      } else if (0 == strncmp("$DTYPE MOL:BULK.FINE(", line, 15)) {
         p = getline(a->fpin) + 7;
         if      ('F' == *p) { bf = ('B' == *(p+1)) ? "fine, bulk" : "fine"; }
         else if ('B' == *p) { bf = ('F' == *(p+1)) ? "fine, bulk" : "bulk"; }
         else                { bf = NULL;                                    }
      }
   }

   /** Write final entry. ***/

   if (0 < nin) writetdt(a);
   zap_entry(a);

   /*** Report. ***/

   printerrors(stderr, DX_ERR_ERROR);
   fprintf(stderr, "%d trees, %d ids, %d xrefs\n",
           a->ntrees, a->nids, a->nids - a->ntrees);

   /*** Clean up. ***/

   acd_zap_units(a->units); a->units = NULL;
   du_free(a);

   /*** Exit with 0. ***/

   exit(0);
}

