/*****************************************************************************
*  acd_units.c -- support for acd unit conversions
*
*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <memory.h>
#include "du_utils.h"
#include "acd2tdt.h"

/*** Unit conversions. ***/

#define UNIT_GRAM 1
#define UNIT_ML   2
#define UNIT_MOLE 3
#define UNIT_USD  4

typedef struct _unit {
   char         *name;
   float         per;
   int           type;
   struct _unit *next;
} UNIT;

/*============================================================================
 *  acd_read_readunits
 */

void *acd_read_units(char *fn)
{
   UNIT   *unit, *units = NULL;
   char    buf[80], *type;
   double  din;
   FILE   *fp;

   /*** Read input line, skip comments. ***/

   if (NULL == (fp = fopen(fn, "r"))) {
      fprintf(stderr, "can't open file %s\n", fn);
      return NULL;
   }

   /*** Read input line, skip comments. ***/

   while (NULL != du_fgetz(buf, 80, fp)) {
      if ('#' == *buf) continue;
      unit = (UNIT *) malloc(sizeof(UNIT));

      /*** First 10 characters contain unit name. ***/

      buf[10] = '\0';
      unit->name = du_strdup(du_trim(buf));

      /*** Next field contains unit multipler. ***/

      /* sscanf(buf + 11, "%lf", &din); */
      unit->per = atof(buf + 11);

      /*** Type flag in ntains unit multipler. ***/

      type = buf + 23;
      if      (0 == strncmp(type, "GRAM", 4)) unit->type = UNIT_GRAM;
      else if (0 == strncmp(type, "ML"  , 2)) unit->type = UNIT_ML;
      else if (0 == strncmp(type, "MOLE", 4)) unit->type = UNIT_MOLE;
      else if (0 == strncmp(type, "USD",  3)) unit->type = UNIT_USD;
      else { fprintf(stderr, "bad unit type: %s\n", type); continue; }

      unit->next = units;
      units      = unit;
   }

   /*** Close file and return list of UNITs. ***/

   fclose(fp);
   return (void *) units;
}

/*============================================================================
 *  acd_zap_units() -- deallocate UNIT list
 */

void acd_zap_units(void *units)
{
   UNIT *unit, *nextunit;

   for (unit = (UNIT *) units; unit; unit = nextunit) {
      nextunit = unit->next;
      du_free(unit->name);
      du_free(unit);
   }
}

/*===========================================================================
 *  acd_find_unit() -- find unit by name
 */

static void *findunit(void *units, char *name)
{
   UNIT *unit;
   for (unit = (UNIT *) units; unit; unit = unit->next)
      if (0 == strcmp(unit->name, name)) return unit;
   return NULL;
}

/*============================================================================
 *  acd_convert2dpg() -- convert to US dollars per gram
 */

float acd_convert2dpg(void *units, char *price, char *currency,
                      char *amount, char *aunits, char *amw)
{
   UNIT    *curunit, *amtunit;
   float    p, a, dpg, mw;
   double   atof(const char *);
   char    *price2, *q, *qq;

   /*** Find currency, return 0.0 if unable. ***/

   if (NULL == currency || '\0' == *currency || ' ' == *currency) return 0.0;
   if (NULL == (curunit = findunit(units, currency))) return 0.0;

   /*** Find amount units, return 0.0 if unable. ***/

   if (NULL == aunits || '\0' == *aunits || ' ' == *aunits) return 0.0;
   if (NULL == (amtunit = findunit(units, aunits))) return 0.0;

   /*** Remove commas from price. ***/

   price2 = (price ? strdup(price) : NULL);
   if (price2) {
      for (q = qq = price2; *qq; qq++)
	 if (',' != *qq) *q++ = *qq;
       *q = '\0';
   }

   /*** Interpret input. ***/

   if (NULL == price2) return 0.0;
   p = atof(price2);
   du_free(price2);
   if (0.0 == p) return 0.0;

   if (NULL == amount || 0.0 == (a = atof(amount)  )) return 0.0;

   /*** Compute price in dollars per standard unit. ***/

   dpg = (p * curunit->per) / (a * amtunit->per);

   /*** Convert from density molar to grams. ***/

   switch (amtunit->type) {
   case UNIT_ML:
      /* dpg *= 0.7; */
      dpg = 0.0;       /* not now */
      break;
   case UNIT_MOLE:
      if (NULL == amw || 0.0 == (mw = atof(amw))) dpg  = 0.0;
      else                                        dpg /= mw;
   case UNIT_GRAM:
   default:
      break;
   }

   /*** Return dollars per gram. ***/

   return dpg;
}
