/*****************************************************************************
*  ctab2dep.c -- convert mdl's CTAB block (.mol, .rdf) to depiction
*
*      dt_setarborder(*aptr, i);
*         case 1:  dt_setadjunct(*aptr, parity1);  break;
*----------------------------------------------------------------------------
*
*----------------------------------------------------------------------------
*  TODO:
*   [x] deal with continuation lines +
*   [x] deal with chirality
*   [x] deal with double bond orientation
*   [ ] deal with singlet, triplet, non-ring radicals?
*   [x] test for leaks
*   [x] check against AMW
*   [x] deal with $NAM's?
*   [?] sort names?
*   [x] deal with casnos
*   [x] sort prices?
*   [x] add $CNO
*   [x] deal with ISM
*   [ ] deal with V2000
*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <memory.h>
#include <dt_smiles.h>
#include <dt_depict.h>
#include "acd2tdt.h"

/*** Error macros. ***/

#define TRUE 1
#define FALSE 0

#define FUNC        "ctab2dep"
#define OOPS(s)     dt_errorsave(FUNC, DX_ERR_ERROR, s)
#define WOOPS(s)    dt_errorsave(FUNC, DX_ERR_WARN,  s)

/*** Parity flags. ***/

static dt_Handle parity1 = NULL_OB;
static dt_Handle parity2 = NULL_OB;
static dt_Handle parity3 = NULL_OB;

/*============================================================================
 *  numbof()
 */

static char *symbol[106] = { "*",       /* Slot for wildcard atom. */
    "H",  "He", "Li", "Be", "B",  "C",  "N",  "O",  "F",  "Ne",
    "Na", "Mg", "Al", "Si", "P",  "S",  "Cl", "Ar", "K",  "Ca",
    "Sc", "Ti", "V",  "Cr", "Mn", "Fe", "Co", "Ni", "Cu", "Zn",
    "Ga", "Ge", "As", "Se", "Br", "Kr", "Rb", "Sr", "Y",  "Zr",
    "Nb", "Mo", "Tc", "Ru", "Rh", "Pd", "Ag", "Cd", "In", "Sn",
    "Sb", "Te", "I",  "Xe", "Cs", "Ba", "La", "Ce", "Pr", "Nd",
    "Pm", "Sm", "Eu", "Gd", "Tb", "Dy", "Ho", "Er", "Tm", "Yb",
    "Lu", "Hf", "Ta", "W",  "Re", "Os", "Ir", "Pt", "Au", "Hg",
    "Tl", "Pb", "Bi", "Po", "At", "Rn", "Fr", "Ra", "Ac", "Th",
    "Pa", "U",  "Np", "Pu", "Am", "Cm", "Bk", "Cf", "Es", "Fm",
    "Md", "No", "Lr", "Rf", "Ha"
};

static int numbof(char *as)
{
   int an;
   for (an = 0; an < 106; an++)
      if (0 == strcmp(as, symbol[an])) return an;
   return 0;
}

/*============================================================================
 *  symbof() -- return symbol for given atomic number
 *
 *  NOT USED
 */

static char *symbof(int an)
{
   return ((an > 0 && an < 106) ? symbol[an] : symbol[0]);
}

/*============================================================================
 *  normal_valence() -- return implicit valence given atomic number and vbond
 *
 *  NOT USED
 */

static int normal_valence(int atno, int vbond)
{
   int val;

   switch (atno) {
      /* H  */  case  1:  val = 1;  break;
      /* B  */  case  5:  val = 3;  break;
      /* C  */  case  6:  val = 4;  break;
      /* N  */  case  7:  val = vbond > 3 ? 5 : 3;  break;
      /* O  */  case  8:  val = 2;  break;
      /* F  */  case  9:  val = 1;  break;
      /* P  */  case 15:  val = vbond > 3 ? 5 : 3;  break;
      /* S  */  case 16:  val = vbond > 5 ? 6 : (vbond > 3 ? 4 : 6);  break;
      /* Cl */  case 17:  val = 1;  break;
      /* Br */  case 35:  val = 1;  break;
      /* I  */  case 53:  val = 1;  break;
                default:  val = 0;
   }

   return val;
}

/*============================================================================
 *   rd_valence() -- return valence for given atom, using .RD rules
 *
 *   This is derived from ACD structures (ignores the published MDL tables).
 */


static int rd_valence(dt_Handle atom, int vbond)
{
   int atno   = dt_number(atom);
   int charge = dt_charge(atom);
   int ncon   = dt_count(atom, TYP_BOND);
   int val;

   switch (dt_number(atom)) {
   case DX_ATN_WILD:
      val = 0;
      break;

   /*** Group 1a ***/

   case DX_ATN_H:  
   case DX_ATN_Li: 
   case DX_ATN_Na: 
   case DX_ATN_K:  
   case DX_ATN_Rb: 
   case DX_ATN_Cs: 
   case DX_ATN_Fr: 
      val = 1 - charge;
      break;

   /*** Group 2a ***/

   case DX_ATN_Be: 
   case DX_ATN_Mg: 
   case DX_ATN_Ca: 
   case DX_ATN_Sr: 
   case DX_ATN_Ba: 
   case DX_ATN_Ra: 
      val = 2 - charge;
      break;

   /*** Group 3a ***/

   case DX_ATN_Al: 
   case DX_ATN_Ga: 
      val = 3 - charge;
      break;

   case DX_ATN_B:   val = 3 - charge;  break;
   case DX_ATN_Tl:  val = 1 - charge;  break;
   case DX_ATN_In:  val = 0;           break;

   /*** Group 4a ***/

   case DX_ATN_C:  
   case DX_ATN_Si: 
   case DX_ATN_Ge: 
      val = 4 - ABS(charge);
      break;

   case DX_ATN_Sn:
      if (-1 == charge)   val = 3;
      else if (2 < vbond) val = 4;
      else                val = 2;
      break;

   case DX_ATN_Pb:  val = 2 - ABS(charge);            break;

   /*** Group 5a ***/

   case DX_ATN_N:  
   case DX_ATN_As: 
      val = 3 + charge;
      break;

   case DX_ATN_Bi:  val = 3 - charge;  break;
   case DX_ATN_Sb:  val = 3 - charge;  break;

   case DX_ATN_P:
      if      (1 == charge) val = 4;
      else if (3 <  vbond ) val = 5;
      else                  val = 3 - charge;
      break;

   /*** Group 6a ***/

   case DX_ATN_O:  
   case DX_ATN_Se: 
   case DX_ATN_Te: 
   case DX_ATN_Po: 
      val = 2 + charge;
      break;

   case DX_ATN_S:
      /* vbond += charge; */
      if (dt_aromatic(atom)) val = 2;  /* aromatic sulfur has 0 H's */
      else if ( 1 == charge) val = 2;
      else if (-1 == charge) val = 1;
      else if ( 4 <  vbond ) val = 6;
      else if ( 2 <  vbond ) val = 4;
      else                   val = 2;
      break;

   /*** Group 7a ***/

   case DX_ATN_F:  
   case DX_ATN_Cl: 
   case DX_ATN_Br: 
   case DX_ATN_I:  
   case DX_ATN_At: 
      val = 1 + charge;
      break;

   /*** Group 0 ***/

   case DX_ATN_He: 
   case DX_ATN_Ne: 
   case DX_ATN_Ar: 
   case DX_ATN_Kr: 
   case DX_ATN_Xe: 
   case DX_ATN_Rn: 
      val = 0;
      break;

   /*** Assume everything else has zero default valence. ***/

   case DX_ATN_Sc: 
   case DX_ATN_Ti: 
   case DX_ATN_V:  
   case DX_ATN_Cr: 
   case DX_ATN_Mn: 
   case DX_ATN_Fe: 
   case DX_ATN_Co: 
   case DX_ATN_Ni: 
   case DX_ATN_Cu: 
   case DX_ATN_Zn: 
   case DX_ATN_Y:  
   case DX_ATN_Zr: 
   case DX_ATN_Nb: 
   case DX_ATN_Mo: 
   case DX_ATN_Tc: 
   case DX_ATN_Ru: 
   case DX_ATN_Rh: 
   case DX_ATN_Pd: 
   case DX_ATN_Ag: 
   case DX_ATN_Cd: 
   case DX_ATN_La: 
   case DX_ATN_Ce: 
   case DX_ATN_Pr: 
   case DX_ATN_Nd: 
   case DX_ATN_Pm: 
   case DX_ATN_Sm: 
   case DX_ATN_Eu: 
   case DX_ATN_Gd: 
   case DX_ATN_Tb: 
   case DX_ATN_Dy: 
   case DX_ATN_Ho: 
   case DX_ATN_Er: 
   case DX_ATN_Tm: 
   case DX_ATN_Yb: 
   case DX_ATN_Lu: 
   case DX_ATN_Hf: 
   case DX_ATN_Ta: 
   case DX_ATN_W:  
   case DX_ATN_Re: 
   case DX_ATN_Os: 
   case DX_ATN_Ir: 
   case DX_ATN_Pt: 
   case DX_ATN_Au: 
   case DX_ATN_Hg: 
   case DX_ATN_Ac: 
   case DX_ATN_Th: 
   case DX_ATN_Pa: 
   case DX_ATN_U: 
   case DX_ATN_Np: 
   case DX_ATN_Pu: 
   case DX_ATN_Am: 
   case DX_ATN_Cm: 
   case DX_ATN_Bk: 
   case DX_ATN_Cf: 
   case DX_ATN_Es: 
   case DX_ATN_Fm: 
   case DX_ATN_Md: 
   case DX_ATN_No: 
   case DX_ATN_Lr: 
      val = 0;
      break;
   default:
      val = 0;
      break;
   };

   /*** Return valence but not less than zero. ***/

   return (0 < val ? val : 0);
}

/*============================================================================
 *  rd_normh() -- return normal H-count for given atom, using .RD rules
 *
 *  Figures normal H-count based on atom's number, charge and connectivity.
 */

static int rd_normh(dt_Handle atom, int fixval)
{
   int       val, abond, vbond, atno, hcount;
   dt_Handle bond, bonds;

   /*** Count bond equivalents. ***/

   vbond = 0;
   abond = 0;
   bonds = dt_stream(atom, TYP_BOND);
   while (NULL_OB != (bond = dt_next(bonds))) {
      switch (dt_bondtype(bond)) {
         case DX_BTY_SINGLE:  vbond += 1;  break;
         case DX_BTY_DOUBLE:  vbond += 2;  break;
         case DX_BTY_TRIPLE:  vbond += 3;  break;
         case DX_BTY_AROMAT:  vbond += 1;  abond += 1;  break;
      }
   }
   if (0 < abond) vbond++;
   dt_dealloc(bonds);

   /*** Get valence if needed. ***/

   atno = dt_number(atom);
   val  = (0 > fixval ? rd_valence(atom, vbond) : fixval);
   
#ifdef OLDWAY
   /*** Deal with special cases ***/

   switch (atno) {
      /* O-[PH2]=O (hypophosphite) is not handled correctly */
      case DX_ATN_P:   if (3 < vbond) val = 5 - dt_charge(atom);  break;
      case DX_ATN_Sn:  if (2 < vbond) val = 4 - dt_charge(atom);  break;
   }
#endif

   /*** Return outstanding valence. ***/

   return (MAX(0, val - vbond));
}

/*============================================================================
 *   radical_rings()
 *
 *   Operates in mod-on mode
 */

static int radical_rings(dt_Handle mol, dt_Handle ratoms)
{
   int       *radat;
   dt_Handle  atoms, atom, bonds, bond, cycles, cycle, acharge;

   /*** Create and zero array corresponding to atoms. ***/

   radat = (int *) malloc(dt_uidrange(mol, TYP_ATOM) * sizeof(int));
   atoms = dt_stream(mol, TYP_ATOM);
   while (NULL_OB != (atom = dt_next(atoms)))
      radat[dt_uid(atom)] = FALSE;
   dt_dealloc(atoms);

   /*** Mark radical atoms. ***/

   dt_reset(ratoms);
   while (NULL_OB != (atom = dt_next(ratoms)))
      radat[dt_uid(atom)] = TRUE;

   /*** Loop over cycles in molecule. ***/

   cycles = dt_stream(mol, TYP_CYCLE);
   while (NULL_OB != (cycle = dt_next(cycles))) {

      /*** Loop over atoms in ring. ***/

      atoms = dt_stream(cycle, TYP_ATOM);
      while (NULL_OB != (atom = dt_next(atoms)))
         if (FALSE == radat[dt_uid(atom)]) break;

      /*** If cycle has all radical atoms, it's a radical cycle. ***/

      if (NULL_OB == atom) {

         /*** Set aromatic bonds. ***/

         bonds = dt_stream(cycle, TYP_BOND);
         while (NULL_OB != (bond = dt_next(bonds)))
            dt_setbondtype(bond, DX_BTY_AROMAT);
         dt_dealloc(bonds);

         /*** Set aromatic atoms. ***/

	 acharge = NULL_OB;
         dt_reset(atoms);
         while (NULL_OB != (atom = dt_next(atoms))) {
            dt_setaromatic(atom, TRUE);
            dt_setimp_hcount(atom, rd_normh(atom, -1));
	    if (NULL_OB == acharge && 0 == dt_charge(atom)) acharge = atom;
         }

         /*** Now we have cycles with all "radical" atoms: set charge ***/

	 if (acharge) {
            switch (dt_count(cycle, TYP_ATOM)) {
               case 5:  dt_setcharge(acharge, -1);  break;
               case 7:  dt_setcharge(acharge,  1);  break;
            }
         }
      }
      dt_dealloc(atoms);
   }
   dt_dealloc(cycles);

   du_free(radat);
   return TRUE;
}

/*============================================================================
 *  ifixed() -- return integer in fixed format (0-origin pos)
 *
 *  This doesn't check that pos <= strlen(str).
 */

static int ifixed(char *str, int pos, int len)
{
   char buf[20];
   strncpy(buf, str + pos, len);
   buf[len] = '\0';
   return atoi(buf);
}

/*============================================================================
 *  ffixed() -- return float in fixed format (0-origin pos)
 */

static float ffixed(char *str, int pos, int len)
{
   char buf[20];
   strncpy(buf, str + pos, len);
   buf[len] = '\0';
   return atof(buf);
}

/*============================================================================
 *  parity2chirality() -- convert MDL parity to tetrahedral chirality
 *----------------------------------------------------------------------------
 *        3              3            Where:
 *        |  parity      |  parity      * = chiral atom
 *        *   = 1        *   = 2        4 is behind *
 *       / \            / \             1234 is MOL-order
 *      2   1          1   2            4 is imp H if present
 *
 *     4[*@@]321        4[*@]321  <- SMILES equivalent
 *
 *  Note: parity = 0 -> not chiral; = 3 -> either/unmarked
 */
 
static dt_Integer parity2chirality(dt_Handle atom, dt_Integer parity)
{
   dt_Handle bond, bonds, chibond, chibonds, xatom;
   int       imph, ncon, chirality;
 
   /*** Return DX_CHI_NONE if chirality not specified. ***/

   if (1 != parity && 2 != parity) return DX_CHI_NONE;

   /*** Can't be chiral if more than one implicit hydrogen. ***/

   if (1 < (imph = dt_imp_hcount(atom))) return DX_CHI_NONE;

   /*** Can't be chiral if not four-connected. ***/

   ncon = dt_count(atom, TYP_BOND);
   if (4 != ncon + imph) return DX_CHI_NONE;
 
   /*** Make sequence of bonds in input order. ***/

   chibonds = dt_alloc_seq();
   bonds    = dt_stream(atom, TYP_BOND);
   while (NULL_OB != (bond = dt_next(bonds))) {
      xatom = dt_xatom(atom, bond);
      dt_reset(chibonds);
      while (NULL_OB != (chibond = dt_next(chibonds))) {
         if (chibond != bond) {
            if (dt_arborder(dt_xatom(atom, chibond)) > dt_arborder(xatom)) {
               dt_insert(chibonds, bond);
               break;
            }
         }
      }
      if (NULL_OB == chibond) dt_append(chibonds, bond);
   }
 
   /*** Add implicit hydrogen at end if needed. ***/

   if (imph) dt_append(chibonds, dt_isohydro());

   /*** Assign chirality. ***/

   chirality = ((1 == parity) ? DX_CHI_THCW : DX_CHI_THCCW);
   dt_setchival(atom, chibonds, chirality);
   dt_dealloc(chibonds);
   return chirality;
}

/*============================================================================
 *  setcistrans() -- CIS/TRANS assignment
 *
 *     aa0     aa1
 *       \     /
 *       a0 = a1
 *
 *  Look for double bonded atoms that have at least one single bonded atom,
 *  check coordinates for CIS/TRANS.  In x-y plane, consider line through a0
 *  and a1: if aa0 and aa1 are on same side set, CIS; if opposite, TRANS.
 */

static dt_Integer setcistrans(dt_Handle dep)
{
   dt_Handle  mol, dbond, dbonds, bond, bonds, atoms, results, iob;
   dt_Handle  a0, aa0, a1, aa1, b0, b1;
   dt_Real    x0, x1, xx0, xx1, y0, y1, yy0, yy1, z, m, b;
   dt_Integer ndbs, dboval, ok = TRUE;

   /*** Reject trivial cases. ***/

   if (NULL_OB == (mol = dt_base(dep))) return TRUE;

   /*** Check for acyclic double bonds. ***/

   ndbs   = 0;
   dbonds = dt_alloc_seq();
   bonds  = dt_stream(mol, TYP_BOND);
   while (NULL_OB != (bond = dt_next(bonds))) {
      if (2 == dt_bondtype(bond) && 0 == dt_count(bond, TYP_CYCLE)) {
         dt_append(dbonds, bond);
         ndbs++;
      }
   }
   dt_dealloc(bonds);

   /*** No-op if there aren't any acyclic double bonds. ***/

   if (0 == ndbs) { dt_dealloc(dbonds);  return TRUE; }

   /*** Create sequence to hold temporary results. ***/

   results = dt_alloc_seq();

   /***  Loop over double bonds. ***/

   ndbs = 0;
   dt_reset(dbonds);
   while (NULL_OB != (dbond = dt_next(dbonds))) {

      /*** Find end atoms. ***/

      atoms = dt_stream(dbond, TYP_ATOM);
      a0 = dt_next(atoms);
      a1 = dt_next(atoms);
      dt_dealloc(atoms);

      /*** Find atom aa0 attached to a0. ***/

      bonds = dt_stream(a0, TYP_BOND);
      if (2 == dt_bondtype(bond = dt_next(bonds))) bond = dt_next(bonds);
      aa0 = (NULL_OB == bond || 1 != dt_bondtype(bond)) ? NULL_OB
							: dt_xatom(a0, bond);
      dt_dealloc(bonds);

      /*** Find atom aa1 attached to a1. ***/

      bonds = dt_stream(a1, TYP_BOND);
      if (2 == dt_bondtype(bond = dt_next(bonds))) bond = dt_next(bonds);
      aa1 = (NULL_OB == bond || 1 != dt_bondtype(bond)) ? NULL_OB
							: dt_xatom(a1, bond);
      dt_dealloc(bonds);

      /*** No dbo for this bond if connected atoms not found. ***/

      if (NULL_OB == aa0 || NULL_OB == aa1) continue;

      /*** Get coordinates. ***/

      dt_getcoord(dep, a0,  &x0,  &y0,  &z);
      dt_getcoord(dep, aa0, &xx0, &yy0, &z);
      dt_getcoord(dep, a1,  &x1,  &y1,  &z);
      dt_getcoord(dep, aa1, &xx1, &yy1, &z);

     /*
      *  y = mx + b is line through a1 and a1
      *  where m = (y1 - y0) / (x1 -x0), b = y0 - (m) * x0
      *
      *  (xx0,yy0) is cis to (xx1,yy1) IFF
      *  (yy0 - (m*xx0 + b)) * (yy1 - (m*xx1 + b)) > 0
      */

      m      = (y1 - y0) / (x1 - x0);
      b      =  y0 - (m * x0);
      dboval = ( ( ( (yy0 - (m * xx0 + b)) * (yy1 - (m * xx0 + b)) ) > 0 )
               ? DX_CHI_CIS : DX_CHI_TRANS);

      /*** Save results. ***/

      dt_append(results, dbond           );
      dt_append(results, dt_bond(a0, aa0));
      dt_append(results, dt_bond(a1, aa1));
      iob = dt_alloc_integer();
      dt_setintegervalue(iob, dboval);
      dt_append(results, iob);
      ndbs++;
   }

   /*** Assign double bond orientation if any were deduced. ***/

   if (0 < ndbs) {
      dt_mod_on(mol);
      dt_reset(results);
      while (NULL_OB != (dbond = dt_next(results))) {
         b0     = dt_next(results);
         b1     = dt_next(results);
	 iob    = dt_next(results);
	 dboval = dt_integervalue(iob);
	 if (0 != dboval) dt_setdbo(dbond, b0, b1, dboval);
	 dt_dealloc(iob);
      }

      /*** Return value is from dt_mod_off(). ***/

      ok = dt_mod_off(mol);
   }

   /*** Clean up. ***/

   dt_dealloc(dbonds);
   dt_dealloc(results);

   /*** Return TRUE unless dt_mod_off() failed. ***/

   return ok;
}


/*============================================================================
 *  ctab2dep() -- interpret RD file's Ctab block to a depiction
 *
 *  This expects input stream fp to be positioned at the Ctab block (i.e., at
 *  the count block), reads Ctab block (i.e., count, atom, bond, properties,
 *  blocks), and leaves the stream positioned after the end block ("M  END").
 *
 *  This returns NULL on EOF or error (if error, messages will be queued).
 */

#define MX_CTLINE 512

dt_Handle ctab2dep(FILE *fp)
{
   dt_Handle    dep, mol, atom, adj;
   dt_Handle    rads;                             /* radical sequences */
   dt_Handle   *alist, *aptr;                     /* raw atom array, ptr */
   dt_Handle   *blist, *bptr;                     /* raw atom array, ptr */
   int         *vlist;                            /* special valence array */
   static char  line[MX_CTLINE+1];                /* Ctab block input line */
   char         as[3];                            /* 2 char atomic symbol */
   int          n, ne, i, j, ok, lens, ifchiral, nprop, size;
   int          nb, end1, end2, order, bster;
   int          na, ia, number, hcount, charge, weight, mass;
   int v, parity;
   float *xlist, *ylist;

   /*** Make global parity flags. ***/

   if (NULL == parity1) {
      parity1 = dt_alloc_string(2, "P1");
      parity2 = dt_alloc_string(2, "P2");
      parity3 = dt_alloc_string(2, "P3");
   }

   /*** Read counts. ***/

   if (NULL == du_fgetz(line, MX_CTLINE, fp)) return NULL;
   na       = ifixed(line,  0, 3);
   nb       = ifixed(line,  3, 3);
   ifchiral = ifixed(line, 12, 3);
   nprop    = ifixed(line, 30, 3);
   if (1 > na) return NULL;

   /*** Allocate molecule & atom array, return unhappy if out of memory. ***/

   mol   = dt_alloc_mol();
   alist = (dt_Handle *) malloc(na * sizeof (dt_Handle));
   blist = (0 == nb ? NULL : (dt_Handle *) malloc(nb * sizeof (dt_Handle)));
   xlist = (float     *) malloc(na * sizeof (float)    );
   ylist = (float     *) malloc(na * sizeof (float)    );
   vlist = NULL;
   if (NULL_OB == mol || !alist || !xlist || !ylist) {
      OOPS("out of memory creating temporary space");
      goto errorexit;
   }
   dt_mod_on(mol);

   /*** Allocate sequence to contain radical atoms. ***/

   rads = dt_alloc_seq();

   /*** Read in atoms and their attributes. ***/

   for (i = 0, aptr = alist; i < na; i++, aptr++) {
      if (NULL == du_fgetz(line, MX_CTLINE, fp)) goto eofexit;
      xlist[i]  = ffixed(line,  0, 10);
      ylist[i]  = ffixed(line, 10, 10);
      weight    = ifixed(line, 33,  3);
      charge    = ifixed(line, 36,  3);
      parity    = ifixed(line, 39,  3);

      if (0 != (v = ifixed(line, 48, 3))) {
         if (NULL == vlist) {
	    size  = na * sizeof(int);
	    vlist = (int *) malloc(size);
	    memset(vlist, 0, size);
	 }
	 vlist[i] = v;
/* if (vlist[i]) WOOPS("valence specified"); */
      }

      /*** Interpret atomic symbol, with D and T exceptions ***/

      as[0] = line[31];
      as[1] = line[32];
      as[2] = '\0';

      mass = 0;
      if (' ' == as[1]) {
         if      ('D' == as[0]) { as[0] = 'H';  mass = 2; }
         else if ('T' == as[0]) { as[0] = 'H';  mass = 3; }
         as[1] = '\0';
      }
      number = numbof(as);

      /*** Add atom to molecule, save in list, save attributes. ***/

      *aptr = dt_addatom(mol, number, 0);

      /*** Interpret charge. ***/

      switch (charge) {
         case 0:  charge =  0;                             break;
         case 1:  charge =  3;                             break;
         case 2:  charge =  2;                             break;
         case 3:  charge =  1;                             break;
         case 4:  charge =  0;  dt_append(rads, *aptr); break;
         case 5:  charge = -1;                             break;
         case 6:  charge = -2;                             break;
         case 7:  charge = -3;                             break;
      }
      if (0 != charge) dt_setcharge(*aptr, charge);

      /*** Interpret atomic mass from weight difference. ***/

      if (0 != weight) mass = weight + normal_mass(*aptr);
      if (0 != mass)   dt_setweight(*aptr, mass);

      /*** Save input order as arborder. ***/

      dt_setarborder(*aptr, i);

      /*** Save atomic stereo parity as adjunct. ***/
      
      switch(parity) {
         case 0:  break;
         case 1:  dt_setadjunct(*aptr, parity1);  break;
         case 2:  dt_setadjunct(*aptr, parity2);  break;
         case 3:  dt_setadjunct(*aptr, parity3);  break;
         default: OOPS("unknown atomic stereo parity");
      }
   }

   /*** Read in bonds and their attributes. ***/

   for (i = 0, bptr = blist; i < nb; i++, bptr++) {
      if (NULL == du_fgetz(line, MX_CTLINE, fp)) goto eofexit;
      end1  = ifixed(line,  0, 3);
      end2  = ifixed(line,  3, 3);
      order = ifixed(line,  6, 3);

      *bptr = dt_addbond(alist[end1-1], alist[end2-1], order);

#ifdef OTHERNEWERWAY
      /*** sb: 1 = up, 4 = either, 6 = down, db: 0 coords, 3 = either **/

      bster = ifixed(line,  9, 3);
      switch (bster) {
         case 1: dt_setadjunct(*bptr, upbond);   break;
         case 6: dt_setadjunct(*bptr, downbond); break;
         case 4: OOPS("chiral bond EITHER"); break;
         case 3: OOPS("chiral bond DBO");    break;
      }
#endif
      /*** sb: 1 = up, 4 = either, 6 = down, db: 0 coords, 3 = either **/

      bster = ifixed(line,  9, 3);
      if (2 == order && 3 == bster) WOOPS("DBO=3 specified");
   }

   /*** Read in properties. ***/
   /*** was:  for (i = 0; i < nprop; i++) ***/

   for (i = 0; 1; i++) {
      if (NULL == du_fgetz(line, MX_CTLINE, fp)) goto eofexit;

      if (0 == strncmp("M  END", line, 6)) break;

      if (0 == strncmp("M  CHG", line, 6)) {
         n = ifixed(line,  7, 3);
         for (i = 0, j = 10; i < n; i++, j += 8) {
            ia     = ifixed(line, j,     3);
            charge = ifixed(line, j + 4, 3);
            if (0 != charge) dt_setcharge(alist[ia - 1], charge);
         }

      /*** Read radical atoms into sequences. ***/

      } else if (0 == strncmp("M  RAD", line, 6)) {
         n = ifixed(line,  7, 3);
         for (i = 0, j = 10; i < n; i++, j += 8) {
            ia = ifixed(line, j,     3);
            ne = ifixed(line, j + 4, 3);
	    switch (ne) {
	       case 1: case 2: case 3:
	          if (NULL_OB == rads) rads = dt_alloc_seq();
	          dt_append(rads, alist[ia - 1]);
	          break;
	    }
         }

      } else if (0 == strncmp("M  ISO", line, 6)) {
         n = ifixed(line,  7, 3);
         for (i = 0, j = 10; i < n; i++, j += 8) {
            ia   = ifixed(line, j,     3);
            mass = ifixed(line, j + 4, 3);
            if (-19 == mass) dt_setweight(alist[ia - 1], 0);
            else             dt_setweight(alist[ia - 1], mass);
         }
      }
   }

   /*** Set implicit H-counts (radical O might be special, 0 H's). ***/

   for (i = 0, aptr = alist; i < na; i++, aptr++) {
      if (NULL == vlist || 0 == vlist[i])
	 dt_setimp_hcount(*aptr, rd_normh(*aptr, -1));
      else if (15 > vlist[i])
	 dt_setimp_hcount(*aptr, rd_normh(*aptr, vlist[i]));
      else
	 dt_setimp_hcount(*aptr, 0);
   }

  /*** Radicals have no implicit hydrogens and 0 charge. ***/

   if (rads) {
      dt_reset(rads);
      while (NULL_OB != (atom = dt_next(rads))) {
         dt_setimp_hcount(atom, 0);
         dt_setcharge    (atom, 0);
      }
   }

   /*** Deal with atom stereo parity. ***/

   for (i = 0, aptr = alist; i < na; i++, aptr++) {
      if (NULL_OB != (adj = dt_adjunct(*aptr))) {
         ok = TRUE;
         if      (parity1 == adj) ok = parity2chirality(*aptr, 1);
         else if (parity2 == adj) ok = parity2chirality(*aptr, 2);
         if (0 == ok) WOOPS("failed to set chirality from parity");
      }
   }

   /*** Return NULL if molecule is invalid. ***/

   if (FALSE == dt_mod_off(mol)) {
      OOPS("can't interpret resultant molecule");
      goto errorexit;
   }

   /*** Deal with radical rings. ***/

   if (NULL_OB != rads) {
      dt_mod_on(mol);
      radical_rings(mol, rads);
      dt_dealloc(rads);
      rads = NULL_OB;
      if (FALSE == dt_mod_off(mol)) WOOPS("can't fix radical rings(s)");
   }

   /*** Create depiction, set x and y coordinates. ***/

   if (NULL == (dep = dt_alloc_depiction(mol))) {
      OOPS("out of memory creating depiction");
      goto errorexit;
   }

   for (i = 0; i < na; i++)
      dt_setcoord(dep, alist[i], xlist[i], ylist[i], 0.0);

   /*** Set cis and trans. ***/

   if (FALSE == setcistrans(dep)) WOOPS("failed to set CIS/TRANS");

   /*** Deallocate temporary sequence and arrays. ***/

   dt_dealloc(rads);
   du_free(alist);
   du_free(blist);
   du_free(xlist);
   du_free(ylist);
   du_free(vlist);

   /*** Return handle to molecule on success. ***/

   return dep;

   /*** eofexit (i.e. unexpected EOF): set eof and do error exit.  ***/

eofexit:
   OOPS("unexpected EOF");

   /*** errorexit: deallocate memory as needed & return. ***/

errorexit:
   dt_dealloc(mol);
   dt_dealloc(rads);
   du_free(alist);
   du_free(blist);
   du_free(xlist);
   du_free(ylist);
   du_free(vlist);
   return NULL_OB;
}
