/***************************************************************************
* du_acd2tdt_utils.c
* 
* (This stuff should probably all be in libdu.a.)
***************************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <memory.h>
#include <sys/types.h>
#include <sys/time.h>
#include <dt_smiles.h>
#include <du_utils.h>
#include "acd2tdt.h"


#define ISWHITE(c) \
        ((c) == ' ' || (c) == '\t' || (c) == '\n' || (c) == '\r')


/***************************************************************************
* FUNCTION: du_cas_standardize
*
* DESCRIPTION:
*       Converts a CAS number (in string form) to a standardized CAS
*       number.  The value returned is in a static buffer; the value
*       passed in is not altered.
***************************************************************************/
 
#define CSERROR(e) \
  {sprintf(errmsg, "%s: %s", e, orig_cas); \
   dt_errorsave("du_cas_standardize", DX_ERR_ERROR, errmsg); \
   return 0;}
 
int du_cas_standardize(int *plen, char *cas_str, char **pnew)
{
  static char cas[15];
  char  checkchar;
  char *ps, *pd, *end;
  char errmsg[200], *orig_cas;
  int casnum, checkdigit;
 
  /**** copy the string, skip '-' ****/
  cas_str = du_skipwhite(cas_str);      /* skip leading whitespace */
  orig_cas = cas_str;                   /* save for error reporting */
  while ('0' == *cas_str)
    cas_str++;  /* skip leading zeros */
  pd = cas;
  ps = cas_str;
  end = cas_str + *plen;
  while (*ps && (isdigit(*ps) || '-' == *ps) && ps < end) {
    if ('-' == *ps)
      ps++;                     /* skip '-' */
    else
      *(pd++) = *(ps++);
  }
  checkchar = *(--pd);          /* remember check digit, remove from string */
  *pd = '\0';
 
  /**** compute a new check digit ****/
  if (1 != sscanf(cas, "%d", &casnum))
    CSERROR("bad CAS number, can't normalize");
  checkdigit = du_cas_checkdigit(casnum);
  if (checkchar != (checkdigit + '0'))
    CSERROR("bad checksum on CAS number");
  /**** put a dash in before the last two digits ****/
  *pd = *(pd-1);
  *(pd-1) = *(pd-2);
  *(pd-2) = '-';
 
  /**** add another dash and the check digit ****/
  *(pd+1) = '-';
  *(pd+2) = checkchar;
  *(pd+3) = '\0';;
 
  *pnew = cas;
  *plen = strlen(cas);
  return TRUE;
}


/***************************************************************************
* FUNCTION: du_cas_checkdigit
*
* DESCRIPTION:
*       Generates a check digit for a CAS number.
*
***************************************************************************/
int du_cas_checkdigit(int cas)
{
  int n = 1, sum = 0;
 
  while (cas > 0) {
    sum += (cas % 10) * n;
    n++;
    cas /= 10;
  }
  return sum % 10;
}

char *du_white0(char *s)
{
  char *psrc, *pdest;
 
  if (NULL == s)
    return s;
 
  for (psrc = s, pdest = s; *psrc; psrc++)
    if (! ISWHITE(*psrc))
      *(pdest++) = *psrc;
  *pdest = '\0';
  return s;
}

char *du_upcase(char *s)
{
  char *b = s;
 
  while (*s) {
    if (islower(*s)) *s = toupper(*s);
    s++;
  }
  return b;
}

char *du_skipwhite(char *p)
{
  while (*p == ' ' || *p == '\t' || *p == '\n' || *p == '\r')  p++;
  return p;
}

/*============================================================================
 *  du_millisec(start) -- return elapsed time in milliseconds or -1 on error
 *
 *  On success, range of output value is 0 to 86399999.
 *  If called with zero, returns milliseconds since local midnight.
 */
 
#define DU_MSPERDAY  86400000

int du_millisec(int start)
{
  struct timeval  now;
  struct timezone zone;
  int        ms;
 
  if (0 != gettimeofday(&now, &zone))
    return -1;
  ms = now.tv_sec * 1000 + now.tv_usec / 1000 - start;
  return (0 <= ms ? ms : ms + DU_MSPERDAY);
}

/***************************************************************************
* FUNCTION: du_mergesort_ptr
*
* DESCRIPTION:
*       Identical to du_mergesort(), above, but sorts arrays of pointers,
*       which may not necessarily be 32-bit quantities.
***************************************************************************/
int du_mergesort_ptr(int nitems, void **array, void **a2,
                          int (*compare)(void *, void *))
{
  void **a1, **atmp;
  int n, runsize;
  int own_a2 = FALSE;
 
  /**** If no second array passed in, make one ****/
  a1 = array;
  if (NULL == a2) {
    if (NULL == (a2 = (void **) malloc(nitems * sizeof(void *))))
      return FALSE;
    own_a2 = TRUE;
  }
 
  /**** Sort the array ****/
  for (runsize = 1; runsize < nitems; runsize *= 2) {
    merge_ptr(a1, a2, runsize, nitems, compare);
    atmp = a1;
    a1 = a2;
    a2 = atmp;
  }
 
  /**** a1 points to sorted array.  If not in orig location, copy back ****/
  if (array != a1) {
    for (n = 0; n < nitems; n++)
      array[n] = a1[n];
  }
 
  /**** If we malloced an array, free it.  But a1 & a2 may have switched ****/
  if (own_a2) {
    if (array == a1)
      free(a2);
    else
      free(a1);
  }
  return TRUE;
}

/***************************************************************************
* FUNCTION: merge_ptr
*
* DESCRIPTION:
*       Part of du_mergesort_ptr(), below.
*       Identical to merge(), above, but works on pointers rather than
*       on 32-bit integers.
***************************************************************************/
 
void merge_ptr(void **src, void **dest, int runsize,
                      int len, int (*cmp)(void *, void *))
{
  void **p1, **p2, **pd, **p1end, **p2end, **srcend;
  int nruns, runpairs, i;
  long n;
 
  srcend = src + len;
 
  nruns = len / runsize;                /* number of runs of sorted numbers */
  if (len > nruns * runsize)
    nruns += 1;                         /* account for fractional runs */
  runpairs = nruns/2;                   /* number of pairs of runs */
  if (nruns > runpairs * 2)
 
    runpairs += 1;                      /* number of pairs of runs + "tail" */
 
  for (i = 0; i < runpairs; i++) {
    p1 = src  + i * runsize * 2;        /* p1 & p2 point to adjacent runs */
    p2 = p1 + runsize;
    if (p2 > srcend) p2 = srcend;
    p1end = p2;                         /* p1end and p2end are run's ends */
    p2end = p2 + runsize;
    if (p2end > srcend) p2end = srcend;
    pd = dest + i * runsize * 2;        /* pd points to destination run */
 
    n = (p1end - p1) + (p2end - p2);    /* length of combined run */
    while (n-- > 0) {
      if (p1 >= p1end)
        *(pd++) = *(p2++);              /* 1st run exhausted, use 2nd */
     else if (p2 >= p2end)
        *(pd++) = *(p1++);              /* 2nd run exhausted, use 1st */
     else {
        if ((*cmp)(*p1, *p2) <= 0)      /* compare values */
          *(pd++) = *(p1++);            /* 1st is less */
        else
          *(pd++) = *(p2++);            /* 2nd is less */
      }
    }   /* while */
  }     /* for */
}

