/*****************************************************************************
* FILE: masses.c
*
* SYNOPSIS:
*
*   dt_Real atomic_mass(dt_Handle atom);
*
*	Returns the atomic mass of the given atom, if known.  The mass is
*	determined as follows:
*
*	  - If the atom object's atomic number has been explicitely set
*	    (e.g. [2H], [12C], [15N] in a SMILES, or via dt_setweight()),
*	    and the exact isotopic mass is known for that weight, the exact
*	    isotopic mass is returned.
*
*	  - If the atom object's atomic number is not explicitely set (e.g.
*	    "C", "[CH4]", or dt_setweight(atom, 0)), then the average
*	    weight of naturally-occuring isotopes is returned.
*
*   dt_Integer normal_mass(dt_Handle atom)
*
*	Returns the mass (as an integer) of the most common isotope, or if
*	the element is not stable, returns the atomic mass of the most
*	stable isotope.
*
*   dt_Real amw(dt_Handle mol)
*       
*       Return average molecular weight for given mol.
*
*****************************************************************************/

#include <dt_smiles.h>
#include "acd2tdt.h"

#define MX_ELEM 105

#define AVE_HYDROGEN_MASS 1.007947


/*** aaw -- average atomic weight data ***/

static float aaw[MX_ELEM + 1] = {
0.000, 1.008, 4.0026, 6.94, 9.01218, 10.81, 12.011,
14.0067, 15.9994, 18.99846, 20.17, 22.98977, 24.305, 26.9815,
28.086, 30.9738, 32.06, 35.453, 39.948, 39.1, 40.08, 44.9559,
47.9, 50.941, 51.996, 54.938, 55.847, 58.9332, 58.71, 63.543,
65.38, 69.72, 72.59, 74.9216, 78.96, 79.904, 83.8, 85.467,
87.62, 88.9059, 91.22, 92.9064, 95.94, 98.9062, 101.07,
102.9055, 106.4, 107.868, 112.4, 114.82, 118.69, 121.75, 127.6,
126.9045, 131.3, 132.9055, 137.34, 138.9055, 140.12, 140.9077,
144.24, 145., 150.4, 151.96, 157.25, 158.9254, 162.5, 164.9303,
167.26, 168.9342, 173.04, 174.97, 178.49, 180.947, 183.85,
186.2, 190.2, 192.22, 195.09, 196.9665, 200.59, 204.37, 207.2,
208.9806, 209., 210., 222., 223., 226., 227., 232.0381,
231.0359, 238.029, 237.0408, 244., 243., 247., 247., 251.,
254.09, 257.08, 258.098, 259.10, 260.1, 261.1, 262.11
};

/*============================================================================
*   atomic_mass() -- return atomic mass for given atom
*/

dt_Real atomic_mass(dt_Handle atom)
{
  int atno, atwt;

  /*** Reject out-of-bounds requests. ***/

  atno = dt_number(atom);
  if (0 > atno || MX_ELEM < atno) return 0;

  /*** Return average mass if no isotope specified. ***/

  atwt = dt_weight(atom);
  if (0 == atwt) return aaw[atno];

  /*** Return isotopic mass for common/high-defect isotopes. ***/

  switch (atno) {

  case DX_ATN_H:
    switch (atwt) {
    case  1:  return (float) 1.007825;
    case  2:  return (float) 2.0140;
    case  3:  return (float) 3.01605;
    }
    break;

  case DX_ATN_C:
    switch (atwt) {
    case 12:  return (float) 12.000000;
    case 13:  return (float) 13.003355;
    case 14:  return (float) 14.003241;
    case 15:  return (float) 15.010599;
    case 16:  return (float) 16.014701;
    }
    break;

  case DX_ATN_N:
    switch (atwt) {
    case 14:  return (float) 14.003074;
    case 15:  return (float) 15.000108;
    }
    break;

  case DX_ATN_O:
    switch (atwt) {
    case 16:  return (float) 15.994915;
    case 17:  return (float) 16.999131;
    case 18:  return (float) 17.999160;
    }
    break;

  case DX_ATN_F:
    switch (atwt) {
    case 19:  return (float) 18.998403;
    }
    break;

  case DX_ATN_Si:
    switch (atwt) {
    case 28:  return (float) 27.976927;
    case 29:  return (float) 28.976495;
    case 30:  return (float) 29.973770;
    }
    break;

  case DX_ATN_P:
    switch (atwt) {
    case 31:  return (float) 30.973762;
    }
    break;

  case DX_ATN_S:
    switch (atwt) {
    case 32:  return (float) 31.972070;
    case 33:  return (float) 32.971456;
    case 34:  return (float) 33.967866;
    }
    break;

  case DX_ATN_Cl:
    switch (atwt) {
    case 35:  return (float) 34.968852;
    case 37:  return (float) 36.965903;
    }
    break;

  case DX_ATN_Br:
    switch (atwt) {
    case 79:  return (float) 78.918336;
    case 81:  return (float) 80.916289;
    }

  case DX_ATN_I:
    switch (atwt) {
    case 127:  return (float) 126.904473;
    }
    break;

  }

  /*** Otherwise, float integer atomic mass and return it ***/

  return (float) atwt;
}

/*
 *  nomass -- normal atomic mass data
 *
 *  If naturally occurring, for the most common isotope
 *  If not naturally occurring, for the most stable known isotope
 *  Reference: CRC Handbook of Chemistry and Physics, 60th Ed.
 */

static int nomass[MX_ELEM + 1] = { 0,
    1,   4,   7,   9,  11,  12,  14,  16,  19,  20,
   23,  24,  27,  28,  31,  32,  35,  40,  39,  40,
   45,  48,  51,  52,  55,  56,  59,  58,  63,  64,
   69,  74,  75,  80,  79,  84,  85,  88,  89,  90,
   93,  98,  97, 102, 103, 106, 107, 114, 115, 120,
  121, 130, 127, 132, 133, 138, 139, 140, 141, 142,
  145, 152, 153, 158, 159, 164, 165, 166, 169, 174,
  175, 180, 181, 184, 187, 192, 193, 195, 197, 202,
  205, 208, 209, 209, 210, 222, 223, 226, 227, 232,
  231, 238, 237, 244, 243, 247, 247, 251, 254, 257,
  257, 255, 256, 260, 262
};


/*============================================================================
 *  normal_mass() -- return integral normal atomic mass for given atom
 *
 *  If naturally occurring, for the most common isotope
 *  If not naturally occurring, for the most stable known isotope
 *  If out-of range 1-105, returns 0.
 *
 *  Reference: CRC Handbook of Chemistry and Physics, 60th Ed.
 */

dt_Integer normal_mass(dt_Handle atom)
{
   int atno = dt_number(atom);

   /*** Reject out-of-bounds requests. ***/

   if (0 > atno || MX_ELEM < atno) return 0;

   /*** Return table value. ***/

   return nomass[atno];
}

/*============================================================================
 *  amw() -- return average molecular weight for given mol.
 */

dt_Real amw(dt_Handle mol)
{
   double    amw;
   dt_Handle atom, atoms;

   amw   = 0.0;
   atoms = dt_stream(mol, TYP_ATOM);
   while (NULL_OB != (atom = dt_next(atoms))) {
      amw += atomic_mass(atom);
      amw += dt_imp_hcount(atom) * AVE_HYDROGEN_MASS;
   }
   dt_dealloc(atoms);

   return (float) amw;
}


/*============================================================================
 *  rdamw() -- return average molecular weight by RD rules
 *
 *  This function accounts for atomic mass and implicit hydrogens.
 */

/*** rdaaw -- average atomic weight data -- RD rules ***/
 
static double rdaaw[MX_ELEM + 1] = {
    0.00000,   1.00797,   4.00260,   6.93900,   9.01220,  10.81100,  12.01115,
   14.00670,  15.99940,  18.99840,  20.18300,  22.98980,  24.31200,  26.98150,
   28.08600,  30.97380,  32.06400,  35.45300,  39.94800,  39.10200,  40.08000,
   44.95600,  47.90000,  50.94200,  51.99600,  54.93800,  55.84700,  58.93320,
   58.71000,  63.54600,  65.37000,  69.72000,  72.59000,  74.92160,  78.96000,
   79.90400,  83.80000,  85.47000,  87.62000,  88.90500,  91.22000,  92.90600,
   95.94000,  98.90620, 101.07000, 102.90500, 106.40000, 107.86800, 112.40000,
  114.82000, 118.69000, 121.75000, 127.60000, 126.90440, 131.30000, 132.90500,
  137.33000, 138.91000, 140.12000, 140.90700, 144.24000, 145.00000, 150.35000,
  151.96000, 157.25000, 158.92400, 162.50000, 164.93000, 167.26000, 168.93400,
  173.04000, 174.97000, 178.49000, 180.94800, 183.85000, 186.20000, 190.20000,
  192.20000, 195.09000, 196.96700, 200.59000, 204.37000, 207.19000, 208.98000,
  209.00000, 210.00000, 222.00000, 223.00000, 226.03000, 227.00000, 232.03800,
  231.04000, 238.03000, 237.05000, 244.00000, 243.00000, 247.00000, 247.00000,
  251.00000, 254.00000, 257.00000, 258.00000, 259.00000, 260.00000, 261.00000,
  262.00000
};

dt_Real rd_amw(dt_Handle mol)
{
   int         atno, imph, wt;
   double      amw = 0.0;
   dt_Handle   atom, atoms = dt_stream(mol, TYP_ATOM);
 
   /*** Loop over atoms, adding up molecular weight. ***/
 
   while (NULL != (atom = dt_next(atoms))) {
     atno = dt_number(atom);

     /*** Add atomic weight of non-explicit hydrogens. ***/
 
     if (0 < (imph = dt_imp_hcount(atom))) amw += imph * rdaaw[1];
 
     /*** Add atomic weight of heavy atoms. ***/
 
     if (0 < atno && MX_ELEM >= atno) {
	amw += rdaaw[atno];

        /*** Account for isotopes (heavy approx, [2H] & [3H] exact). ***/

        if (0 != (wt = dt_weight(atom))) {
           if      (1 != atno) amw += wt - rdaaw[atno];
	   else if (2 == wt  ) amw += 1.00603;
	   else if (3 == wt  ) amw += 2.00808;
        }
     }
   }
   dt_dealloc(atoms);
 
   /*** Return average molecular weight. ***/
 
   return (dt_Real) amw;
}
