/***********************************************************************
*** mol2smi.c -- convert MOL-file to SMILES or TDTs
***
*** This program converts MDL format molfiles to Daylight SMILES and
*** TDTs.  SD (structure-data) files are converted to TDTS with SMILES
*** and other data.
***
*** Stereochemistry is inferred from the molfile and output as an 
*** Isomeric SMILES.  Atom "parity", "up/down bonds", 2D and 3D
*** coordinates are all taken into account in accordance with the
*** program options specified.
***
*** Authors: Jeremy Yang, Jack Delany
*** Rev:   09 Sep 2003
*** Please report bugs or requests to support@daylight.com.
************************************************************************
*** Refs: (1) JCICS #32, 1992, Dalby et al.
***       (2) "CTfile Formats", December 1999, MDLIS Inc.
************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include "dt_smiles.h"
#include "du_utils.h"
#include "du_3d.h"

#define VERSION "09 Sep 2003"

#define D3D_SOURCE "n/a"

#define MAX(x,y)  (((x) > (y)) ?  (x) : (y))
#define MIN(x,y)  (((x) < (y)) ?  (x) : (y))
#define XOR(a,b)  ((a) && !(b) || !(a) && (b))

#define MAXLINE  220   /*** Max length of molfile line (?) ***/

#define TAG_CHAR(c)   \
                    (  c == '$'                  \
                    || c == '_'                  \
                    || c == '/'                  \
                    || (c >= 'A' && c <= 'Z')    \
                    || (c >= 'a' && c <= 'z')    \
                    || (c >= '0' && c <= '9') )

void help()
{
  fprintf(stderr,
  "\t************************************************************\n"
  "\t| mol2smi - Convert MOL file to SMILES or TDT file         |\n"
  "\t************************************************************\n"
  "\t|                                                          |\n"
  "\t| mol2smi [options] <infile.mol >outfile.smi|tdt           |\n"
  "\t|                                                          |\n"
  "\t| options:                                                 |\n"
  "\t|   [-output_format SMI]  write SMILES (default)           |\n"
  "\t|   -output_format ISM    write Isomeric SMILES            |\n"
  "\t|   -output_format TDT    write TDTs (w/ ISM)              |\n"
  "\t|   -write_2d TRUE|FALSE  include xy's in TDTs             |\n"
  "\t|                         TRUE is default                  |\n"
  "\t|   -write_3d TRUE|FALSE  include xyz's in TDTs            |\n"
  "\t|                         TRUE is default                  |\n"
  "\t|   -parity_over_bonds    parity supercedes bondstyles     |\n"
  "\t|                         for chirality spec               |\n"
  "\t|   -smi_is_ism           TDT SMI<> contains ISM           |\n"
  "\t|   -ring_cistrans        write ring cis/trans bonds       |\n"
  "\t|   -m__iso_are_weights   M  ISO are weights not defects   |\n"
  "\t|   -one_ditem_per_line   each data line is a dataitem     |\n"
  "\t|   -id smiid             $SMIG run id field               |\n"
  "\t|   -comment comment      $SMIG comment                    |\n"
  "\t|   -minor_report n       progress report counter          |\n"
  "\t|   -major_report N       progress report counter          |\n"
  "\t|   -v                    report warnings                  |\n"
  "\t|   -vv                   report notes and warnings        |\n"
  "\t|   -help                 this help message                |\n"
  "\t************************************************************\n"
  "\t|                   Daylight CIS, Inc.                     |\n"
  "\t|               Toolkit/Contributed Program                |\n"
  "\t************************************************************\n");
  exit(1);
}

/********************************************************************
***  Function Declarations:
***     nooblanks -- strip outer blanks from char array
***     get_sdata -- read and write SD text data
***   printquoted -- fix/quote data with special chars
********************************************************************/
char *nooblanks(char *string);
int get_sdata(char *line,char *name,char *errtxt,int one_ditem_per_line);
int printquoted(char *text);

/********************************************************************
*** sameword - compares two strings ignoring case.
********************************************************************/
int sameword(char *p, char *q)
{
  while (*p||*q) {if (((*p++)-(*q++))%('a'-'A')) return 0;}
  return 1;
}


/************************************************************
*** Global variables.
************************************************************/
int ct_in=0;                /*** # connection tables read       ***/
char *progname=NULL;        /*** program name                   ***/

main(int argc,char *argv[])
{
  dt_Handle mol, atomstream, atom;

  char *line, *p;
  char *cansmi = NULL;        /*** Canonical (unique) SMILES      ***/
  char *canismi = NULL;       /*** Isomeric SMILES                ***/
  char errtxt[500] = "";      /*** Error msg buffer               ***/
  char name[100] = "";        /*** From 1st line                  ***/
  char *smiid = NULL;         /*** $SMIG run id                   ***/
  char *comment = NULL;       /*** $SMIG comment                  ***/

  int i, j, k, len;
  int rc, got2d, got3d, gotsymbols;

  int smi_out = 0;            /*** N SMILES written               ***/
  int tdts_out = 0;           /*** N TDTs written                 ***/
  int max_lensmi = 0;         /*** Max length of SMILES           ***/
  int max_lenismi = 0;        /*** Max length of iso SMILES       ***/
  int error_count = 0;        /*** Total count of errors          ***/
  int major_report = 1000;    /*** Progress report N              ***/
  int minor_report = 25;      /*** Progress report n              ***/

  int nonempty_tdt = 0;       /*** Boolean: Anything in TDT?      ***/
  int nonstruct = 0;          /*** Boolean: Non-structure?        ***/
  int write_ism = 0;          /*** Boolean: Iso-SMILES output?    ***/
  int write_smi = 0;          /*** Boolean: SMILES output?        ***/
  int write_tdt = 0;          /*** Boolean: TDT output?           ***/
  int write_2d = 0;           /*** Boolean: 2D included?          ***/
  int write_3d = 0;           /*** Boolean: 3D included?          ***/
  int smi_is_ism = 0;         /*** Boolean: SMI<> is Iso-SMILES   ***/
  int one_ditem_per_line = 0; /*** Boolean: one dataitem per line ***/
  int errlev=DX_ERR_ERROR;    /*** Error level (min) reported     ***/

  /*****************************************
  ***  Parse command line.
  *****************************************/
  write_smi = write_2d = write_3d = 1;
  progname = *argv;
  for (++argv; --argc; ++argv) {
    if (sameword(*argv, "-help")) help();
    if (sameword(*argv, "-output_format")) {
      if (--argc == 0) {
        fprintf(stderr, "ERROR (MOL2SMI): \"%s\" requires value...\n", *argv);
        help();
      } else if (sameword(*++argv, "SMI")) {
        write_smi = 1; write_ism = 0; write_tdt = 0;
      } else if (sameword(*argv, "ISM")) {
        write_smi = 0; write_ism = 1; write_tdt = 0;
      } else if (sameword(*argv, "TDT")) {
        write_smi = 0; write_ism = 0; write_tdt = 1;
      } else {
        fprintf(stderr, "ERROR (MOL2SMI): Unknown format: \"%s\"\n", *argv);
        help();
      }
    } else if (sameword(*argv, "-s")) {
        write_smi = 1; write_ism = 0; write_tdt = 0;
    } else if (sameword(*argv, "-i")) {
        write_smi = 0; write_ism = 1; write_tdt = 0;
    } else if (sameword(*argv, "-t")) {
        write_smi = 0; write_ism = 0; write_tdt = 1;
    } else if (sameword(*argv, "-write_2d")) {
      if (--argc == 0) {
        fprintf(stderr, "ERROR (MOL2SMI): \"%s\" requires value...\n", *argv);
        help();
      } else if (sameword(*++argv, "TRUE")) { write_2d = 1;
      } else if (sameword(*argv, "FALSE"))  { write_2d = 0;
      } else {
        fprintf(stderr, "ERROR (MOL2SMI): Bad option value: \"%s\"\n", *argv);
        help();
      }
    } else if (sameword(*argv, "-write_3d")) {
      if (--argc == 0) {
        fprintf(stderr, "ERROR (MOL2SMI): \"%s\" requires value...\n", *argv);
        help();
      } else if (sameword(*++argv, "TRUE")) { write_3d = 1;
      } else if (sameword(*argv, "FALSE"))  { write_3d = 0;
      } else {
        fprintf(stderr, "ERROR (MOL2SMI): Bad option value: \"%s\"\n", *argv);
        help();
      }
    } else if (sameword(*argv, "-parity_over_bonds")) { 
      du_mol2smi_setparity(TRUE);
    } else if (sameword(*argv, "-smi_is_ism")) { smi_is_ism = 1;
    } else if (sameword(*argv, "-one_ditem_per_line")) { one_ditem_per_line = 1;
    } else if (sameword(*argv, "-ring_cistrans")) { 
      du_mol2smi_setringEZ(TRUE);
    } else if (sameword(*argv, "-m__iso_are_weights")) {
      du_mol2smi_setisoweight(TRUE);
    } else if (sameword(*argv, "-id")) {
      if (--argc == 0) {
        fprintf(stderr, "ERROR (MOL2SMI): \"%s\" requires value...\n", *argv);
        help();
      } else {
        smiid = *++argv;
      }
    } else if (sameword(*argv, "-comment")) {
      if (--argc == 0) {
        fprintf(stderr, "ERROR (MOL2SMI): \"%s\" requires value...\n", *argv);
        help();
      } else {
        comment = *++argv;
      }
    } else if (sameword(*argv, "-minor_report")) {
      if (--argc == 0) {
        fprintf(stderr, "ERROR (MOL2SMI): \"%s\" requires value...\n", *argv);
        help();
      } else {
        minor_report = atoi(*++argv);
      }
    } else if (sameword(*argv, "-major_report")) {
      if (--argc == 0) {
        fprintf(stderr, "ERROR (MOL2SMI): \"%s\" requires value...\n", *argv);
        help();
      } else {
        major_report = atoi(*++argv);
      }
    } else if (sameword(*argv, "-v")) {
      errlev=DX_ERR_WARN;
    } else if (sameword(*argv, "-vv")) {
      errlev=DX_ERR_NOTE;
    } else {
      fprintf(stderr, "ERROR (MOL2SMI): Bad option \"%s\"\n", *argv);
      help();
    }
  }

  if (write_tdt)
    printf("$SMIG<%s;%s;%s;%s>\n|\n",
      smiid ? smiid : "", progname, VERSION, comment ? comment : "");

  /********************************************************************
  ***  Molecule loop:
  ********************************************************************/
  while (1) {
    nonstruct = nonempty_tdt = 0;
    strcpy(name, "");

    /********************************************************************
    ***  Header block:
    ***  Skip RDfile headers.
    ***  Read name from first line; strip trailing blanks.
    ********************************************************************/
    do {
      if (!(line = du_fgetline(&len, stdin))) {
        fprintf(stderr,"NOTE (mol2smi): EOF... (NORMAL EXIT)\n");
        goto byebye;
      }
    } while ('$' == line[0]);

    ++ct_in;

    for (k = strlen(line); ' ' == line[k - 1]; --k)   ;
    for (j = 0; j <= k; ++j) name[j] = line[j];
    name[k] = '\0';

    /*** Header block: User, program ***/
    if (!(line = du_fgetline(&len, stdin))) {
      dt_errorsave("mol2smi",DX_ERR_ERROR,"Unexpected EOF...");
      goto byebye;
    }
    /*** Header block: Comment ***/
    if (!(line = du_fgetline(&len, stdin))) {
      dt_errorsave("mol2smi",DX_ERR_ERROR,"Unexpected EOF...");
      goto byebye;
    }

    mol = du_mol2smi(stdin, &rc);
    if (rc == -1)
      {
	error_count++;
	goto nextct;
      }
    if (NULL_OB==mol && rc==0) {
      sprintf(errtxt, "Non-struct; name=\"%s\",mol#=%d",name,ct_in);
      dt_errorsave("mol2smi",DX_ERR_NOTE,errtxt);
      if (!write_tdt) goto nextct;
      if (strlen(name)) {
        printf("$NAM<%s>\n", name);
        nonempty_tdt = 1;
      }
      nonstruct = 1;
      if (write_tdt) goto endstruct;
      else goto nextct;
    }
    
    /****************************************************
    ***  Make SMILES
    ****************************************************/

    p = dt_cansmiles(&len, mol, 0);
    if (len <= 0) {
      sprintf(errtxt,"dt_cansmiles failure; name=\"%s\",mol#=%d",name,ct_in);
      dt_errorsave("mol2smi",DX_ERR_ERROR,errtxt);
      ++error_count;
      goto nextct;
    }
    if (len > max_lensmi) {
      if (!cansmi) cansmi = (char *) malloc((len+1)*sizeof(char));
      else cansmi = (char *) realloc(cansmi, (len+1)*sizeof(char));
      max_lensmi = len;
    }
    strncpy(cansmi, p, len); cansmi[len] = '\0';
    if (write_ism || write_tdt) {
      p = dt_cansmiles(&len, mol, 1);
      if (len <= 0) {
        sprintf(errtxt,"dt_cansmiles failure; name=\"%s\",mol#=%d",name,ct_in);
        dt_errorsave("mol2smi",DX_ERR_ERROR,errtxt);
        ++error_count;
        goto nextct;
      }
      if (len > max_lenismi) {
        if (!canismi) canismi = (char *) malloc((len+1)*sizeof(char));
        else canismi = (char *) realloc(canismi, (len+1)*sizeof(char));
        max_lenismi = len;
      }
      strncpy(canismi, p, len); canismi[len] = '\0';
    }
    if (write_tdt) {
      printf("$SMI<%s>\n", smi_is_ism ? canismi : cansmi);
      if (strlen(name)) printf("$NAM<%s>\n", name);
      if (strcmp(cansmi, canismi)) printf("ISM<%s>\n", canismi);
      nonempty_tdt = 1;
    }
    else if (write_ism) {
      printf("%s %s\n", canismi, name);
    }
    else {
      printf("%s %s\n", cansmi, name);
    }  
    ++smi_out;

    /********************************************************************
    *** FOR TDT ONLY: Write D3D data if wegot_3d; write 2D
    *** data if wegot_2d.  If there are MDL "special" atom symbols
    *** (e.g., R, X, Pg, Rsn, etc.), write these as ALAB datatype.
    *********************************************************************
    *** Note that the SMILES_NTUPLE normalization allows for two 
    *** cases: (1) data for all non-hydrogen atoms only, in SMILES order,
    *** or (2) data for all non-hydrogen atoms, in SMILES order, followed
    *** by data for all hydrogens, in SMILES order.  Implicit and 
    *** explicit hydrogens are treated the same.
    ********************************************************************/
    if (write_tdt) {
      if (dt_boolean(mol, 9, "gotsymbol"))
	{
	  char *str_tmp;
	  int slen;
	  
	  printf("ALAB<");
	  atomstream = dt_canstream(mol, TYP_ATOM, 0, 0);
	  for (i = 0; NULL_OB != (atom = dt_next(atomstream)); ++i)
	    {
	      if (i) printf(",");
	      str_tmp = dt_string(&slen, atom, 6, "symbol");
	      if ((str_tmp != NULL) && (slen > 0))
		printf("%.*s", slen, str_tmp);
	    }
	  printf(">\n");
	  dt_dealloc(atomstream);
	}

      /*** Only write one or the other.  If we have 3D, don't
        output the 2D.  JB 20030821 ***/

      if (write_3d && dt_boolean(mol, 5, "got3d"))
	printf("$D3D<0.0;%s;%s;%s;>\n", name, du_mol2d3d(mol), D3D_SOURCE);
      else if (write_2d && dt_boolean(mol, 5, "got2d"))
	printf("2D<%s;>\n", du_mol2d2d(mol));

endstruct:   /*** End of structural data. ***/

      /*****************************************************
      *** Call get_sdata() to read and write text data.
      *** get_sdata returns the number of bytes written.
      *****************************************************/
      nonempty_tdt += get_sdata(line,name,errtxt,one_ditem_per_line);

      /*** Done reading text data; "$$$$" or EOF found... ***/
      if (nonempty_tdt) {
	  printf("|\n");
	  ++tdts_out;
      } else {
        sprintf(errtxt,"No TDT (no data); name=\"%s\",mol#=%d",name,ct_in);
        dt_errorsave("mol2smi",DX_ERR_NOTE,errtxt);
      }
    }
    if (!nonstruct && NULL_OB!=mol) dt_dealloc(mol);

    /********************************************************************
    *** Drop to here if problem with CT.  Mol should have already been
    *** deallocated.  Report errors.
    ********************************************************************/
nextct:
    du_printerrors(stderr,errlev);
    dt_errorclear();

    /*** Find end of CT or EOF (?) ***/
    while (strncmp("$$$$", line, 4)) {
      if (!(line = du_fgetline(&len, stdin))) {
        if (ct_in==1) { /*** Must be ISIS-MOL (not SD) file...  ***/
          goto byebye;
        } else {
          ++error_count;
          dt_errorsave("mol2smi",DX_ERR_ERROR,"Unexpected EOF...");
          goto byebye;
        }
      }
    }

   /******************************************************
    *** Progress reports...
    ******************************************************/
    if (minor_report&&!((ct_in)%minor_report))
      { fprintf(stderr, "."); fflush(stderr); }
    if (major_report&&!((ct_in)%major_report))
      fprintf(stderr," %d MOLs, %d errors\n",ct_in,error_count);

    du_printerrors(stderr,errlev);
    dt_errorclear();
  } /*** End of molecule loop ***/

  /*****************************************************
  *** Byebye: drop to here on EOF or fatal error.
  *****************************************************/
byebye:
  du_printerrors(stderr,errlev);
  fprintf(stderr, "MOL2SMI results:  MOLs in:    %d\n", ct_in);
  if (write_tdt)
    fprintf(stderr, "MOL2SMI results:  TDTs out:   %d\n", tdts_out);
  fprintf(stderr, "MOL2SMI results:  SMILES out: %d\n", smi_out);
  fprintf(stderr, "MOL2SMI results:  Errors:     %d\n", error_count);
  fprintf(stderr, "Adios!\n");

  free(cansmi); free(canismi);
  exit (error_count);
}

/********************************************************************
***  nooblanks - Strips outer blanks from char array.
********************************************************************/
char *nooblanks(char *buff)
{
  int i;
  char *p;

  for (p = buff; *p == ' '; ++p) ;
  if (p != buff) strcpy(buff, p);
  for (i = strlen(buff); (buff[i-1] == ' ') && (i > 0); --i) ;
  buff[i] = '\0';
  return (buff);
}

/********************************************************************
*** get_sdata - Read "SD" data if present.  Tags are
*** taken from text after starting ">".  Data concatenated from
*** lines till next starting ">" and stored in linked list.
*** Note that linked list starts with and keeps one empty node
*** (*head).
********************************************************************/
int get_sdata(char *line, char *name, char *errtxt, 
              int one_ditem_per_line)
{
  char *p, *q;                /*** Generic (char *)'s             ***/
  char tag[21] = "";          /*** Datatype tag for addl SD data  ***/
  char oldtag[21] = "";       /*** Unfixed datatype tag           ***/
  int k;                      /*** Generic indices                ***/
  int len;                    /*** String length                  ***/
  int tag_ok;                 /*** Tag legal?                     ***/
  int bytes = 0;              /*** Bytes written                  ***/

  typedef struct data {       /*** Struct for linked-list of addl ***/
    char *text;               /*** data (SD format)               ***/
    struct data *next;
  } Data;

  Data *pdata;                /*** Ptr to current Data struct     ***/
  Data *ptrash;               /*** Ptr to trashable struct        ***/
  static Data *head = NULL;   /*** Ptr to empty head struct       ***/

  if (!head) {
    head = (Data *)malloc(sizeof(Data));
    head->text = NULL;
    head->next = NULL;
  }

  while (strncmp("$$$$", line, 4)) {
    tag_ok = 1;
    if (('>' == line[0]) &&
        (p = strchr(line + 1, '<')) &&
        (q = strchr(line + 1, '>')) &&
        (q - p > 1)) {

      /*****************************************************
      *** Here we get tag, read all records till next
      *** "> <foo>" or "$$$$" for all data items.
      *** Outer "<>"'s are stripped from tags.
      *** The format says a blank line delimits each
      *** datatype-section, e.g.,
      ***
      *** > <Address>
      *** 111 rue Iberville
      *** Suite 610
      *** New Orleans, LA
      ***
      *** > <Phone>
      *** 504-898-5500
      *****************************************************/
      k = MIN(q - p - 1, 20);
      strncpy(tag, p + 1, k);
      tag[k] = '\0';
      strcpy(oldtag, tag);

      /*****************************************************
      *** Check for illegal Thor datatag.
      *****************************************************/
      for (p = tag; *p != '\0'; ++p) {
        if (!TAG_CHAR(*p)) {
          tag_ok = 0;
          *p = '_';
        }
      }
      if (!tag_ok) {
        sprintf(errtxt,"Illegal datatag \"%s\" replaced by \"%s\".",oldtag,tag);
        dt_errorsave("mol2smi",DX_ERR_NOTE,errtxt);
      }

      /*****************************************************
      *** The default is that each MDL tag is a separate dataitem,
      *** but separate data lines are combined for that one
      *** dataitem.  However, to support some formats, e.g.,
      ***
      *** > <Boiling Point>
      *** 100.2
      *** 99.5
      ***
      *** the "one_line_per_dataitem" option means that
      *** each line is considered a separate dataitem,
      *** and gets its own tag.  Leading and trailing
      *** blanks are stripped (nooblanks()).  Empty or
      *** blank dataitems are not written.
      *****************************************************/
      if (!(line = du_fgetline(&len, stdin))) { return (bytes); }
      pdata = head;
      while (strcmp("", nooblanks(line))) {
        pdata->next = (Data *)malloc(sizeof(Data));
        pdata = pdata->next;
        pdata->next = NULL;
        pdata->text = (char *)malloc((strlen(line) + 1) * sizeof(char));
        strcpy(pdata->text, nooblanks(line));

        if (!(line = du_fgetline(&len, stdin))) { return (bytes); }
      }

      /*****************************************************
      *** Now write out data to TDT; free linked list.
      *****************************************************/
      if (one_ditem_per_line) {
        for(pdata=head; NULL!=pdata->next; pdata=pdata->next) {
          if (strlen(pdata->next->text)) {
            bytes += printf("%s<", tag);
            bytes += printquoted(pdata->next->text);
            bytes += printf(">\n");
          }
        }
      } else {
        bytes += printf("%s<", tag);
        for(pdata=head; NULL!=pdata->next; pdata=pdata->next) {
          if (pdata != head) bytes += printf(" "); 
          if (strlen(pdata->next->text)) {
            bytes += printquoted(pdata->next->text);
          }
        }
        bytes += printf(">\n");
      }
      while (head->next) {
        ptrash = head;
        head = head->next;
        if (ptrash->text) free(ptrash->text);
        free(ptrash);
      }
    }
    else {
      if (!(line=du_fgetline(&len,stdin))) { return (bytes); }
    }
  }
  return (bytes);
}

/********************************************************************
*** printquoted - print dataitem, doubling double-quotes,
*** quoting entire field only if necessary.
********************************************************************/
int printquoted(char *text)
{
  int bytes = 0;
  char *p;

  if (strpbrk(text,";><|\"$")) bytes += printf("\"");
  for (p=text; *p!='\0'; ++p) {
    if ('"'==*p) bytes += printf("\"\"");
    else bytes += printf("%c",*p);
  }
  if (strpbrk(text,";><|\"$")) bytes += printf("\"");
  return (bytes);
}
