/**********************************************************************
 * This source code is public domain, and may be freely distributed,   *
 * modified, and used for any purpose.  IT COMES WITH ABSOLUTELY NO    *
 * WARRANTY OF ANY KIND.                       *
 **********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "dt_smiles.h"

#include "du_utils.h"
#include "du_smiles.h"

#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif

#define MX_LINE            512
#define MX_VARIATION       100

dt_Handle du_ctab2dep(FILE *, int *);
dt_Handle du_mol2smi(FILE *, int *);

/*======================================================================
 * 
 ====================================================================== */

int dump_errors(int regno)
{
  dt_Handle sos, sob;
  dt_String errtxt;
  dt_Integer errlen;

  sos = dt_errors(DX_ERR_NOTE);
  while (NULL_OB != (sob = dt_next(sos)))
    {
      errtxt = dt_stringvalue(&errlen, sob);
      if (errlen > 1)
  fprintf(stderr, "RIREG %d: %.*s\n", regno, errlen, errtxt);
      dt_dealloc(sob);
    }
  dt_dealloc(sos);
  dt_errorclear();
  return (TRUE);
}

/*======================================================================
 * This is a safeguarded free that also clears the pointer that 
 * referenced the data 
 ====================================================================== */

void guardedFree(char **p)
{
   if (*p == NULL)
   {
      fprintf(stderr, "trying to free a NULL pointer\n");
      exit (1);
   }
   else
   {
      free(*p);
      *p = NULL;
   }
}

/*======================================================================
 * Main
 ====================================================================== */

int main(int argc, char **argv)
{
  dt_Handle rxn[MX_VARIATION];
  dt_String ereg[MX_VARIATION];
  dt_Integer var_count, variation, ireg, next_ireg;
  char line[MX_LINE];
  
  dt_Handle mol, rmol, pmol, dep;
  dt_Handle ratoms, ratom, patoms, patom;
  int lens, i, rn, pn, mapvalue, rc;
  char *s, *extreg, *preg;
  int record_is_ok;

  if ((argc == 2) && (argv[1] != NULL))
    extreg = du_strdup(argv[1]);
  else
    extreg = "EXTREG";
  
  /*** Scan until we find the start of a reaction. ***/

  while(NULL!=du_fgetz(line,MX_LINE,stdin))
    if(!strncmp(line, "$RFMT $RIREG", 12))
      {
	/*** Get the internal regno from the line. ***/
	sscanf(line + 13, "%d", &ireg);
	break;
      }
  
  /*** Main loop.  Once per rd-file ***/
  
  while(NULL!=du_fgetz(line,MX_LINE,stdin))
    if(!strncmp(line, "$RXN", 3))
      {
        /*** Initialize some variables. ***/
	record_is_ok = TRUE;
	rmol = pmol = NULL_OB;
	var_count = 0;
	for (i = 0; i < MX_VARIATION; i++)
	  rxn[i] = NULL_OB;
	
        /*** Advance the input file. ***/
	du_fgetz(line, MX_LINE, stdin);
	du_fgetz(line, MX_LINE, stdin);
	du_fgetz(line, MX_LINE, stdin);
	du_fgetz(line, MX_LINE, stdin);
	
        /*** Now we know how many reactants and products ***/
	
        sscanf(line, "%d %d", &rn, &pn);
	
        /*** Note: rxn[0] is used as a template.  Copies are made and
	  the agents are added to the copies for the 1 - n variations
	  in a record. ***/
	
        rxn[0] = dt_alloc_reaction();
        dt_mod_on(rxn[0]);
	
        /*** Get the reactants. ***/
	
        for (i = 0; i < rn; i++)
          {
	    du_fgetz(line, MX_LINE, stdin);
	    while ((0 != strncmp(line, "  REACCS", 8)) &&
		   (0 != strncmp(line, "  -ISIS-", 8)))
	      du_fgetz(line, MX_LINE, stdin);
	    du_fgetz(line, MX_LINE, stdin);

	    mol = du_mol2smi(stdin, &rc);
	    if (rc == -1)
	      {
		dt_errorsave("du_ctab2dep", DX_ERR_WARN, "Error for reactant");
		record_is_ok = FALSE;
	      }
	    else if (mol == NULL_OB)
	      dt_errorsave("du_ctab2dep", DX_ERR_WARN, "Empty reactant");
	    else
              {
		rmol = dt_addcomponent(rxn[0], mol, DX_ROLE_REACTANT);
		dt_dealloc(mol);
              }
	  }
	
        /*** Get the products. ***/
	
        for (i = 0; i < pn; i++)
          {
	    du_fgetz(line, MX_LINE, stdin);
	    while ((0 != strncmp(line, "  REACCS", 8)) &&
		   (0 != strncmp(line, "  -ISIS-", 8)))
	      du_fgetz(line, MX_LINE, stdin);
	    du_fgetz(line, MX_LINE, stdin);

	    mol = du_mol2smi(stdin, &rc);
	    if (rc == -1)
	      {
		dt_errorsave("du_ctab2dep", DX_ERR_WARN, "Error for product");
		record_is_ok = FALSE;
	      }
	    else if (mol == NULL_OB)
	      dt_errorsave("du_ctab2dep", DX_ERR_WARN, "Empty product");
	    else
              {
		pmol = dt_addcomponent(rxn[0], mol, DX_ROLE_PRODUCT);
		dt_dealloc(mol);
              }
	  }
	
        /*** Set the map classes based on the integer property
	  which was stored. ***/
        
        patoms = dt_stream(pmol, TYP_ATOM);
	ratoms = dt_stream(rmol, TYP_ATOM);
	while (NULL_OB != (ratom = dt_next(ratoms)))
	  {
	    if (0 != (mapvalue = dt_integer(ratom, 3, "map")))
	      {
		dt_reset(patoms);
		while (NULL_OB != (patom = dt_next(patoms)))
		  if (dt_integer(patom, 3, "map") == mapvalue)
		    dt_setmap(ratom, patom);
	      }
	  }
	dt_dealloc(ratoms);
	dt_dealloc(patoms);
	
	variation = 1; /* always the first variation after the rxn block */

	if (rxn[variation] == NULL_OB)
	  {
	    rxn[variation] = dt_copy(rxn[0]);
	    var_count = variation;
	  }
	
	/*** Now, consume the rest of the lines for the reaction.  Get
	  agents, if available.  Otherwise, throw away the stuff.  This
	  section must deal with multiple variations as separate
          subtrees.  ***/

	while(NULL != du_fgetz(line,MX_LINE,stdin))
	  {
	    /*** check if we've got a new variation and copy the reaction
	      prototype to this slot ***/
	    if (strstr(line, "VARIATION"))
	      {
		if (1 != sscanf(line + 20, "(%d)", &variation))
		  {
		    dt_errorsave("main", DX_ERR_WARN,
				 "Couldn't read variation for extreg");
		    record_is_ok = FALSE;
		  }
		
		if (rxn[variation] == NULL_OB)
		  {
		    rxn[variation] = dt_copy(rxn[0]);
		    var_count = (variation>var_count)?variation:var_count;
		    ereg[variation] = NULL;
		  }
	      }
	    
	    if (strstr(line, "REGNO") && 
		(strstr(line, "CATALYST") || 
		 strstr(line, "SOLVENT")))
	      {
		/*** We found an agent!!!  ***/
		/*** Advance the file. ***/

		du_fgetz(line, MX_LINE, stdin);
		while (0 != strncmp(line, "  REACCS", 8))
		  du_fgetz(line, MX_LINE, stdin);
		du_fgetz(line, MX_LINE, stdin);

		mol = du_mol2smi(stdin, &rc);
		if (rc == -1)
		  {
		    dt_errorsave("du_ctab2dep", DX_ERR_WARN,
				 "Error for agent");
		    record_is_ok = FALSE;
		  }
		else if (mol == NULL_OB)
		  dt_errorsave("du_ctab2dep", DX_ERR_WARN, "Empty agent");
		else
		  {
		    dt_addcomponent(rxn[variation], mol, DX_ROLE_AGENT);
		    dt_dealloc(mol);
		  }
	      }
	    
	    /*** External registration number. ***/
	    
	    if (strstr(line, extreg))
	      {
		/*** Figure out which variation it belongs to. ***/
		if (1 != sscanf(line + 20, "(%d)", &variation))
		  {
		    dt_errorsave("main", DX_ERR_WARN,
				 "Couldn't read variation for extreg");
		    record_is_ok = FALSE;
		  }
		
		if (rxn[variation] == NULL_OB)
		  {
		    rxn[variation] = dt_copy(rxn[0]);
		    var_count = variation;
		  }
		
                du_fgetz(line, MX_LINE, stdin);
		/*** These show up as $RNO<> ##### ***/
		ereg[variation] = du_strdup(line + 7);
	      }
	    
	    /*** Start of another reaction. ***/
	    
            if (!strncmp(line, "$RFMT $RIREG", 12))
	      {
		sscanf(line + 13, "%d", &next_ireg);
		break;
	      }
	  }
	
        /*** Read the entire record, now write the SMILES out ***/
	
	dump_errors(ireg);
	if (record_is_ok)
	  {
	    /*** Generate and output the datatree.  Start with the root. ***/
	    dt_mod_off(rxn[0]);
	    s = dt_cansmiles(&lens, rxn[0], 0);
	    if (0 < lens) 
	      printf("$SMI<\"%.*s\">\n", lens, s);
	    dt_dealloc(rxn[0]);
	    
	    /*** Now, write the subtrees. ***/
	    for (i = 1; i <= var_count; i++)
	      {
		dt_mod_off(rxn[i]);
		s = dt_cansmiles(&lens, rxn[i], 1);
		
		/*** ereg found.  Make it $RID<>ISM<>. ***/
		if (ereg[i] != NULL)
		  {
		    if (*ereg[i] == ' ')
                      {
			preg = ereg[i] + strspn(ereg[i], " ");
			printf("$RID<%s>\n", preg);
		      }
		    else
                      printf("$RID<%s>\n", ereg[i]);
		    guardedFree(&ereg[i]);
		  }
		else
		  printf("$RID_INTERNAL<%d_%d>\n", ireg, i);
		if (0 < lens)
		  printf("ISM<\"%.*s\">\n", lens, s);
		dt_dealloc(rxn[i]);
	      }
	    printf("|\n");
	  }
	fflush(stdout);
	fflush(stderr);
	ireg = next_ireg;
      }
  return(1);
}

