/*************************************************************************
*** smi2mol.c - convert SMI or TDT file to a MOLfile (MDL SD format), 
***             optionally adding Depict 2D coordinates.
*** 
*** $D3D or 2D data is written to coordinate fields; Depict 2Ds are 
*** added if needed and not de-requested.
*** 
*** Wildcard atoms (*) with ALAB<> labels are interpreted as MDL
*** pseudo-atoms and written to the MDL symbol field.
*** 
*** Other datatypes are appended in SD format.
*** 
*** R/S chirality is translated to stereo parity and up/down bonds.
*** Cis/trans information MAY NOT BE in generated coordinates due to
*** the Depict algorithm.
***
*** Author: Jeremy Yang
*** Rev:   12 Feb 2002
*************************************************************************
*** To compile and run this program without the Depict Toolkit, comment
*** out the _DO_DEPICT preprocessor definition below, and remove 
*** libdt_depict.a and libdl_stubs.a from the Makefile.
*************************************************************************
*** Thanks to George Lauri and John Bradshaw for their contributions.
*************************************************************************
*** Reference: JCICS #32, 1992, Dalby et al.
*************************************************************************
****-cut-here-**************** Makefile *******************-cut-here-****
*** CONTRIB = $(DY_ROOT)/contrib
*** LOCAL_INCLUDES = -I$(CONTRIB)/include -L$(CONTRIB)/lib
*** include $(CONTRIB)/src/makefile_include
***  
*** LIBS = -ldu -ldt_depict -ldl_stubs -ldt_smiles
***  
*** smi2mol: smi2mol.o
*** 	$(CC) -o $@ $(CFLAGS) smi2mol.o $(LIBS) $(SYSLIBS)
****-cut-here-****************-cut-here-*******************-cut-here-****
*************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include "dt_smiles.h"
#include "dt_depict.h"
#include "du_utils.h"
#include "du_3d.h"

#define MAX(x, y) (((x) > (y)) ?  (x) : (y))
#define MIN(x, y) (((x) < (y)) ?  (x) : (y))

#define MAXATOMS  1000
#define MAXBONDS  1000

/************************************************************
*** Comment out or delete this preprocessor define to enable
*** compilation without the Depict Toolkit.
************************************************************/
#define _DO_DEPICT TRUE

/************************************************************
*** Unifying struct for atom-tuple info
************************************************************/
typedef struct {
  dt_Handle h;
  float     x;
  float     y;
  float     z;
  char      symbol[4];
  int       parity;
  int       visited;
} ATOMINFO;
 
/************************************************************
*** Unifying struct for bond-tuple info
************************************************************/
typedef struct {
  dt_Handle h;
  dt_Handle from_atom;
  dt_Handle to_atom;
  int stereo;
} BONDINFO;

int setparity(dt_Handle mol);
int setchibonds(dt_Handle mol);
int setezstereo(dt_Handle mol);
dt_Handle stream2seq(dt_Handle stream);
int atnum2mass(int atnum);
int sameword(char *str1, char *str2);

void help()
{
  fprintf(stderr,
  "\t************************************************************\n"
  "\t| smi2mol - Convert TDT file to MOL file (MDL SD format)   |\n"
  "\t************************************************************\n"
  "\t|                                                          |\n"
  "\t| smi2mol [options] <infile.smi|tdt >outfile.mol           |\n"
  "\t|                                                          |\n"
  "\t| options:                                                 |\n"
  "\t|   -input_format SMI|TDT (-s and -t also work)            |\n"
  "\t|   -explicit_hydrogens ... Make H's explicit atoms        |\n");

#ifdef _DO_DEPICT
  fprintf(stderr,
  "\t|   -add_2d TRUE|FALSE ... Add Depict 2Ds if needed.       |\n");
#endif

  fprintf(stderr,
  "\t|   -name_datatype <tag>  TDT tag for molfile name.        |\n"
  "\t|   -write_hcount ....... Write hcount field.              |\n"
  "\t|   -comment_is_smi ..... Write smi in comment field       |\n"
  "\t|   -split_fields ... multiple fields -> separate data     |\n"
  "\t|   -minor_report n ..... progress report counter          |\n"
  "\t|   -major_report N ..... progress report counter          |\n"
  "\t|   -v .................. verbose: report warnings         |\n"
  "\t|   -vv ................. report notes and warnings        |\n"
  "\t|   -help ............... This help                        |\n"
  "\t|                                                          |\n"
  "\t************************************************************\n"
  "\t|                   Daylight CIS, Inc.                     |\n"
  "\t|             Toolkit Demonstration Program                |\n"
  "\t************************************************************\n");
  exit (1);
}

ATOMINFO *atoms = NULL;       /*** Atom info array (dynamic)           ***/
BONDINFO *bonds = NULL;       /*** Bond info array (dynamic)           ***/
int natoms;                   /*** Number of atoms to molfile          ***/
int nbonds;                   /*** Number of bonds to molfile          ***/
int max_natoms = 0;           /*** Max number of atoms so far          ***/
int max_nbonds = 0;           /*** Max number of bonds so far          ***/
int imp_htotal;               /*** Total implicit H count              ***/
char errtxt[500]="";          /*** Error msg buffer                    ***/

main(int argc, char *argv[])
{
  dt_Handle atom;             /*** Atom object                         ***/
  dt_Handle atom1;            /*** Atom object #2                      ***/
  dt_Handle atom2;            /*** Atom object #3                      ***/
  dt_Handle atomseq;          /*** Atom sequence                       ***/
  dt_Handle atomstream;       /*** Atom stream                         ***/
  dt_Handle xatoms;           /*** Atom stream #2                      ***/
  dt_Handle batoms;           /*** Bond Atom stream                    ***/
  dt_Handle bond;             /*** Bond object                         ***/
  dt_Handle bondstream;       /*** Bond stream                         ***/
  dt_Handle dep;              /*** Depiction                           ***/
  dt_Handle hyd;              /*** Hydrogen                            ***/
  dt_Handle mol;              /*** Molecule                            ***/
  char *smi;                  /*** Input SMILES                        ***/
  char *cansmi;               /*** Canonical SMILES                    ***/
  char name[81];              /*** Name for line 1                     ***/
  char tag[80];               /*** Datatype tag                        ***/
  char nametag[10];           /*** TDT tag for molfile name            ***/
  char *p,*q,*s,*t,*pp;       /*** Generic char ptrs                   ***/
  char *s2d;                  /*** 2D data string                      ***/
  char *sd3d;                 /*** D3D data string                     ***/
  char *sxyzs;                /*** 3D xyz string                       ***/
  char *sym;                  /*** Atom symbol                         ***/
  char *tdt = NULL;           /*** TDT string                          ***/
  int a1;                     /*** "From" atom                         ***/
  int a2;                     /*** "To" atom                           ***/
  int ai;                     /*** Atom index                          ***/
  int bi;                     /*** Bond index                          ***/
  int border;                 /*** Bond order (1, 2, or 3)             ***/
  int btype;                  /*** Bond type (1, 2, 3, or 4)           ***/
  int bufsize = 0;            /*** TDT buffer size                     ***/
  int charge;                 /*** Charge                              ***/
  int chiflag;                /*** Bond chirality flag                 ***/
  int dborder;                /*** Double Bond chirality order         ***/
  int count;                  /*** Coordinate count                    ***/
  int error_count = 0;        /*** Error count                         ***/
  int i, j, k;                /*** Generic ints                        ***/
  int ii;                     /*** Generic int #4                      ***/
  int len;                    /*** Generic length                      ***/
  int lencan;                 /*** Length of canonical SMI str         ***/
  int lensym;                 /*** Length of symbol str                ***/
  int mol_out = 0;            /*** Count of MOLs written               ***/
  int depict_count = 0;       /*** Count of 2D depictions created      ***/
  int smilen;                 /*** Length of SMI str                   ***/
  int smi_in = 0;             /*** Count of SMI read                   ***/
  int tdt_in = 0;             /*** Count of TDTs read                  ***/
  int tdtlen;                 /*** Length of TDT str                   ***/
  int wegot_2d;               /*** Bool: 2D data is present            ***/
  int wegot_3d;               /*** Bool: 3D data is present            ***/
  int wegot_name;             /*** Bool: Name is present               ***/
  int explicit_hs = 0;        /*** Bool: Add explicit H's to molfile   ***/
  int input_is_smi = 1;       /*** Bool: Input format is SMILES.       ***/
  int input_is_tdt = 0;       /*** Bool: Input format is TDTs.         ***/
  int add_2d = 1;             /*** Bool: Add Depict 2Ds if needed.     ***/
  int write_hcount = 0;       /*** Bool: Write hcount field.           ***/
  int comment_is_smi=0;       /*** Bool: Write smi in comment          ***/
  int split_fields=0;         /*** Bool: multifields->multi-data       ***/
  int atnum;                  /*** Atomic number                       ***/
  int mass_diff;              /*** Mass difference                     ***/
  int weight;                 /*** Atomic weight                       ***/
  int orig_imp_htotal;        /*** Original implicit H total           ***/
  int minor_report = 25;      /*** Progress report counter             ***/
  int major_report = 1000;    /*** Progress report counter             ***/
  int errlev=DX_ERR_ERROR;    /*** Error level reported                ***/

  strcpy(nametag, "$NAM");

  /*****************************************
  ***  Parse command line.
  *****************************************/
  for (++argv; --argc; ++argv) {
    if (sameword(*argv, "-help")) help();
    if (sameword(*argv, "-explicit_hydrogens") ||
        sameword(*argv, "+h")) {
      explicit_hs = 1;
    } else if (sameword(*argv, "-s")) {
        input_is_smi = 1; input_is_tdt = 0;
    } else if (sameword(*argv, "-t")) {
        input_is_smi = 0; input_is_tdt = 1;
    } else if (sameword(*argv, "-input_format")) {
      if (--argc == 0) {
        fprintf(stderr, "ERROR (SMI2MOL): \"%s\" requires value...\n", *argv);
        help();
      } else if (sameword(*++argv, "SMI")) {
        input_is_smi = 1; input_is_tdt = 0;
      } else if (sameword(*argv, "TDT")) {
        input_is_smi = 0; input_is_tdt = 1;
      } else {
        fprintf(stderr, "ERROR (SMI2MOL): Unknown format: \"%s\"\n", *argv);
        help();
      }

#ifdef _DO_DEPICT
    } else if (sameword(*argv, "-add_2d")) {
      if (--argc == 0)       {
        fprintf(stderr, "ERROR (SMI2MOL): \"%s\" requires value...\n", *argv);
        help();
      } else if (sameword(*++argv, "TRUE"))       { add_2d = 1;
      } else if (sameword(*argv, "FALSE")) { add_2d = 0;
      } else {
        fprintf(stderr, "ERROR (SMI2MOL): Bad option value: \"%s\"\n", *argv);
        help();
      }
#endif

    } else if (sameword(*argv, "-name_datatype")) {
      if (--argc == 0) {
        fprintf(stderr, "ERROR (SMI2MOL): \"%s\" requires value...\n", *argv);
        help();
      } else {
        strcpy(nametag, *++argv);
      }
    } else if (sameword(*argv, "-write_hcount")) {
      write_hcount = 1;
    } else if (sameword(*argv, "-split_fields")) {
      split_fields = 1;
    } else if (sameword(*argv, "-comment_is_smi")) {
      comment_is_smi = 1;
    } else if (sameword(*argv, "-minor_report")) {
      if (--argc == 0) {
        fprintf(stderr, "ERROR (SMI2MOL): \"%s\" requires value...\n", *argv);
        help();
      } else {
        minor_report = atoi(*++argv);
      }
    } else if (sameword(*argv, "-major_report")) {
      if (--argc == 0) {
        fprintf(stderr, "ERROR (SMI2MOL): \"%s\" requires value...\n", *argv);
        help();
      } else {
        major_report = atoi(*++argv);
      }
    } else if (sameword(*argv,"-v")) {
      errlev=DX_ERR_WARN;
    } else if (sameword(*argv,"-vv")) {
      errlev=DX_ERR_NOTE;
    } else {
      fprintf(stderr, "ERROR (SMI2MOL): Bad option \"%s\"\n", *argv);
      help();
    }
  }

  /***************************************************
  *** TDT/SMILES-reading loop
  ***************************************************/
  tdt_in = mol_out = smi_in = depict_count = error_count = 0;
  while (input_is_tdt && du_readtdt(stdin,&tdtlen,&tdt,&bufsize) ||
         input_is_smi && (smi=du_fgetline(&len,stdin))) {

    if (input_is_tdt) ++tdt_in;

    wegot_name = wegot_2d = wegot_3d = 0;

    /***************************************************
    *** Find and interpret SMILES
    ***************************************************/
    if (input_is_smi || !strncmp("$SMI<", tdt, 5)) {
      ++smi_in;
      if (input_is_tdt) {
        smilen = du_find_character(-1, tdt, ">") - (tdt + 5);
        smi = tdt + 5;
        if (*smi == '"' && smi[smilen-1] == '"') {
          ++smi;  smilen -= 2;
        }
        mol = dt_smilin(smilen, smi);
      } else {
        smilen = strlen(smi);
        mol = dt_smilin(smilen, smi);
      }
      if (NULL_OB == mol) {
        sprintf(errtxt,"dt_smilin ... \"%.*s\"\n",smilen,smi);
        dt_errorsave("smi2mol",DX_ERR_ERROR,errtxt);
        ++error_count;
        goto nexttdt;
      }
    }
    else {
      mol = NULL_OB; smi = NULL;
      natoms = nbonds = 0;
      dt_errorsave("smi2mol",DX_ERR_NOTE,"No-struct read...");
    }

    /***********************************************************
    *** For both no-structs and structs, we get name.
    ***********************************************************/
    if (input_is_tdt && (p = strstr(tdt, nametag))) {
      q = du_find_character(-1, p, ">");
      k = MIN((q - p) - strlen(nametag) - 1, 80);
      strncpy(name, p + strlen(nametag) + 1, k);
      name[k] = '\0';
      wegot_name = 1;
    } else if (input_is_smi && (p = strchr(smi, ' '))) {
        strcpy(name, p + 1);
        if (*name) wegot_name = 1;
    }

    /***************************************************************
    *** If not a no-struct, get natoms, nbonds, 2D or D3D if present.
    ***************************************************************/
    if (NULL_OB != mol) {
      natoms = dt_count(mol, TYP_ATOM);
      nbonds = dt_count(mol, TYP_BOND);

      /***********************************************************
      *** (Re)Allocate ATOMINFO array of structs.
      *** (Re)Allocate BONDINFO array of structs.
      *** Make extra room for possible added H's (kludgy! --
      *** we're assuming no more than (3*natoms) H's.)
      ***********************************************************/
      if (!atoms) {
        max_natoms = 4 * natoms;
        atoms = (ATOMINFO *) malloc((max_natoms+1) * sizeof(ATOMINFO));
      } else if (max_natoms < 4 * natoms) {
        max_natoms = 4 * natoms;
        atoms = (ATOMINFO *) realloc(atoms, (max_natoms+1)*sizeof(ATOMINFO));
      }
      if (!bonds) {
        max_nbonds = MAX(4 * nbonds, 4 * natoms);
        bonds = (BONDINFO *) malloc((max_nbonds+1) * sizeof(BONDINFO));
      } else if (max_nbonds < MAX(4 * nbonds, 4 * natoms)) {
        max_nbonds = MAX(4 * nbonds, 4 * natoms);
        bonds = (BONDINFO *) realloc(bonds, (max_nbonds+1)*sizeof(BONDINFO));
      }

      orig_imp_htotal = 0;
      atomstream = dt_stream(mol, TYP_ATOM);
      while (NULL_OB != (atom = dt_next(atomstream))) {
        orig_imp_htotal += dt_imp_hcount(atom);
      }
      dt_dealloc(atomstream);

      /*****************************************************************
      *** Look for ALAB which may be MDL pseudo-atom symbols.
      *****************************************************************/
      if (input_is_tdt && (p = strstr(tdt, "ALAB<"))) {
        q = du_find_character(-1, p, ">");
        for (s = p + 5, i = 1; s < q; s = t + 1, ++i) {
          t = du_find_character(-1, s, ",");
          if (t > s) strncpy(atoms[i].symbol, s, t - s);
          (atoms[i].symbol)[t - s] = '\0';
        }
      }

      /*****************************************************************
      *** If explicit_hydrogens, make all implicit H's explicit in
      *** molecule.
      *****************************************************************/
      if (explicit_hs) {
        atomstream = dt_stream(mol, TYP_ATOM);
        atomseq = stream2seq(atomstream);
        dt_dealloc(atomstream);
        dt_mod_on(mol);
        while (NULL_OB != (atom = dt_next(atomseq))) {
          for (j = dt_imp_hcount(atom);j > 0; --j) {
            hyd = atoms[++natoms].h = dt_addatom(mol, DX_ATN_H, 0);
            atoms[natoms].parity = 0;
            dt_setadjunct(hyd, (dt_Handle)natoms);
            bond = bonds[++nbonds].h = dt_addbond(atom, hyd, 1);
            dt_setadjunct(bond, (dt_Handle)nbonds);
            bonds[nbonds].from_atom = atom;
            bonds[nbonds].to_atom = hyd;
            bonds[nbonds].stereo = 0;
            dt_setimp_hcount(atom, 0);
          }
        }
        dt_dealloc(atomseq);
      }

      /*****************************************************************
      *** Associate atoms with their index as per original SMILES order.
      *** Compute total implicit H-count.  Note that this is needed for
      *** the 2D or D3D xyz's to maintain their association.
      *** (Not necessary for the bonds.)
      *****************************************************************/
      imp_htotal = 0;
      atomstream = dt_origstream(mol, TYP_ATOM);
      for (i = 1; NULL_OB != (atom = dt_next(atomstream)); ++i) {
        imp_htotal += dt_imp_hcount(atom);
        dt_setadjunct(atom, (dt_Handle) i);
        atoms[i].h = atom;
        atoms[i].x = atoms[i].y = atoms[i].z = 0.0;
        atoms[i].parity = 0;
      }
      bondstream = dt_stream(mol, TYP_BOND);
      for (i = 1; NULL_OB != (bond = dt_next(bondstream)); ++i) {
        dt_setadjunct(bond, (dt_Handle) i);
        bonds[i].h = bond;
        xatoms = dt_stream(bond, TYP_ATOM);
        bonds[i].from_atom = dt_next(xatoms);
        bonds[i].to_atom = dt_next(xatoms);
        dt_dealloc(xatoms);
        bonds[i].stereo = 0;
      }

      /***************************************************
      *** Read in data to string s2d.  Replace
      *** commas with spaces; find spaces and sscan
      *** each real.
      ***************************************************/
      if (input_is_tdt && (p = strstr(tdt, "2D<"))) {
        wegot_2d = 1;
        q = du_find_character(-1, p, ">");
        s2d = (char *)malloc((q - p - 3 + 1)*sizeof(char));
        strncpy(s2d, p + 3, q - p - 3);
        s2d[q - p - 3] = '\0';
        for (j = 0; s2d[j] != '\0'; ++j)  if (',' == s2d[j]) s2d[j] = ' ';
        for (j = count = 0, ii = 1; j < strlen(s2d); ++ii) {
          count += sscanf(s2d + j, "%f", &atoms[ii].x);
          do ++j; while (' ' != s2d[j] && '\0' != s2d[j]);
          count += sscanf(s2d + j, "%f", &atoms[ii].y);
          do ++j; while (' ' != s2d[j] && '\0' != s2d[j]);
        }
        free(s2d);
      }
  
      /***************************************************
      *** Read in D3D data to sd3d.
      *** Copy only xyz's to sxyzs (between 2nd and
      *** 3rd semicolon).  Replace commas w/ spaces.
      *** Find spaces and sscan each real.
      ***************************************************/
      if (input_is_tdt && (p = strstr(tdt, "$D3D<"))) {
        q = du_find_character(-1, p, ">");
        sd3d = (char *) malloc((q - p - 5 + 1)*sizeof(char));
        strncpy(sd3d, p + 5, q - p - 5);
        sd3d[q - p - 5] = '\0';
  
        i = j = 0;
        while (';' != sd3d[j++])  ;
        while (';' != sd3d[j++])  ;
        while (';' != sd3d[j + i++])  ;
        sxyzs = (char *) malloc(i*sizeof(char));
        strncpy(sxyzs, sd3d + j, i - 1);
        free(sd3d);
        sxyzs[i - 1] = '\0';
        for(j = 0; j < i; ++j) if (',' == sxyzs[j]) sxyzs[j] = ' ';
        
        for (j = count = 0, ii = 1; j < i - 1; ++ii) {
          count += sscanf(sxyzs + j, "%f", &atoms[ii].x);
          do ++j; while (' ' != sxyzs[j] && '\0' != sxyzs[j]);
          count += sscanf(sxyzs + j, "%f", &atoms[ii].y);
          do ++j; while (' ' != sxyzs[j] && '\0' != sxyzs[j]);
          count += sscanf(sxyzs + j, "%f", &atoms[ii].z);
          do ++j; while (' ' != sxyzs[j] && '\0' != sxyzs[j]);
        }
        free(sxyzs);
  
        /***************************************************
        *** D3D is allowed to have all coordinates, or
        *** only coordinates for explicit atoms (no
        *** implicit H's), or some coordinates but all
        *** commas.  Check for legal count.
        ***************************************************/
        if (3*natoms == count  || 
            3*(natoms + imp_htotal) == count ||
            3*(natoms - orig_imp_htotal) == count
           ) wegot_3d = 1;
        else {
          wegot_3d = 0;
          dt_errorsave("smi2mol",DX_ERR_ERROR,"Bad D3D field.");
          ++error_count;
        }
      }
  
      /************************************************************
      *** Generate 2D coordinates if needed & requested.
      *** This section will be commented out if _NO_DEPICT is defined
      *** to allow this program to be compiled without a Depict
      *** Toolkit license.
      ************************************************************/

#ifdef _DO_DEPICT
      if (add_2d && !wegot_2d && !wegot_3d) {  
        dep = dt_alloc_depiction(mol);
        dt_setisomeric(dep, 1);
        dt_calcxy(dep);
        for (i = 1; i <= natoms; ++i) {
          dt_getcoord(dep, atoms[i].h, &atoms[i].x, &atoms[i].y, &atoms[i].z);
        }
        dt_dealloc(dep);
        ++depict_count;
        wegot_2d = 1;
      }
#endif

      /********************************************************
      * Note parity is a write only feature it is never read.
      * Yes really !!! -JB.
      ********************************************************/
      setparity(mol);

      /************************************************************
      *** Convert tetrahedral chirality to up/down bonds.
      *** Does this make sense if wegot_3d?  I think not.
      *** Note that setchibonds() may set mol mod_on.
      *** Explicit H's may be added and natoms and nbonds increased.
      ************************************************************/
      if (wegot_2d) {
        dt_dealloc(atomstream);
        atomstream = NULL_OB;
        if (!setchibonds(mol)) {
          dt_errorsave("smi2mol",DX_ERR_ERROR,"setchibonds failed...");
          ++error_count;
        }

        /***************************************************
        *** Convert cis/trans chirality to stereo.
        ***************************************************/
        if (!setezstereo(mol)) {
          dt_errorsave("smi2mol",DX_ERR_ERROR,"setezstereo failed...");
          ++error_count;
        }
      }
    }

    /*************************************************************
    *** Let's write out the MOL file.
    *** First the header: name, source, comment.
    *************************************************************/
    printf("%s\n", wegot_name?name:"");
    printf("  SMI2MOL \n");
    if (NULL_OB != mol) {
      if (dt_mod_is_on(mol)) {
        if (!dt_mod_off(mol)) {
          dt_errorsave("smi2mol",DX_ERR_ERROR,"dt_mod_off failed...");
          ++error_count;
          mol = NULL_OB;
        }
      }
    }
    if (comment_is_smi && NULL_OB!=mol && smi!=NULL) {
      printf("%.*s", smilen, smi);
    }
    printf("\n");

    /*************************************************************
    *** Next the counts-line: 
    ***      aaabbblllfffcccsssxxxrrrpppiiimmmvvvvvv
    *** aaa: number of atoms
    *** bbb: number of bonds
    *** lll: number of atom lists
    *** fff: (obsolete)
    *** ccc: chiral flag
    *** sss: number of stext entries
    *** xxx: number of reaction components + 1
    *** rrr: number of reactants
    *** ppp: number of products
    *** iii: number of intermediates
    *** mmm: number of lines of additional properties (obsolete;
    ***      default set to 999)
    *** vvvvvv: Ctab version; V2000 or V3000
    **************************************************************
    *** ccc is set to 1 = CHIRAL if the absolute chirality is
    *** known.  Can only guarantee this if wegot_3d = TRUE.
    *************************************************************/
    printf("%3d%3d  0  0%3d  0  0  0  0  0999 V2000\n",
            natoms, nbonds, wegot_3d);

    if (NULL_OB != mol) {

      /*************************************************************
      * Now the atom lines:
      * xxxxx.xxxxyyyyy.yyyyzzzzz.zzzz aaaddcccssshhhbbbvvvHHHrrriiimmmnnneee
      *************************************************************/
      for (i = 1; i <= natoms; ++i) {
        sym = NULL;
        mass_diff = 0;
        if (NULL_OB != atoms[i].h) {
          /*** xxxxx.xxxxyyyyy.yyyyzzzzz.zzzz: atom coordinates ***/
          printf("%10.4f%10.4f%10.4f ", atoms[i].x, atoms[i].y, atoms[i].z); 

          /***************************************************
          *** Deuterium, tritium, and mass-difference.
          ***************************************************/
          if (0 < (weight = dt_weight(atoms[i].h))) {
            atnum = dt_number(atoms[i].h);
            if (atnum == 1 && weight == 2) {
              sym = "D"; mass_diff = 0;
            } else if (atnum == 1 && weight == 3) {
              sym = "T"; mass_diff = 0;
            } else {
              mass_diff = weight - atnum2mass(atnum);
            }
          }

          /***************************************************
          *** Check for wildcard atoms which might have
          *** MDL symbols from ALAB<>.
          ***************************************************/
          if (0 == dt_number(atoms[i].h) && *(atoms[i].symbol))
            { sym = atoms[i].symbol; }

          /*** aaa: atom symbol ***/
          /*** dd: mass difference ***/
          if (!sym) sym = dt_symbol(&lensym, atoms[i].h);
          printf("%-3s%2d", sym, mass_diff);

          /***************************************************
          *** Note format: (charge) = 4 - (MDL charge)
          *** --> except 0.
          ***************************************************/
          charge = dt_charge(atoms[i].h);
          if (0 != charge) charge = 4 - charge;
          printf("%3d", charge); /*** ccc: charge ***/

          /*** sss: atom stereo parity ***/
          printf("  %1d", atoms[i].parity);

          /***************************************************
          *******   jb2314@ggr.co.uk 1-October-1996    *******
          *** hhh: hydrogen count + 1
          *** This line is used for query and means the following
          *** 1 = H0  i.e. no H atoms allowed unless explicitly drawn
          *** 2 = H1, 3 = H2 ...5 = H4 i.e. Hn means atom must have
          *** n or more H's in excess of explicit H's
          ***
          *** Should be set to zero for normal use.
          ***************************************************/
          if (write_hcount) {
            printf("%3d", dt_hcount(atoms[i].h) + 1); 
          } else { printf("  0"); }

          printf("  0");  /*** bbb: stereo care box ***/
          printf("  0");  /*** vvv: valence ***/
          printf("  0");  /*** HHH: H0 designator ***/
          printf("  0");  /*** rrr: reaction component type ***/
          printf("  0");  /*** iii: reaction component number ***/
          printf("  0");  /*** mmm: atom-atom mapping number ***/
          printf("  0");  /*** nnn: inversion/retention flag ***/
          printf("  0");  /*** eee: exact change flag ***/
          printf("\n");
        }
      }
  
      /***********************************************************
      *** Now the bond lines:
      *** 111222tttsssxxxrrrccc
      *** 111: first atom number
      *** 222: second atom number
      *** ttt: bond type
      *** sss: bond stereo
      *** xxx: not used
      *** rrr: bond topology
      *** ccc: reacting center status
      ************************************************************
      *** Now the bonds.  Note that we use
      *** dt_bondorder and not dt_bondtype, because
      *** aromatic bonds are not allowed in the MDL
      *** SD format (only in query molfiles).
      ***********************************************************/
      for (bi = 1; bi <= nbonds; ++bi) {
        if (NULL_OB != bonds[bi].h) {
          atom1 = bonds[bi].from_atom;
          atom2 = bonds[bi].to_atom;
          border = dt_bondorder(bonds[bi].h);
          a1 = (int) dt_adjunct(atom1);
          a2 = (int) dt_adjunct(atom2);
          printf("%3d%3d%3d%3d", a1, a2, border, bonds[bi].stereo);
          printf("  0");  /*** xxx: not used ***/
          printf("  0");  /*** rrr: bond topology ***/
          printf("  0");  /*** ccc: reacting center status ***/
          printf("\n");
        }
      }
    }
    /***************************************************
    *******   jb2314@ggr.co.uk 1-October-1996    *******
    *** No blank lines allowed after M END before DATA
    ***************************************************/
    printf("M  END\n"); 
        
    /***************************************************
    *** Now extract other data, and add in MDL "SD"
    *** format to end of table.  Skip 2D and $D3D.
    *** Remove Daylight quotes.
    ***************************************************/ 
    if (input_is_tdt) {
      if (NULL_OB == mol) { p = tdt; }
      else { p = du_find_character(-1, tdt, ">") + 1; }
      while ((NULL != p) && (p < tdt + strlen(tdt))) {
        if (p == du_find_character(-1, p, "|")) break;
        if ((!strncmp(p, "$D3D", 4)) || (!strncmp(p, "2D", 2))) {
          p = du_find_character(-1, p, ">") + 1;
          continue;
        }
        q = du_find_character(-1, p, "<");
        strncpy(tag,p,q-p);
        tag[q-p] = '\0';
        p = q + 1;
        t = du_find_character(-1,p,">");
        q = du_find_character(-1,p,";");

        /***************************************************
        *** multi-field datatype - handle by separating into
        *** multiple SD data, with additional data header
        *** info as follows:
        *** 
        *** CP<-3.124;-0P;4.72>  converted to...
        *** 
        *** > <CP> FIELD 1 OF 3
        *** -3.124
        ***
        *** > <CP> FIELD 2 OF 3
        *** -0P
        ***
        *** > <CP> FIELD 3 OF 3
        *** 4.72
        **************************************************** 
        *** Question: are null data allowed in SD?  No.
        *** Kludge: use '~' to mean null content.
        ***************************************************/ 
        if (split_fields && q!=NULL && q<t) {

          for (q=p,k=1; NULL!=(q=du_find_character(-1,q,";")) && q<t;++q) ++k;
          for (i=1; i<=k; ++i) {
            q=du_find_character(-1,p,";");
            if (NULL==q) q=du_find_character(-1,p,">");
            s = du_unquote(q-p,p); /*** NULL if len==0 ***/
            printf("> <%s> FIELD %d of %d\n%s\n\n",tag,i,k,s?s:"~");
            if (s) free(s);
            p = q + 1;
          }

        /*** single-field datatype ***/
        } else {
          s = du_unquote(t-p,p);
          printf("> <%s>\n%s\n\n",tag,s);
          free(s);
        }
        p = t + 1;
      }
    }

    /***************************************************
    ***  Done with mol.  Find end of TDT or EOF.
    ***************************************************/
    printf("$$$$\n");
    ++mol_out;
    if (NULL_OB!=mol) dt_dealloc(mol);

    /******************************************************
    *** Progress reports...
    ******************************************************/
    k = (input_is_tdt ? tdt_in : smi_in);
    if (minor_report && !(k%minor_report))
      { fprintf(stderr, "."); fflush(stderr); }
    if (major_report && !(k%major_report))
      fprintf(stderr, " %d %s, %d errors\n",
        k,input_is_tdt?"TDTs":"SMIs",error_count);

nexttdt:
    du_printerrors(stderr,errlev);
    dt_errorclear();
    continue;
  }
  fprintf(stderr, "NOTE (SMI2MOL): Normal exit (EOF)...\n");

errorexit:
  du_printerrors(stderr,errlev);
  if (input_is_tdt)
    fprintf(stderr, "SMI2MOL: TDT's in:  %d\n", tdt_in);
  fprintf(stderr, "SMI2MOL: SMI's in:  %d\n", smi_in);
  fprintf(stderr, "SMI2MOL: MOL's out: %d\n", mol_out);
  fprintf(stderr, "SMI2MOL: 2D Depictions added: %d\n", depict_count);
  fprintf(stderr, "SMI2MOL: errors: %d\n", error_count);
  fprintf(stderr, "Adios!\n");
}

/*************************************************************************
*** setparity - set the MDL stereo/chiral parity for each atom.
*************************************************************************/
int setparity(dt_Handle mol)
{
  dt_Handle atomstream;            /*** Atom stream                   ***/
  dt_Handle atom;                  /*** Atom                          ***/
  dt_Handle xatoms;                /*** Connected atom stream         ***/
  dt_Handle bondstream;            /*** Bond stream                   ***/
  dt_Handle bond;                  /*** Bond                          ***/
  dt_Handle bseq;                  /*** Bond sequence                 ***/
  dt_Handle aseq;                  /*** Atom sequence                 ***/
  dt_Handle a1;                    /*** Atom 1                        ***/
  dt_Handle a2;                    /*** Atom 2                        ***/
  int ai;                          /*** Atom index                    ***/
  int i, j;                        /*** Generic indices               ***/
  int chival;                      /*** Chiral value                  ***/
  int nconn;                       /*** Number of connections         ***/

  atomstream = dt_stream(mol, TYP_ATOM);
  while (NULL_OB != (atom = dt_next(atomstream))) {

    bondstream = dt_stream(atom, TYP_BOND);
    nconn = dt_count(bondstream, TYP_BOND);
    if (nconn != 3 && nconn != 4) {
      dt_dealloc(bondstream);
      continue;
    }

    bseq = stream2seq(bondstream);
    if (nconn == 3) dt_append(bseq, dt_isohydro());

    chival = dt_chival(atom, bseq);
    if (DX_CHI_THCW != chival && DX_CHI_THCCW != chival) {
      dt_dealloc(bondstream);
      dt_dealloc(bseq);
      continue;
    }

    aseq = dt_alloc_seq();
    dt_reset(bondstream);
    while (NULL_OB != (bond = dt_next(bondstream)))
      dt_append(aseq, dt_xatom(atom, bond));

    /***************************************************
    *** Sort aseq with respect to order in MOLfile.
    ***************************************************/
    for (i = 1; i < nconn; ++i) {
      dt_reset(aseq);
      a1 = dt_next(aseq);
      while (NULL_OB != (a2 = dt_next(aseq))) {
        if ((int) dt_adjunct(a1) > (int) dt_adjunct(a2)) {
          dt_delete(aseq);
          dt_insert(aseq, a2);
        } else a1 = a2;
      }
    }

    /*******************************************
    *** Reassemble bond-sequence in order.
    *******************************************/
    dt_reset(bseq);
    while (NULL_OB != dt_next(bseq)) dt_delete(bseq);

    dt_reset(aseq);
    while (NULL_OB != (a1 = dt_next(aseq)))
      dt_append(bseq, dt_bond(atom, a1));
    if (nconn == 3) dt_append(bseq,dt_isohydro());

    chival = dt_chival(atom, bseq);
    ai = (int) dt_adjunct(atom);
    switch (chival) {
      case DX_CHI_THCW:  atoms[ai].parity = 1; break;
      case DX_CHI_THCCW: atoms[ai].parity = 2; break;
      case DX_CHI_NONE:  atoms[ai].parity = 0; break;
      default:           atoms[ai].parity = 0;
    }
    dt_dealloc(bondstream);
    dt_dealloc(bseq);
    dt_dealloc(aseq);
  }
  return (1);
}


/*************************************************************************
*** setchibonds - set molfile up/down bonds from SMILES chirality.
***               This is only done if writing 2D coordinates.
*************************************************************************/
int setchibonds(dt_Handle mol)
{
  dt_Handle atom;             /*** Atom object                        ***/
  dt_Handle xatom;            /*** Other atom object                  ***/
  dt_Handle atomstream;       /*** Atom stream                        ***/
  dt_Handle atomseq;          /*** Atom sequence                      ***/
  dt_Handle batoms;           /*** Bond Atom stream                   ***/
  dt_Handle bond;             /*** Bond object                        ***/
  dt_Handle bondstream;       /*** Bond stream                        ***/
  dt_Handle bondseq;          /*** Bond sequence                      ***/
  dt_Handle ccwbseq;          /*** CCW bond sequence                  ***/
  int hcount;                 /*** Atom hydrogen count                ***/
  int ai;                     /*** Atom index                         ***/
  int bi;                     /*** Bond index                         ***/
  int i;                      /*** Generic index                      ***/
  int nbs;                    /*** Number of bonds to atom            ***/
  int j, k, l;                /*** Generic indices                    ***/
  int chival;                 /*** Chiral value                       ***/
  float signed_ang;           /*** Signed angle                       ***/

  struct {                         /*** Struct for chiral info        ***/
    dt_Handle bond;                /*** Attached chiral bond          ***/
    dt_Handle atom;                /*** Attached chiral atom          ***/
    float x; float y; float z;     /*** Relative coordinates          ***/
    float r; float theta;          /*** Relative 2D polar coords      ***/
  } chiinfo[4];

  int ccworder[4];                 /*** chiinfo indices in ccw order  ***/
  
  atomstream = dt_stream(mol, TYP_ATOM);
  atomseq = stream2seq(atomstream);
  dt_dealloc(atomstream);

  if (!dt_mod_is_on(mol)) dt_mod_on(mol);

  while (NULL_OB != (atom = dt_next(atomseq))) {
    batoms = dt_stream(atom, TYP_BOND);
    bondseq = stream2seq(batoms);
    dt_dealloc(batoms);

    /********************************************************
    *** Must have four bonds or three + one hydrogen.
    *** Now (nbs) == 3 or 4 and (hcount) == 0 or 1.
    ********************************************************/
    nbs = dt_count(bondseq, TYP_BOND);
    hcount = dt_imp_hcount(atom);
    if (nbs + hcount != 4 || hcount > 1) {
      dt_dealloc(bondseq);
      continue;
    }
    if (hcount) dt_append(bondseq, dt_isohydro());

    /********************************************************
    *** If not chiral center, go to next atom.
    ********************************************************/
    chival = dt_chival(atom, bondseq);
    if (DX_CHI_THCW != chival && DX_CHI_THCCW != chival) {
      dt_dealloc(bondseq);
      continue;
    }

/************************DEBUG {
dt_Handle xatom;
int len;
char *p;
  fprintf(stderr, "DEBUG: chival = %s...\n",
  chival == DX_CHI_THCW ? "DX_CHI_THCW" : "DX_CHI_THCCW");
  dt_reset(bondseq);
  p = dt_symbol(&len, atom);
  fprintf(stderr,
  "\tBonds from %.*s %d (handle = %d) are:\n", len, p,
  (int) dt_adjunct(atom), atom);
  while(NULL_OB != (bond = dt_next(bondseq))) {
    if (TYP_ATOM == dt_type(bond)) {
      fprintf(stderr,
      "\t(ain't a bond, must be H)\n");
    } else {
      xatom = dt_xatom(atom, bond);
      p = dt_symbol(&len, xatom);
      fprintf(stderr,
      "\tbond %d (h=%d) to %.*s %d (h=%d)...\n",
      (int)dt_adjunct(bond),
      (int)bond,
      len, p,
      (int)dt_adjunct(xatom),
      (int)xatom);
    }
  }
}************************/

    /********************************************************
    *** Now it's geometry time.  We locate the center atom
    *** at (0,0) and translate the other 4 (or 3).  Then
    *** we compute the polar angle to sort these atoms
    *** (anticlockwise).
    ********************************************************/
    ai = (int) dt_adjunct(atom);
    dt_reset(bondseq);
    for (j = 0; (j < nbs) && (NULL_OB != (bond = dt_next(bondseq))); ++j) {
      chiinfo[j].bond = bond;
      xatom = dt_xatom(atom, bond);
      i = (int) dt_adjunct(xatom);
      chiinfo[j].atom = xatom;
      chiinfo[j].x = atoms[i].x - atoms[ai].x;
      chiinfo[j].y = atoms[i].y - atoms[ai].y;
      chiinfo[j].r = sqrt((chiinfo[j].x) * (chiinfo[j].x) +
                          (chiinfo[j].y) * (chiinfo[j].y));
      if (fabs(chiinfo[j].r) <= FLT_EPSILON) {
        dt_dealloc(bondseq);
        goto next_atom;
      }
      chiinfo[j].theta = (chiinfo[j].y >= 0.0) ?
                         (acos(chiinfo[j].x / chiinfo[j].r)) :
                         (2.0*PI - acos(chiinfo[j].x / chiinfo[j].r));
    }

    /********************************************************
    *** Bubble sort of ccworder[].  Start with 0123 order.
    ********************************************************/
    for (j = 0; j < nbs; ++j) { ccworder[j] = j; }
    for (j = 0; j < nbs - 1; ++j) {
      for (k = 0; k < nbs - 1 - j; ++k) {
        if (chiinfo[ccworder[k + 1]].theta < chiinfo[ccworder[k]].theta) {
          l = ccworder[k];
          ccworder[k] = ccworder[k + 1];
          ccworder[k + 1] = l;
        }
      }
    }

/************************DEBUG {
int len;
fprintf(stderr, "DEBUG: ccw-ordered chi bonds at angles:\n");
for (j = 0; j < nbs; ++j) {
  fprintf(stderr,
  "\tchiinfo[ccworder[%d]=%d] bond %d (h=%d) to %c %d (h=%d) : %6.1fdeg\n",
  j, ccworder[j],
  (int) dt_adjunct(chiinfo[ccworder[j]].bond),
  chiinfo[ccworder[j]].bond,
  *dt_symbol(&len, chiinfo[ccworder[j]].atom),
  (int) dt_adjunct(chiinfo[ccworder[j]].atom),
  chiinfo[ccworder[j]].atom,
  chiinfo[ccworder[j]].theta * 180.0 / PI);
}
} ************************/

    /*********************************************************
    *** Next create a new sequence in ccw order.  Determine
    *** chival wrt this order, and set bond stereo up/down
    *** accordingly (in one of several possible stereo-
    *** equivalent arrangements).
    *********************************************************/
    ccwbseq = dt_alloc_seq();
    for (j = 0; j < nbs; ++j) {
      dt_append(ccwbseq, chiinfo[ccworder[j]].bond);
    }
    if (hcount) dt_append(ccwbseq, dt_isohydro());
    chival = dt_chival(atom, ccwbseq);


    /********************************************************
    *** Now we have ccw order:
    *** 
    ***    chiinfo[ccworder[0]].atom         nbs == 4
    ***              |
    ***              |
    ***  [1] --------*-------- [3]
    ***              |
    ***              |
    ***             [2]
    ***                          
    ***    ...or...
    ***
    ***    chiinfo[ccworder[0]].atom          nbs == 3
    ***              |
    ***              |
    ***  [1]---------* - - - - - H  <- not-yet explicit H
    ***              |                 
    ***              |                 
    ***             [2]                
    ***                          
    *** We "average" both the bond lengths and angles of
    *** the 1st and 3rd bonds to locate the new explicit H
    *** between them.  Then translate back from relative
    *** coordinates.
    ********************************************************/
    if (3 == nbs) {
      ++natoms;
      --imp_htotal;
      atoms[natoms].h = dt_addatom(mol, DX_ATN_H, 0);
      dt_setadjunct(atoms[natoms].h, (dt_Handle)natoms);
      atoms[natoms].parity = 0;
      ccworder[3] = 3;
      chiinfo[3].atom = atoms[natoms].h;
      ++nbonds;
      bonds[nbonds].h = chiinfo[3].bond = dt_addbond(atom, atoms[natoms].h, 1);
      dt_setimp_hcount(atom, 0);
      dt_setadjunct(bonds[nbonds].h, (dt_Handle)nbonds);
      chiinfo[3].r = (chiinfo[ccworder[0]].r + chiinfo[ccworder[2]].r) / 2.0;
      signed_ang = 
        (chiinfo[ccworder[2]].theta - 2*PI + chiinfo[ccworder[0]].theta)/2.0;
      chiinfo[3].theta = signed_ang > 0 ?  signed_ang : 2 * PI + signed_ang;
      chiinfo[3].x = chiinfo[3].r * cos(chiinfo[3].theta);
      atoms[natoms].x = chiinfo[3].x + atoms[ai].x;
      chiinfo[3].y = chiinfo[3].r * sin(chiinfo[3].theta);
      atoms[natoms].y = chiinfo[3].y + atoms[ai].y;
      atoms[natoms].z = 0.0;
      nbs = 4;

/***********************DEBUG {
fprintf(stderr, "\t with r = %.1f, theta = %.1fdeg, x = %.1f, y = %.1f ...\n",
chiinfo[3].r, chiinfo[3].theta * 180 / PI, chiinfo[3].x, chiinfo[3].y);
fprintf(stderr, "\t bond h = %d , atom h = %d ...\n",
chiinfo[3].bond, chiinfo[3].atom);
}****************************/

    }

    /********************************************************
    *** Reassign same chival with new explicit H.
    ********************************************************/
    dt_dealloc(ccwbseq);
    ccwbseq = dt_alloc_seq();
    for (j = 0; j < nbs; ++j) {
      dt_append(ccwbseq, chiinfo[ccworder[j]].bond);
    }
    dt_setchival(atom, ccwbseq, chival);


/***********************DEBUG {
dt_Handle bond, xatom;
int len;
fprintf(stderr, "DEBUG: chival = %d ...\n", chival);
if (chival < 0)  {
  fprintf(stderr, "DEBUG: dt_chival error (%d) ...\n", chival);
  dt_reset(ccwbseq);
  while (NULL_OB != (bond = dt_next(ccwbseq))) {
   if (TYP_ATOM == dt_type(bond)) {
      fprintf(stderr,
      "\t(ain't a bond, must be H)\n");
    } else {
      xatom = dt_xatom(atom, bond);
      fprintf(stderr,
      "\tbond %d (h=%d) to %c %d (h=%d)...\n",
      (int)dt_adjunct(bond),
      (int)bond,
      *dt_symbol(&len, xatom),
      (int)dt_adjunct(xatom),
      (int)xatom);
    }
  }
}
} ***********************/

    if (DX_CHI_THCW == chival || DX_CHI_THCCW == chival) {
      dt_reset(ccwbseq);
      while (NULL_OB != (bond = dt_next(ccwbseq))) {
        bi = (int) dt_adjunct(bond);
        bonds[bi].from_atom = atom;
        bonds[bi].to_atom = dt_xatom(atom, bond);
      }
      if (DX_CHI_THCW == chival) {
        bonds[(int)dt_adjunct(chiinfo[ccworder[0]].bond)].stereo = 6; /*DOWN*/
        bonds[(int)dt_adjunct(chiinfo[ccworder[1]].bond)].stereo = 1; /*UP*/
        bonds[(int)dt_adjunct(chiinfo[ccworder[2]].bond)].stereo = 6; /*DOWN*/
        bonds[(int)dt_adjunct(chiinfo[ccworder[3]].bond)].stereo = 1; /*UP*/
      } else {
        bonds[(int)dt_adjunct(chiinfo[ccworder[0]].bond)].stereo = 1; /*UP*/
        bonds[(int)dt_adjunct(chiinfo[ccworder[1]].bond)].stereo = 6; /*DOWN*/
        bonds[(int)dt_adjunct(chiinfo[ccworder[2]].bond)].stereo = 1; /*UP*/
        bonds[(int)dt_adjunct(chiinfo[ccworder[3]].bond)].stereo = 6; /*DOWN*/
      }
    }
    dt_dealloc(ccwbseq);

next_atom: continue;
  }
  dt_dealloc(atomseq);
  return (1);
}
 
/*************************************************************************
*** setezstereo - set bond stereo from cis/trans chirality.
*** 0 = as per 2Ds; 3 = don't care.
*************************************************************************/
int setezstereo(dt_Handle mol)
{
  dt_Handle ob1;              /***                                    ***/
  dt_Handle ob2;              /***                                    ***/
  dt_Handle tb;               /***                                    ***/
  dt_Handle bs1;              /*** Bondstream 2                       ***/
  dt_Handle bs2;              /*** Bondstream 1                       ***/
  dt_Handle atom1;            /*** Atom object                        ***/
  dt_Handle atom2;            /*** Atom object                        ***/
  dt_Handle atomstream;       /*** Atom stream                        ***/
  dt_Handle batoms;           /*** Bond Atom stream                   ***/
  dt_Handle bond;             /*** Bond object                        ***/
  dt_Handle bondstream;       /*** Bond stream                        ***/
  dt_Handle bondseq;          /*** Bond sequence                      ***/
  int bi;                     /*** Bond index                         ***/
  int dbval;                  /*** Double bond chiral value           ***/

  bondstream = dt_stream(mol, TYP_BOND);
  while (NULL_OB != (bond = dt_next(bondstream))) {
    if (dt_bondorder(bond) == DX_BTY_DOUBLE) {
      bi = (int) dt_adjunct(bond);
      batoms = dt_stream(bond, TYP_ATOM);
      atom1 = dt_next(batoms);
      atom2 = dt_next(batoms);
      dt_dealloc(batoms);

      ob1 = ob2 = NULL_OB;
      bs1 = dt_stream(atom1, TYP_BOND);
      bs2 = dt_stream(atom2, TYP_BOND);
      while ((tb = dt_next(bs1)) != NULL_OB) {
        if (tb != bond) { ob1 = tb; break; }
      }
      while ((tb = dt_next(bs2)) != NULL_OB) {
        if (tb != bond) { ob2 = tb; break; }
      }
      dt_dealloc(bs1);
      dt_dealloc(bs2);
      if (ob1 != NULL_OB && ob2 != NULL_OB) {
        dbval = dt_dbo(bond, ob1, ob2);
        if (dbval == DX_CHI_NO_DBO) {
          bonds[bi].stereo = 3;
        } else {
          bonds[bi].stereo = 0;
        }
      }
    }
  }
  dt_dealloc(bondstream);
  return (1);
}

/********************************************************************
***  stream2seq - Copy a stream to a new sequence.
********************************************************************/
dt_Handle stream2seq(dt_Handle stream)
{
  dt_Handle seq = dt_alloc_seq(), ob;
 
  while (NULL_OB != (ob = dt_next(stream))) {
    dt_append(seq, ob);
  }
  return seq;
}

/********************************************************************
***  atnum2mass - Return mass given atomic number.
***  Returns total number of neutron and protons for given atnum.
***  If naturally occurring, for the most common isotope.
***  If not naturally occurring, for the most stable isotope.
***  Zero (0) if not known - error.
***  Reference: CRC Handbook of Chemistry and Physics, 60th Ed.
********************************************************************/
int atnum2mass(int atnum)
{
  static int mass[] = {
    0,
    1,  4,  7,  9,  11, 12, 14, 16, 19, 20,
    23, 24, 27, 28, 31, 32, 35, 40, 39, 40,
    45, 48, 51, 52, 55, 56, 59, 58, 63, 64,
    69, 74, 75, 80, 79, 84, 85, 88, 89, 90,
    93, 98, 97, 102,103,106,107,114,115,120,
    121,130,127,132,133,138,139,140,141,142,
    145,152,153,158,159,164,165,166,169,174,
    175,180,181,184,187,192,193,195,197,202,
    205,208,209,209,210,222,223,226,227,232,
    231,238,237,244,243,247,247,251,254,257,
    257,255,256,260,262,0 };
 
  if ((atnum < 1) || (atnum > 105)) {
    sprintf(errtxt,"Unknown atom number --> %d (mol#=%d)",atnum);
    dt_errorsave("smi2mol",DX_ERR_ERROR,errtxt);
    return (0);
  }
  return (mass[atnum]);
}

 
/********************************************************************
*** sameword - compares two strings ignoring case.
********************************************************************/
int sameword(char *p, char *q)
{
  while (*p||*q) {if (((*p++)-(*q++))%('a'-'A')) return 0;}
  return 1;
}
