
    /**********************************************************************
    * This source code is public domain, and may be freely distributed,   *
    * modified, and used for any purpose.  IT COMES WITH ABSOLUTELY NO    *
    * WARRANTY OF ANY KIND.                                               *
    **********************************************************************/

/*****************************************************************************
 *  PDB2TDT -- "Protein Data Bank (PDB) format" to TDT function
 *
 *  Note the function DY_PDBIN().
 *
 *  void dy_pdbinit()
 *  void dy_pdbfree(dt_Handle conf)
 *  dt_Handle dy_pdbin(FILE *fp, int newlabels)
 *
 *  BUG!    There is currently no way of differentiating between an invalid
 *          structure in the middle of a PDB file and the end of a PDB file.
 *          Should an extra EOF switch be added?
 *
 *  BUG!    There is currently no mechanism for calculating connectivity
 *          from ATOM records only.  No geometry-to-connectivity algorithm.
 *        
 *  NOTE!   DY_PDBIN does not check for atom-label uniqueness automatically.
 *          Input arg. NEWLABELS always needs to be set by calling program.
 *
 *  BUG!    RES fields are ignored.
 *
 *  NOTE!   Either COMPND or END records delimit molecule connection tables.
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "dt_smiles.h"
#include "dt_depict.h"
#include "du_utils.h"

#define MX_LINE 2048


/*** misc. functions. ***/


/*============================================================================
 *  lenstr() -- return number of characters before last white space
 */

static int lenstr(cs)
char *cs;
{
  char *p;

  p = cs + strlen(cs) - 1;
  while ( isspace(*p) && p>=cs ) p--;
  return (int) ( p<cs ? 0 : p-cs+1 );
}


/*============================================================================
 *  getline() -- reads at most NC characters from file FP into array S;
 *               does not include '\n' in string.
 */

static char *getline(s,fp,nc)
char *s;
FILE *fp;
int   nc;
{
  char *tmp,*tmp2;
  tmp=fgets(s,nc+1,fp);
  if (!(NULL == tmp)) {               /* assume it works! */
     tmp2    = strchr(s,'\n');        /* Assume all line len < MXREAD */
     tmp2[0] = '\0';
  }

              /*   while (!(tmp = strchr(s,'\n')))   Deal with > MXREADS later!
                   slen =+ MXREAD;
                   tmp  = (char *)malloc(slen);   
                   return((char *)0);               */

  return(tmp);
}


/*============================================================================
 * sym2num -- function returns atomic number from atomic symbol
 *            Returns -1 if symbol not recognized.
 */

static char *sym[] = { "*",  "H",  "He", "Li", "Be", "B",
                       "C",  "N",  "O",  "F",  "Ne", "Na",
                       "Mg", "Al", "Si", "P",  "S",  "Cl",
                       "Ar", "K",  "Ca", "Sc", "Ti", "V",
                       "Cr", "Mn", "Fe", "Co", "Ni", "Cu",
                       "Zn", "Ga", "Ge", "As", "Se", "Br",
                       "Kr", "Rb", "Sr", "Y",  "Zr", "Nb",
                       "Mo", "Tc", "Ru", "Rh", "Pd", "Ag",
                       "Cd", "In", "Sn", "Sb", "Te", "I",
                       "Xe", "Cs", "Ba", "La", "Ce", "Pr",
                       "Nd", "Pm", "Sm", "Eu", "Gd", "Tb",
                       "Dy", "Ho", "Er", "Tm", "Yb", "Lu",
                       "Hf", "Ta", "W",  "Re", "Os", "Ir",
                       "Pt", "Au", "Hg", "Tl", "Pb", "Bi",
                       "Po", "At", "Rn", "Fr", "Ra", "Ac",
                       "Th", "Pa", "U",  "Np", "Pu", "Am",
                       "Cm", "Bk", "Cf", "Es", "Fm"  };

static int sym2num(str)
char *str;
{ int i,n;

  if ( lenstr(str) == 1 ) 
     str[1] = '\0';
  if ( lenstr(str) == 2 ) {
     str[2] = '\0';
     if (isalpha(str[1]))
        str[1] = tolower(str[1]); }

  n = -1;
  for (i=0;i<101;i++)
     if (0 == strcmp(str,sym[i])) {
        n = i;
        break; }

  return (n);
}


/*============================================================================
 *  dbug() is useful for development.
 */

static void dbug(dt_Handle atoms, dt_Handle conf) 
{ dt_Handle atom, bonds, bond, batoms, batom;
  int       i, count, bcount, len;
  float     x, y, z;

  dt_reset(atoms);
  printf("\n");
  printf("ATOM index  ATOM number  ATOM label     X         Y          Z\n");
  printf("----------  -----------  ----------  --------  --------  --------\n");
  count = 0;
  while (NULL_OB != (atom = dt_next(atoms))) {
     count++;
     dt_getcoord(conf,atom,&x,&y,&z);
     printf("  %4d        %4d        %5s     %+7f  %+7f  %+7f\n",
            count, dt_number(atom), dt_stringvalue(&len,dt_adjunct(atom)),x,y,z);
  }
  printf("\n");
  printf("BOND index    ATOM 1       ATOM2      order   \n");
  printf("----------  -----------  ----------  -------- \n");
  bonds = dt_stream(dt_base(conf),TYP_BOND);
  bcount = 0;
  while (NULL_OB != (bond = dt_next(bonds))) {
     bcount++;
     batoms = dt_stream(bond,TYP_ATOM);
     count = 0;
     batom = dt_next(batoms);
     dt_reset(atoms);
     while (NULL_OB != (atom = dt_next(atoms))) {
        count++;
        if (atom == batom) 
            break;
     }           
     i = count;
     count = 0;
     batom = dt_next(batoms);
     dt_reset(atoms);
     while (NULL_OB != (atom = dt_next(atoms))) {
        count++;
        if (atom == batom) 
            break;
     }           
     printf("  %4d        %4d        %4d         %2d\n",
              bcount, i, count, dt_bondorder(bond));
  }           
}


/*============================================================================
 *  MAKE2CHARLABEL - Returns a left-justified, two-character string.
 *
 *  This string is an ad hoc 2-character representation of numbers 1 to 2802:
 *
 *     [1..2802] => [1,2,3..98,99,aa,ab,ac...,ay,az,aA,aB,...ZY,ZZ]
 */

static char *make2charlabel(int number)
{
  static char cmap[52] = {
     'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm',
     'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z',
     'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
     'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z'
  };

  char *str = "  ";

  if (100 > number) {
     sprintf(str,"%-2d",number);
  } else {
     number -= 99;
     str[0] = cmap[number/52];
     str[1] = cmap[number%52];
     str[2] = '\0';
  }

  return str;
}


/*============================================================================
 *  WRITE_D3D - Writes conformation to stdout as $D3D item.
 */

static void write_d3d(FILE *fp, dt_Handle conf, char *sid, char *rem)
{
   static int       pass0 = TRUE;
   static dt_Handle marker;
   dt_Handle        mol, uatoms, atoms, atom, xatom, bonds, bond;
   dt_Handle        *os, *oldsobs;
   float            fx, fy, fz;
   int              heavy, cnt = 0;

   if (pass0) {                                    /* Make static marker    */
      marker = dt_alloc_string(2, "X");            /* object on first pass  */
      pass0  = FALSE;
   }

   mol = dt_base(conf);                            /* Get mol from conf */

   if (NULL != sid && *sid)                        /* Start $D3D w/string id */
      fprintf(fp,"$D3D<0.0;%s;", sid);
   else
      fprintf(fp,"$D3D<0.0;;");
                                                   /* Save existing atom    */ 
                                                   /* adjuncts in oldsobs.  */
                                                   /* Count heavy atoms.    */
   os = oldsobs = (dt_Handle *)
                  malloc(dt_count(mol,TYP_ATOM) * sizeof(dt_Handle *));
   heavy = 0;
   atoms = dt_stream(mol, TYP_ATOM);
   while (NULL_OB != (atom = dt_next(atoms))) {
      *(os++) = dt_adjunct(atom);
      dt_setadjunct(atom, NULL_OB);
      if (1 < dt_number(atom)) heavy++;
   }
   dt_dealloc(atoms);
                                                   /* Add coordinates for   */
                                                   /* explicit atoms in 3rd */
                                                   /* field, and mark them. */
   uatoms = dt_alloc_seq();
   atoms  = dt_stream(mol, TYP_ATOM);
   while (heavy && NULL_OB != (atom = dt_next(atoms))) {
      if (1 < dt_number(atom)) {
         dt_getcoord(conf, atom, &fx, &fy, &fz);
         if (cnt++) fprintf(fp,",");
         fprintf(fp,"%.2f,%.2f,%.2f", fx, fy, fz);
         dt_setadjunct(atom, marker);
         dt_append(uatoms, atom);
         heavy--;
      }
   }
   dt_dealloc(atoms);

                                                   /* Add coord. for other */
                                                   /* atoms (implicit H's) */
   dt_reset(uatoms);
   while (NULL_OB != (atom = dt_next(uatoms))) {
      bonds = dt_stream(atom, TYP_BOND);
      while (NULL_OB != (bond = dt_next(bonds))) {
         xatom = dt_xatom(atom, bond);
         if (NULL_OB == dt_adjunct(xatom)) {
            dt_getcoord(conf, xatom, &fx, &fy, &fz);
            fprintf(fp,",%.2f,%.2f,%.2f", fx, fy, fz);
         }
      }
      dt_dealloc(bonds);
   }
   dt_dealloc(uatoms);
                                                   /* Write source; remark  */
                                                   /* if not null or empty. */
   if (NULL != rem && *rem)
      fprintf(fp,";%s;%s>", "PDB2TDT", rem);
   else
      fprintf(fp,";%s>", "PDB2TDT");
                                                   /* restore atom adjuncts */
   os = oldsobs;
   atoms = dt_stream(mol,TYP_ATOM);
   while (NULL_OB != (atom = dt_next(atoms))) {
      dt_setadjunct(atom, *(os++));
   }
   dt_dealloc(atoms);
   dy_free(oldsobs);
}


static  char  line[MX_LINE];



/*============================================================================
 *  DY_PDBINIT - Sets DY_PDBIN to read from a new file pointer.
 *
 *  Should also be used when position of a file pointer
 *  has been changed between successive DY_PDBIN calls.
 */

void dy_pdbinit()
{ 
  strncpy(line," ",MX_LINE);
}


/*============================================================================
 * DY_PDBFREE - Frees objects allocated by DY_PDBIN.
 *
 *  molecule, conformation, molecule str adjunct: name, atom str adjuncts
 */

void dy_pdbfree(dt_Handle conf)
{ dt_Handle m, atom, atoms, strob;

  if (TYP_CONFORMATION != dt_type(conf)) {           /* check conformation */
     fprintf(stderr,"DY_PDBFREE argument NOT of type conformation.\n");
     return;
  }

  if (NULL_OB == (m = dt_base(conf))) {              /* check molecule */
     dt_dealloc(conf);
     return;
  }

  if (NULL_OB != (strob = dt_adjunct(m)))            /* dealloc name */
     dt_dealloc(strob);

  if (NULL_OB != (atoms = dt_stream(m,TYP_ATOM))) {
     while (NULL_OB != (atom = dt_next(atoms))) {
        if (NULL_OB != (strob = dt_adjunct(atom)))   /* dealloc atom label */
           dt_dealloc(strob);
     }
     dt_dealloc(atoms);
  }

  dt_dealloc(conf);               
  dt_dealloc(m);
}


/*============================================================================
 *  DY_PDBIN - Reads PDB on stream fp and returns conformation.
 *
 *  If newlabels is true, unique atom labels are generated; else atom
 *  adjuncts are assumed to contain string objects with unique values.
 *
 *  Note: Caller is responsible for deallocating dy_pdbin created objects.
 *        This may be done by calling dy_pdbfree(dt_handle conf).
 */

dt_Handle dy_pdbin(FILE *fp, int newlabels)
{ dt_Handle  m, conf, atoms, atom, bond, fromatom, toatom, strob;
  int        from, to, lastto, symcount[107];
  int        i, j, k, slen, order, natoms, linecount, skip, atnum;
  char       symbol[5], str[6], *ps;
  float      x, y, z;

  natoms    = 0;
  linecount = 0;
  skip      = FALSE;
  if (newlabels)
     for (i=0;i < 106;i++)
        symcount[i] = 0;

  if (NULL_OB == (atoms = dt_alloc_seq())) {         /* sequence for atoms */
     fprintf(stderr,"Failed to allocate atom sequence.\n");
     return (NULL_OB);
  }

  if (NULL_OB == (m = dt_alloc_mol())) {             /* get molecule */
     fprintf(stderr,"Failed to allocate molecule.\n");
     dt_dealloc(atoms);
     return (NULL_OB);
  }

  if (NULL_OB == (conf = dt_alloc_conformation(m))){ /* get conformation */
     fprintf(stderr,"Failed to allocate conformation.\n");
     dt_dealloc(atoms);
     dt_dealloc(m);
     return (NULL_OB);
  }
  dt_mod_on(m);

  if (' ' != line[0]) {                           /* last call buffered name? */
     if (NULL_OB == dt_adjunct(m)) {              /* name->mol str adjunct */
        slen  = strlen(line) - 6;
        strob = dt_alloc_string(slen,line+6);
        dt_setadjunct(m,strob);
     }
  }

  while (NULL != (ps = getline(line,fp,MX_LINE))) {
     linecount++;
     if ((0 == strncmp(line,"END",3))) {
        dy_pdbinit();
        break;
     } else if ((0 == strncmp(line,"COMPND",6))) {
        if (NULL_OB == dt_adjunct(m)) {              /* name->mol str adjunct */
           slen  = strlen(line) - 6;
           strob = dt_alloc_string(slen,line+6);
           dt_setadjunct(m,strob);
        } else {
           break;
        }
     } else if ((0 == strncmp(line,"ATOM  ",6)) || 
                (0 == strncmp(line,"HETATM",6))) {   /* interpret atom record */

        if (' ' == *(line+12)) {
           symbol[0] =  *(line+13);
           symbol[1] =  '\0';
        } else {
           symbol[0] =  *(line+12);
           symbol[1] =  *(line+13);
           symbol[2] =  '\0';
        }
        if ( 0 > (atnum = sym2num(symbol))) {
           atnum = 0;
           fprintf(stderr,
                   "Failed to recognize atomic symbol for record #%d :\n%s\n",
                   linecount,line);
        }
        if (NULL_OB == (atom = dt_addatom(m,atnum,0))) {
           fprintf(stderr,"Failed to add atom for record #%d :\n%s\n",
                   linecount,line);
           skip = TRUE;
           continue;
        } 
        natoms++;
      
        if (newlabels) {                            /* make atom label or ... */
           symcount[dt_number(atom)]++;
           ps = dt_symbol(&i,atom);
           strncpy(symbol,ps,i);
           sprintf(symbol+i,"%2s",make2charlabel(symcount[dt_number(atom)]));
           slen = i + 1;    
        } else if (' ' == *(line+12)) {             /* ...get atom label */
           symbol[1] =  *(line+14);
           symbol[2] =  *(line+15);
           symbol[3] =  '\0';
           slen      = 4;
        } else {
           symbol[2] =  *(line+14);
           symbol[3] =  *(line+15);
           symbol[4] =  '\0';
           slen      = 5;
        }

        x = y = z = 0.0;                            /* scan x,y,z coord */
        if (3 != sscanf(line+30,"%f%f%f",&x,&y,&z)) {
           fprintf(stderr,
                   "Invalid x,y,z coordinates in CONECT record, line #%d:\n",
                    linecount,line); 
           skip = TRUE;
        } 
                                                    /* store x,y,z coord */
        if (FALSE == dt_setcoord(conf,atom,x,y,z)) {
           fprintf(stderr,"Failed to set coordinates for atom on line #%d:%s\n",
                          linecount, line);
           skip = TRUE;
           continue;
        }

        strob = dt_alloc_string(slen,symbol);       /*atom label->str adjunct*/
        dt_setadjunct(atom,strob);
                                                    /* add atom to sequence */
        if (NULL_OB == dt_append(atoms,atom)) {
           fprintf(stderr,"Failed to append to sequence.\n");
           skip = TRUE;
           continue;
        } 
     } else if (0 == strncmp(line,"CONECT",6)) {    /* interpret bond record */
                                                    /* make bonds */
        lastto = 0;
        for (i = 0;(i < MX_LINE) && ('\0' != line[i+6]);i++)
           ;
        if ( 5 > i ) {
           fprintf(stderr,
                   "CONECT record with no \"to\" specifications, line #%d:\n",
                   linecount,line);
           continue;
        } 
        strncpy(str,line+6,5);                      /* scan "from" */
        if (0 >= sscanf(str,"%d",&from)) {
           fprintf(stderr,
                 "Invalid \"from\" specification in CONECT record, line #%d:\n",
                 linecount,line);
           continue;
        } 
        if (from > natoms) {
           fprintf(stderr,
                 "CONECT record's \"from\" > total # atoms, line #%d:\n",
                 linecount,line);
           continue;
        } 

        j = (i / 5) - 1;
        dt_reset(atoms);                           /* get "from" dt_Handle */
        for (i = 0;i < from;i++)
            atom = dt_next(atoms);
        fromatom = atom;
                                                   /* loop over "to" specs. */
        for (i = 0; i < j; i++) {
           strncpy(str,line+6+((i+1)*5),5);        /* scan "to" */
           if (0 >= sscanf(str,"%d",&to)) {
        /*    fprintf(stderr,
                   "Invalid \"to\" specification in CONECT record, line #%d:\n",
                    linecount,line);  */
              continue;
           } 
           if (to > natoms) {
              fprintf(stderr,
                    "CONECT record's \"to\" > total # atoms, line #%d:\n",
                     linecount,line);
              continue;
           } 
           if (to == lastto) {                     /* increment bond order */
              bond = dt_bond(fromatom,toatom);
              order = dt_bondorder(bond);
              dt_setbondorder(bond,order+1);
           } else if (from >= to ) {               /* skip redundant info */
              continue; 
           } else {
              dt_reset(atoms);                     /* get "to" dt_Handle */
              for (k = 0;k < to;k++)
                  atom = dt_next(atoms);
              toatom = atom;
              lastto = to;
                                                   /* either                */
                                                   /*   inc. existing bond  */
                                                   /* or                    */
                                                   /*   add new bond        */
              if (NULL_OB == (bond = dt_bond(fromatom,toatom))) {
                 bond = dt_addbond(fromatom,toatom,DX_BTY_SINGLE);
                 dt_setbondorder(bond,1);
              } else {
                 order = dt_bondorder(bond);
                 dt_setbondorder(bond,order+1);
              }
           }
        }

        dt_toend(atoms);
     }
  }

  if (NULL == ps) {                               /* at EOF */
     dy_pdbinit();
     if (0 >= natoms) {
        dt_dealloc(atoms);
        dt_dealloc(conf);
        dt_dealloc(m);
        return NULL_OB;
     }
  }
                       /*******               dbug(atoms,conf);        *******/

  if (NULL_OB == dt_mod_off(m)) {
     fprintf(stderr, 
             "Failed to interpret PDB structure previous to line #%d:\n",
             linecount);
     dt_dealloc(atoms);
     dt_dealloc(conf);
     dt_dealloc(m);
     return NULL_OB;
  }

  dt_dealloc(atoms);


  return conf;                                    /* return conformation */

}

/*============================================================================
 *  main()
 */

main(int argc,char **argv)
{ FILE      *fpout, *fpin;
  int        i, badarg = FALSE;
  int        len;
  dt_Handle  conf;
                                                     /* Check command args */
  if (1 < argc) {
     badarg =  ((NULL == strstr(argv[1],".tdt")) &&
                (NULL == strstr(argv[1],".pdb")));
     if (!badarg && (2 < argc)) {
        badarg =  ((NULL == strstr(argv[2],".tdt")) &&
                   (NULL == strstr(argv[2],".pdb")));
        if (!badarg) 
           badarg = ((NULL == strstr(argv[1],".tdt")) &&
                     (NULL == strstr(argv[2],".tdt")));
        if (!badarg) 
           badarg = ((NULL == strstr(argv[1],".pdb")) &&
                     (NULL == strstr(argv[2],".pdb")));
     }
  }

  if ((3 < argc) || badarg) {
      fprintf(stderr,"USAGE:  pdb2tdt  [filename.pdb] [filename.tdt]\n\n");
      fprintf(stderr,"        If suffixed arguments are not present,\n");
      fprintf(stderr,"        then stdin  is assumed pdb-input and\n");
      fprintf(stderr,"             stdout is assumed tdt-output.\n");
      exit(0);
  }
                                                     /* Assign fileIO */
  fpin = stdin;
  fpout = stdout;
                                                     
  if (1 < argc) {
     if (NULL != strstr(argv[1],".tdt")) {
        if (NULL == (fpout = fopen(argv[1],"w"))) {
            fprintf(stderr,"ERROR! Can't open output file: %s\n",argv[1]);
            exit(0);
        }
     } else {
        if (NULL == (fpin = fopen(argv[1],"r"))) {
            fprintf(stderr,"ERROR! Can't open input file: %s\n",argv[1]);
            exit(0);
        }
     }
     if (2 < argc) {
        if (NULL != strstr(argv[2],".tdt")) {
           if (NULL == (fpout = fopen(argv[2],"w"))) {
               fprintf(stderr,"ERROR! Can't open output file: %s\n",argv[2]);
               exit(0);
           }
        } else {
           if (NULL == (fpin = fopen(argv[2],"r"))) {
               fprintf(stderr,"ERROR! Can't open input file: %s\n",argv[2]);
               exit(0);
           }
        }
     }
  }
                                                 /* read in pdbs, write tdts */
  dy_pdbinit();
  while (NULL_OB != (conf = dy_pdbin(fpin,TRUE))) {
     fprintf(fpout,"$SMI<%s>", dt_cansmiles(&len,dt_base(conf),FALSE));
     write_d3d(fpout,conf, dt_stringvalue(&len,dt_adjunct(dt_base(conf))), "");
     fprintf(fpout,"|\n");
     dy_pdbfree(conf);
  }

  fclose(fpout);
  exit(0);
}
