
    /**********************************************************************
    * This source code is public domain, and may be freely distributed,	  *
    * modified, and used for any purpose.  IT COMES WITH ABSOLUTELY NO    *
    * WARRANTY OF ANY KIND.						  *
    **********************************************************************/

/*****************************************************************************
 *  TDT2PDB -- SMILES to "Protein Data Bank (PDB) format" function
 *
 *   Note the function PDBOUT().
 *   int du_pdbout(FILE *fp, dt_Handle conf, int noh, int newlabels)
 *
 *****************************************************************************

/*
 * BUG!   A partial PDB may be written out on PDBOUT failure.
 *        Should strings be buffered and written out all at once only if
 *        the entire structure is successful?  (Not done currently...)
 *
 * BUG!   DU_PDBOUT does not generate atom labels when input argument
 *        NEWLABELS is false; all atoms are assumed to have unique labels.
 *        (i.e. either all atoms labels are supplied as string-object
 *        atom-adjuncts or all atom labels are generated.)
 *
 * BUG!   DU_PDBOUT does not check that supplied atom labels are unique.
 *
 * NOTE!  Only hydrogens with conformations get an atom record made.
 *        A switch is available to skip all hydrogens.
 * BUG?   Non-hydrogen atoms without conf. values are given 0.0 as coord.
 *
 * NOTE!  Numbers are concatenated to symbols to make unique symbol.
 *        Having separate counter per atom extends total labels possible.
 *
 * NOTE!  Atom symbols are right-justified on the 14th char. (PDB syntax)
 *
 * NOTE!  TER records without trailing integer parameters are used to
 *        delimit ATOM and CONECT records.
 *
 * NOTE!  REMARK, HEADER, SOURCE, AUTHOR, JOURNAL records are not used.
 *
 * NOTE!  A limit of 10 "to" atoms is imposed on CONECT records.
 *
 * NOTE!  "COMPND" record is used for NAME.
 * NOTE!  "REMARK" record is used for SMILES.
 *
 * NOTE!  HETATM convention: all atoms other than H,C,N,O,P,S.
 *
 * NOTE!  Residue name are all "RES"; part number used for res number.
 *
 * NOTE!  Multiple "END" (EOF) records are used in a single file to allow
 *        multiple structures per file.  (concatenated file method...)
 */

#include <stdio.h>
#include <string.h>
#include "dt_smiles.h"
#include "dt_depict.h"
#include "du_utils.h"


/*** functional definitions ***/
int        du_part      (/* dt_Handle ob  */);


/*============================================================================
 *  MAKE2CHARLABEL - Returns a left-justified, two-character string.
 *
 *  This string is an ad hoc 2-character representation of numbers 1 to 2802:
 *
 *     [1..2802] => [1,2,3..98,99,aa,ab,ac...,ay,az,aA,aB,...ZY,ZZ]
 */

static
char *make2charlabel(int number)
{
  static char cmap[52] = {
     'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm',
     'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z',
     'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
     'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z'
  };
  char *str = "  ";

  if (100 > number) {
     sprintf(str,"%-2d",number);
  } else {
     number -= 99;
     str[0] = cmap[number/52];
     str[1] = cmap[number%52];
     str[2] = '\0';
  }

  return str;
}

/*============================================================================
 *  addbond() - Used by du_pdbout to make bond list.
 *
 *  Returns TRUE if successful, else FALSE.
 */

static int addbond(int *pcount,
	           int *pto1, int *pto2, int *pto3, int *pto4, int *pto5,
                   int *pto6, int *pto7, int *pto8, int *pto9, int *pto10,
	           int ano)
{ 
  if ((1 > *pcount) || (10 < *pcount))
     return (FALSE);
  switch (*pcount) {
     case 1 :  *pto1  = ano;  break;
     case 2 :  *pto2  = ano;  break;
     case 3 :  *pto3  = ano;  break;
     case 4 :  *pto4  = ano;  break;
     case 5 :  *pto5  = ano;  break;
     case 6 :  *pto6  = ano;  break;
     case 7 :  *pto7  = ano;  break;
     case 8 :  *pto8  = ano;  break;
     case 9 :  *pto9  = ano;  break;
     case 10 : *pto10 = ano;  break;
  }
  (*pcount)++;
  return (TRUE);
}
 

/*============================================================================
 *  print_pdb() -- write conformation to stream fp in PDB format
 *
 *  Writes PDB connection table to FILE *fp for conformation conf.
 *  Input arg. NOH should be TRUE if no hydrogens are to be output.
 *  Input arg. NEWLABELS should be TRUE to generate atom labels.
 *  Returns TRUE if successful, else FALSE.
 */

int du_pdbout(FILE *fp, dt_Handle conf, int noh, int newlabels,
	      int lenn, char *name)
{
  dt_Handle  m, atoms, atom, bonds, bond, strob, adj;
  char      *smi, *ps;
  char       cto1[6], cto2[6], cto3[6], cto4[6], cto5[6];
  char       cto6[6], cto7[6], cto8[6], cto9[6], cto10[6];
  char       line[63];
  char       symbol[5];
  int        i, count, from, to1, to2, to3, to4, to5, to6, to7, to8, to9, to10;
  int        bo, len, symcount[107];
  float      x, y, z;

  /*** Get conformation, parent, smiles, or return. ***/

  if (TYP_CONFORMATION != dt_type(conf)                   )  return FALSE;
  if (NULL_OB          == (m   = dt_base(conf))           )  return FALSE;
  if (NULL_OB          == (smi = dt_cansmiles(&i,m,FALSE)))  return FALSE;


  if (NULL != name && 0 < lenn) {                    /* print name */
     fprintf(fp,"COMPND%.*s\n", lenn, name);
     fprintf(fp,"REMARK %.*s\n", i, smi);            /* print smiles */
  } else 
     fprintf(fp,"COMPND%.*s\n", i, smi);             /* print smiles */

  if (NULL_OB == (atoms = dt_stream(m,TYP_ATOM)))    /* get stream of atoms */
     return (FALSE);

  if (newlabels)
     for (i=0;i < 106;i++)                           /* zero lookup table */
        symcount[i] = 0;
  count = 0;                                         
  while ( NULL_OB != (atom = dt_next(atoms))) {      /* print out atoms */
     if (noh && (1 == dt_number(atom)))              /* skip hydrogens? */
	continue;
     count++;
     dt_setadjunct(atom,count);                      /* save atom index */
     if (!dt_getcoord(conf,atom,&x,&y,&z)) {
	x = y = z = 0.0;
        if (1 == dt_number(atom))                    /* skip H without conf */
	   continue;
     }

     if (newlabels) {
        symcount[dt_number(atom)]++;                 /* inc. lookup table  */
        ps = dt_symbol(&i,atom);                     /* PDB-justify symbol */
        if (1 == i) {                                /* build atom serial# */
	   symbol[0] = ' ';
	   symbol[1] = *ps;
        } else
          strcpy(symbol,ps);
        sprintf(symbol+2,"%2s",make2charlabel(symcount[dt_number(atom)]));
     } else {
	if (NULL_OB != (strob = dt_adjunct(atom))) {
           strncpy(symbol,dt_stringvalue(&len,strob),5);
	   for (i=len;i < 5;i++)
	      symbol[i] = ' ';
	} else {
           fprintf(stderr,"%s%s\n",
	           "Aborting PDB: Need string object ",
		   "for atom adjunct when NEWLABELS = FALSE.");
           dt_dealloc(atoms);
           return (FALSE);
        }
     }

     switch (dt_number(atom)) {                      /* CNOPSH are "ATOM"s */
        case 1  :
	case 6  :
	case 7  :
	case 8  :
	case 15 :
	case 16 : fprintf(fp,
	    "ATOM  %5d %4s RES  %4d    %8.3f%8.3f%8.3f                      \n",
	          count,symbol,du_part(atom)+1,x,y,z);
		  break;
        default : fprintf(fp,
	    "HETATM%5d %4s RES  %4d    %8.3f%8.3f%8.3f                      \n",
	          count,symbol,du_part(atom)+1,x,y,z);
     }
  }

  fprintf(fp,"TER   \n");                            /* print out delimiter  */

  if (NULL_OB == dt_reset(atoms)) {                  /* reset stream of atoms */
     dt_dealloc(atoms);
     return (FALSE);
  }

  while ( NULL_OB != (atom = dt_next(atoms))) {      /* for each atom...    */
     if (NULL_OB == (bonds = dt_stream(atom,TYP_BOND))) {
        dt_dealloc(atoms);
        return (FALSE);                              /* get stream of bonds */
     }
     from  = dt_adjunct(atom);
     to1   = to2 = to3 = to4 = to5 = to6 = to7 = to8 = to9 = to10 = 0;
     count = 1;
     while ( NULL_OB != (bond = dt_next(bonds))) {
	bo = dt_bondorder(bond);
	while (bo--) {
	   adj = dt_adjunct(dt_xatom(atom,bond));
	   if (!addbond(&count, &to1, &to2, &to3, &to4, &to5,
				&to6, &to7, &to8, &to9, &to10, adj)) {
              dt_dealloc(atoms);
              dt_dealloc(bonds);
              return (FALSE);
           }
	}
     }

     strcpy( cto1,  "     ");
     strcpy( cto2,  "     ");
     strcpy( cto3,  "     ");
     strcpy( cto4,  "     ");
     strcpy( cto5,  "     ");
     strcpy( cto6,  "     ");
     strcpy( cto7,  "     ");
     strcpy( cto8,  "     ");
     strcpy( cto9,  "     ");
     strcpy( cto10, "     ");

     if (0 == cto1 )
        break;
     else
        sprintf(cto1,"%5d",to1);
     if (0 != cto2 )
        sprintf(cto2,"%5d",to2);
     if (0 != cto3 )
        sprintf(cto3,"%5d",to3);
     if (0 != cto4 )
        sprintf(cto4,"%5d",to4);
     if (0 != cto5 )
        sprintf(cto5,"%5d",to5);
     if (0 != cto6 )
        sprintf(cto6,"%5d",to6);
     if (0 != cto7 )
        sprintf(cto7,"%5d",to7);
     if (0 != cto8 )
        sprintf(cto8,"%5d",to8);
     if (0 != cto9 )
        sprintf(cto9,"%5d",to9);
     if (0 != cto10 )
        sprintf(cto10,"%5d",to10);
                                                     /* print out bonds */

     sprintf(line, "CONECT%5d%5d%5d%5d%5d%5d%5d%5d%5d%5d%5d\n", from, 
	     to1, to2, to3, to4, to5, to6, to7, to8, to9, to10);
     fprintf(fp, "%.*s\n", 6 + (count*5), line);

     dt_dealloc(bonds);                              /* free stream of bonds */
  }

  dt_dealloc(atoms);                                 /* free stuff           */
  
  fprintf(fp,"END\n");                               /* print EOF delimiter  */

  return (TRUE);
}


/*============================================================================
 *  main() program
 */

main (int argc,char **argv)
{
  FILE      *fpout, *fpin;
  int        lenn, ok, i, badarg = FALSE;
  int        tdtlen, bufsize = 0;
  char      *name, *buf = NULL;
  dt_Handle  conf, confs, sos, sob;

  /*** Check command-line arguments. ***/

  if (1 < argc) {
     badarg =  ((NULL == strstr(argv[1],".tdt")) &&
                (NULL == strstr(argv[1],".pdb")));
     if (!badarg && (2 < argc)) {
        badarg =  ((NULL == strstr(argv[2],".tdt")) &&
                   (NULL == strstr(argv[2],".pdb")));
        if (!badarg) 
           badarg = ((NULL == strstr(argv[1],".tdt")) &&
                     (NULL == strstr(argv[2],".tdt")));
        if (!badarg) 
           badarg = ((NULL == strstr(argv[1],".pdb")) &&
                     (NULL == strstr(argv[2],".pdb")));
     }
  }

  /*** Squawk and exit on bad call. ****/

  if ((3 < argc) || badarg) {
      fprintf(stderr,"USAGE:  tdt2pdb  [filename.tdt] [filename.pdb]\n\n");
      fprintf(stderr,"        If suffixed arguments are not present,\n");
      fprintf(stderr,"        then stdin  is assumed tdt-input and\n");
      fprintf(stderr,"             stdout is assumed pdb-output.\n");
      exit(0);
  }

  /*** Assign files. ***/

  fpin  = stdin;
  fpout = stdout;
                                                     
  if (1 < argc) {
     if (NULL != strstr(argv[1],".tdt")) {
        if (NULL == (fpin = fopen(argv[1],"r"))) {
            fprintf(stderr,"ERROR! Can't open input file: %s\n",argv[1]);
            exit(0);
        }
     } else {
        if (NULL == (fpout = fopen(argv[1],"w"))) {
            fprintf(stderr,"ERROR! Can't open output file: %s\n",argv[1]);
            exit(0);
        }
     }
     if (2 < argc) {
        if (NULL != strstr(argv[2],".tdt")) {
           if (NULL == (fpin = fopen(argv[2],"r"))) {
               fprintf(stderr,"ERROR! Can't open input file: %s\n",argv[2]);
               exit(0);
           }
        } else {
           if (NULL == (fpout = fopen(argv[2],"w"))) {
               fprintf(stderr,"ERROR! Can't open output file: %s\n",argv[2]);
               exit(0);
           }
        }
     }
  }

  /*** Read input until end of file or error. ***/

  for (ok = i = 1; 1 == du_readtdt(fpin, &tdtlen, &buf, &bufsize); i++) {

     /*** Require a non-empty datatree. ***/

     if (1 > tdtlen) continue;

     /*** Loop over each conformation ($D3D). ***/

     confs = du_tdt2confseq(buf);

     while (NULL_OB != (conf = dt_next(confs))) {

	/*** Get the name and throw out other data. ***/

	sos = dt_adjunct(conf);
	dt_reset(sos);
	dt_dealloc(dt_next(sos));             /* skip hash number */
	sob  = dt_next(sos);
	name = dt_stringvalue(&lenn, sob);    /* save name */
	dt_dealloc(dt_next(sos));             /* skip source */
	dt_dealloc(dt_next(sos));             /* skip remark */
	dt_dealloc(sos);

	/*** Write out PDB. ***/

        if (!du_pdbout(fpout, conf, FALSE, TRUE, lenn, name)) {
           printf("--> Failed to write PDB for conformation %d\n", i);
	   ok = 0;
        }

	/*** Always deallocate molecule. ***/

        dt_dealloc(dt_base(conf));
     }

     /*** Deallocate now-empty sequence. ***/

     dt_dealloc(confs);

     /*** Warn on error. ***/

     if (!ok) 
        fprintf(stderr,"Failed to produce conformation for TDT:\n  %s\n", buf);
  }

  /*** Close output file and exit. **/

  fclose(fpout);
  exit(0);
}
