    /**********************************************************************
    * This source code is public domain, and may be freely distributed,	  *
    * modified, and used for any purpose.  IT COMES WITH ABSOLUTELY NO    *
    * WARRANTY OF ANY KIND.						  *
    **********************************************************************/

/*****************************************************************************
*  du_rosin() -- interpret Rosdal into molecule or depiction
*
*  This interprets the subset of Rosdal as distributed on the Beilstein tapes.
*  Presumably, this represents "Registration" and other elements used by S4.
*
*  The function du_rosin(ros,typ,isom) interprets a NULL-terminated Rosdal
*  string `ros' into an object of type `typ', which may be TYP_MOLECULE,
*  TYP_DEPICTION, or TYP_CONFORMATION.  Isomeric attributes will not be set
*  if `isom' is FALSE (0).  NULL_OB is returned on error.  Warnings and
*  errors are added to the dt_error utility.
*
*    NOTE:  This is a development version -- it only supports the subset of
*           Rosdal used in the Beilstein "Current Facts" product.
*
*  WARNING: Interpretation to a depiction is *not* supported here (yet).
*
*  WARNING: Interpretation to a conformation sets the coordinates to (x,y,0)
*           and dt_adjunct() of stereogenic double bonds set to themselves
*           and of chiral bonds set to the atom in "front".
*
*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include <dt_smiles.h>
#include <dt_depict.h>
#include <du_utils.h>
#include <du_smiles.h>


/*** Local limits, flags. ***/

#define DM_AINC  100
#define DM_BINC  120
#define DX_HUNK  222
#define DX_NA    122422053
#define SWAP3(a,b,t) { t = a; a = b; b = t; }

/*** Struct representing interpreted Rosdal string. ***/

typedef struct {
   int       an, ch, wt, val, hc, bc, fish, ix, iy, nimp;
   float     x, y;
   dt_Handle atom;
} ROSATOM;

typedef struct {
   int fr, to, bo;
   int ori;  /* (U:0) (A+:+1) (A-:-1) (B+:-2) (B-:+2) (S:3) else DX_NA*/
   dt_Handle bond;
} ROSBOND;

typedef struct {
   int      na, nb;        /* number of atoms & bonds, 0-origin */
   int      ma, mb;        /* alloc'ed array sizes */
   ROSATOM *a;             /* array of atoms */
   ROSBOND *b;             /* array of bonds */
} ROSMOL;

/*** Definition of this file. ***/

dt_Handle du_rosin(char *str, int reqtype, int isom);

/*** Local functions. ***/

static int      numbof(char **str);
static ROSBOND *new_rosbond(ROSMOL *rm);
static ROSATOM *get_rosatom(ROSMOL *rm, int iatom);
static int      atom_modifiers(char **pp, int *ch, int *fish, int *val,
					  int *wt, int *ix,   int *iy);
static int      bond_modifiers(char **pp, int *ori);
static int      setdbo(ROSMOL *rm, int db);
static int      setchi(ROSMOL *rm, int ix);
static void     rotate2d(double ang, int n, double *x, double *y);
static int      rosval(int an);

/*============================================================================
 *  du_rosin
 */

dt_Handle du_rosin(char *ros, int reqtype, int isom)
{
   FILE *fp;
   int  i, j, k, lenin;

   char             *p, *pp;
   int              an, wt, ch, hc, fr, to, bo, skip, stage, lastat, val;
   int              fish, ori, n2d, ix, iy, if2d;
   float            dx, dy, d, dmin;
   dt_Handle        mol, conf, atom, atoms;
   static ROSMOL    *rm = NULL;
   static ROSATOM   *ra, *a, *a1, *a2;
   static ROSBOND   *rb, *b;

   /*** Reject requests for depictions (not supported yet). ***/

   if (TYP_DEPICTION == reqtype) {
      dt_errorsave("du_rosin", DX_ERR_ERROR,
                   "du_rosin(,TYP_DEPICTION,) not implemented yet");
      return NULL_OB;
   }

   /*** Initialize data structure once only. ***/

   if (NULL == rm) {
      rm     = (ROSMOL  *) malloc(sizeof(ROSMOL));
      rm->ma = DM_AINC;
      rm->mb = DM_BINC;
      rm->a  = (ROSATOM *) malloc(rm->ma * sizeof(ROSATOM));
      rm->b  = (ROSBOND *) malloc(rm->mb * sizeof(ROSBOND));
   }

  /*
   *  Read raw rosdal data into local (ROSMOL) struct.
   */

   rm->na = 0;
   rm->nb = 0;
   an     = DX_NA;
   hc     = DX_HUNK;
   wt     = ch = val = fish = ori = ix = iy = DX_NA;
   fr     = to = bo = skip = stage = lastat = if2d = 0;

   /*** Loop over string, interpreting into local struct. ***/

   for (p = ros; '\0' != *p; p++) {

      /*** Set appropriate atom or bond modifiers. ***/

      if ('(' == *p) {
         if (0 != bo) {
            if (0 == bond_modifiers(&p, &ori) || ')' != *p) {
               dt_errorsave("du_rosin", DX_ERR_ERROR, "bad bond modifier");
               return NULL_OB;
            }
            p++;
         } else if (0 < to) {
            if (0 == atom_modifiers(&p, &ch, &fish, &val, &wt, &ix, &iy)
		|| ')' != *p) {
               dt_errorsave("du_rosin", DX_ERR_ERROR, "bad atom modifier");
               return NULL_OB;
            }
            p++;
         } else {
            dt_errorsave("du_rosin", DX_ERR_ERROR, "unexpected modifier");
            return NULL_OB;
         }
      }

      /*** switch on character value ***/

      switch (*p) {

      /*** bond orders ***/

      case '-': bo = 1; fr = to; to = 0; stage = 1; break;
      case '=': bo = 2; fr = to; to = 0; stage = 1; break;
      case '#': bo = 3; fr = to; to = 0; stage = 1; break;

      /*** Letters are atomic symbols; D and T are special. ***/

      case 'A': case 'B': case 'C': case 'D': case 'E': case 'F': case 'G':
      case 'H': case 'I': case 'J': case 'K': case 'L': case 'M': case 'N':
      case 'O': case 'P': case 'Q': case 'R': case 'S': case 'T': case 'U':
      case 'V': case 'W': case 'X': case 'Y': case 'Z':
       /***
       * if (stage) {
       *   dt_errorsave("du_rosin", DX_ERR_ERROR, "atom spec after bond spec");
       *   return NULL_OB;
       * }
       ***/
         an = numbof(&p);
         if (0 >= an || 105 <= an) {
            dt_errorsave("du_rosin", DX_ERR_ERROR, "unknown atom symbol");
            return NULL_OB;
         } else if (1 == an && 'H' != *p) {
            if ('D' == *p)
               wt = 2;
            else if ('T' == *p)
               wt = 3;
         }
         break;
 
      /*** digits are indexes ***/

      case '0': case '1': case '2': case '3': case '4':
      case '5': case '6': case '7': case '8': case '9': 
         to = 10 * to + *p - '0';
         break;

      /*** delimiters ***/

      case '.':
      case ',':
      case ';':    /* InfoChem has these for some reason. */
      case '/':
         if (0 != fr) {
            a1     = get_rosatom(rm, fr);
            a2     = get_rosatom(rm, to);
            b      = new_rosbond(rm);
            b->fr  = fr - 1;
            b->to  = to - 1;
            b->bo  = bo;
            b->ori = ori;
         } else if (0 != to) {
            if (to < lastat) {
               dt_errorsave("du_rosin", DX_ERR_ERROR, "atom spec out-of-order");
               return NULL_OB;
            }
            a       = get_rosatom(rm, to);

            if (DX_NA != an)    a->an   = an;
            if (DX_NA != wt)    a->wt   = wt;
            if (DX_NA != ch)    a->ch   = ch;
            if (DX_NA != hc)    a->hc   = hc;
            if (DX_NA != val)   a->val  = val;
            if (DX_NA != fish)  a->fish = fish;
            if (DX_NA != ix)    a->ix   = ix;
            if (DX_NA != iy)    a->iy   = iy;
            lastat = to;
         }
         an = DX_NA;
         hc = DX_HUNK;
         wt = ch = val = fish = ori = DX_NA;
         fr = to = bo = 0;
         break;
 
      default:
         dt_errorsave("du_rosin", DX_ERR_ERROR, "uninterpretable string");
         return NULL_OB;
      }

      /*** slash is treated as terminator ***/

      /* if ('/' == *p) break; */
      if ('.' == *p) break;
   }

  /*
   *  Interpret 2-D coordinates in "S4 HEX" format if needed and available.
   */

#ifdef NOPE
   if ((isom || TYP_CONFORMATION == reqtype)            /* need 2D data */
      &&           0 == strncmp("/+", p, 2)             /* 2D data present  */
      &&      rm->na == (n2d = strtol(p + 2, &pp, 10))  /* count agrees     */
      &&         '+' == *pp++                           /* + delimiter ok   */
      &&        NULL != strchr(pp, '.')) {              /* . delimeter ok   */
      for (i = 0, p = pp; i < n2d; i++, p += 6) {
         if (1 != sscanf(p,     "%3x", &ix)) break;
         rm->a[i].x = ix;
         if (1 != sscanf(p + 3, "%3x", &iy)) break;
         rm->a[i].y = iy;
      }
#endif

   if (isom || TYP_CONFORMATION == reqtype) {          /* need 2D data */
      n2d = 0;
      for (i = n2d = 0; i < rm->na; i++) {
         if (DX_NA != rm->a[i].ix && DX_NA != rm->a[i].iy) {
            rm->a[i].x = rm->a[i].ix;
            rm->a[i].y = rm->a[i].iy;
	    n2d++;
         }
      }

      if (n2d == rm->na) { 

      /*** Scale coordinates such that minimum bond distance is 1.0. ***/

      /* if ('.' == *p) { */

         for (dmin = 10000.0, i = 0; i < rm->nb; i++) {
            dx    = rm->a[rm->b[i].fr].x - rm->a[rm->b[i].to].x;
            dy    = rm->a[rm->b[i].fr].y - rm->a[rm->b[i].to].y;
            d     = sqrt(dx * dx + dy * dy);
            dmin  = DU_MIN(dmin, d);
         }
         for (i = 0; i < rm->na; i++) {
            rm->a[i].x /= dmin;
            rm->a[i].y /= dmin;
         }
         if2d = 1; /* TRUE */

   /*** Else, generate error or warning. ***/

      } else if (TYP_CONFORMATION == reqtype) {             /* no can do */
         dt_errorsave("du_rosin", DX_ERR_WARN, "2D data not available");
         return NULL_OB;
      } else if (isom) {
         dt_errorsave("du_rosin", DX_ERR_WARN,
                      "2D data not available for isomer resolution");
      }
   }

  /*
   *  Process raw data; the only trick (so far) is determining H-counts.
   *  Rosdal semantics appear to be pooly defined wrt H's and charges.
   *    "Normal valences" are not published, so what's the H-count?
   *    It appears that valence implies charges, can this be right?
   *    No, presumably valence and charge implies H's?
   *
   *  Get this ... the InfoChem version uses published valences!
   */

   /*** Set unspecified atomic numbers to 6 (C). ***/

   for (i = 0; i < rm->na; i++)
      if (DX_NA == rm->a[i].an) rm->a[i].an = DX_ATN_C;

   for (i = 0; i < rm->nb; i++) {                /* validate fr & to */
      if (0 > rm->b[i].fr || rm->na <= rm->b[i].fr) {
         dt_errorsave("du_rosin", DX_ERR_ERROR, "from-atom is out-of-bounds");
         return NULL_OB;
      } else if (0 > rm->b[i].to || rm->na <=  rm->b[i].to) {
         dt_errorsave("du_rosin", DX_ERR_ERROR, "to-atom is out-of-bounds");
         return NULL_OB;
      }
   }

   for (i = 0; i < rm->na; i++) {             /* set valence */
      if (DX_NA == rm->a[i].val)              /* implicit value unless set */
         rm->a[i].val = rosval(rm->a[i].an);
      if (DX_NA != rm->a[i].ch)               /* correct for charge */
         rm->a[i].val = DU_MAX(0, rm->a[i].val - DU_ABS(rm->a[i].ch));
   }

   for (i = 0; i < rm->na; i++)                  /* zero valence bond count */
      rm->a[i].bc = 0;

   for (i = 0; i < rm->nb; i++) {                /* count bond valences */
      rm->a[rm->b[i].fr].bc += rm->b[i].bo;
      rm->a[rm->b[i].to].bc += rm->b[i].bo;
   }

   for (i = 0; i < rm->na; i++)                  /* set hydrogen count */
      if (DX_HUNK == rm->a[i].hc)
         rm->a[i].hc = DU_MAX(0, rm->a[i].val - rm->a[i].bc);

  /*
   *  Create a molecule; allocate, add atoms, add bonds, and return it.
   */

   if (NULL_OB == (mol = dt_alloc_mol()) || 0 == dt_mod_on(mol)) {
      dt_errorsave("du_rosin", DX_ERR_ERROR,
                   "can't dt_alloc_mol() ... out of memory?");
      return NULL_OB;
   }

   /*** Add atoms to molecule; return in error if unable. ***/

   for (i = 0; i < rm->na; i++) {
      rm->a[i].atom = dt_addatom(mol, rm->a[i].an, rm->a[i].hc);
      if (NULL_OB == rm->a[i].atom) {
         dt_errorsave("du_rosin", DX_ERR_ERROR,
                      "can't dt_addatom() ... check syntax");
         dt_dealloc(mol);
         return NULL_OB;
      }
      if (isom && DX_NA != rm->a[i].wt)
         dt_setweight(rm->a[i].atom, rm->a[i].wt);
      if (DX_NA != rm->a[i].ch)
         dt_setcharge(rm->a[i].atom, rm->a[i].ch);
   }

   /*** Add bonds to molecule; return in error if unable. ***/

   for (i = 0; i < rm->nb; i++) {
      fr            = rm->b[i].fr;
      to            = rm->b[i].to;
      rm->b[i].bond = dt_addbond(rm->a[fr].atom, rm->a[to].atom, rm->b[i].bo);
      if (NULL_OB == rm->b[i].bond) {
         dt_errorsave("du_rosin", DX_ERR_ERROR,
                      "can't dt_addbond() ... check syntax");
         dt_dealloc(mol);
         return NULL_OB;
      }
   }

   /*** Deal with chiralities if 2D data is available. ***/

   if (if2d) {

      for (i = 0; i < rm->nb; i++)             /* Double bond orientations. */
         if (3 == rm->b[i].ori)
            if (0 == setdbo(rm, i))
               dt_errorsave("du_rosin", DX_ERR_WARN, "db orientation error");

      for (i = 0; i < rm->na; i++)            /* Other chiralities. */
         if (0 == rm->a[i].fish)
            if (0 == setchi(rm, i))
               dt_errorsave("du_rosin", DX_ERR_WARN, "chirality error");
   }

   /*** Analyze molecule; return in error if unable. ***/

   if (0 == dt_mod_off(mol)) {
      dt_errorsave("du_rosin", DX_ERR_ERROR, "resultant molecule is invalid");
      return NULL_OB;
   } else if (0 == dt_count(mol, TYP_ATOM)) {
      dt_errorsave("du_rosin", DX_ERR_ERROR, "no atoms in resultant molecule");
      return NULL_OB;
   }

   /*** Check for reasonableness. ***/

   atoms = dt_stream(mol, TYP_ATOM);
   k     = 0;
   while (NULL_OB != (atom = dt_next(atoms)))
      if (6 == dt_number(atom) && 0 == dt_count(atom, TYP_BOND)) k++;
   dt_dealloc(atoms);
   if (1 < k) {
      dt_errorsave("du_rosin", DX_ERR_ERROR,
                   "multiple disconnected C's - probable Rosdal error");
      return NULL_OB;
   }

   /*** Return appropriate object if able, else warn and return NULL_OB. ***/

   if (TYP_MOLECULE == reqtype) return mol;

   if (TYP_CONFORMATION == reqtype) {
      if (if2d && NULL_OB != (conf = dt_alloc_conformation(mol))) {
         for (a = rm->a, i = 0; i < rm->na; a++, i++)
            dt_setcoord(conf, a->atom, a->x, a->y, (float) 0.0);
         for (b = rm->b, i = 0; i < rm->nb; b++, i++)
            switch (b->ori) {
               case  1:  dt_setadjunct(b->bond, rm->a[b->to].atom);  break;
               case -1:  dt_setadjunct(b->bond, rm->a[b->fr].atom);  break;
               case  3:  dt_setadjunct(b->bond, b->bond);            break;
            }
         return conf;
      } else {
         dt_errorsave("du_rosin", DX_ERR_ERROR, "Can't create conformation");
         dt_dealloc(mol);
         return NULL_OB;
      }
   }

   dt_errorsave("du_rosin", DX_ERR_ERROR, "unknown request flag");
   return NULL_OB;
}

/*============================================================================
 *  atom_modifiers()
 *
 *  X,Y added
 *
 *  Three atom parameter kinds: sign charge | stereodescriptor | letter value
 *  The working rosdal file contains only: A+ A- B+ B- F0 S U+ U- V3 V4 V5 V6
 *  This responds to [-+FV][0-9].
 *  This returns charge, fisher, and valence (or DX_NA if not available),
 *  and leaves pointer on terminal ')'.
 *
 *  atom modifier mapping: [+|-]<n> .. charge <n>
 *                         F<n> ...... fish <n>
 *                         V<n> ...... val <n>
 *
 *  This only does one i.e. '(' letter digit ')'.
 *
 *  Return 0 on error only, 1 otherwise (success or not a modifier).
 *
 *
 * possible: +[1-4] -[1-9] A A+ A- B B+ B- F F0 F6 S U+ U- V[0-8] W[6-131]
 * responds to: +d -d F0 Vd Wddd
 *
 */

static int atom_modifiers(char **pp, int *ch, int *fish, int *val,
				     int *wt, int *ix, int *iy)
{
   char *p = *pp;
   char letter, digit;
   int  value;

   /*** Initialize. ***/

   *ch = *wt = *fish = *val = DX_NA;

   /*** Require starting '(', point to letter. ***/

   if ('(' != *p++) return 1;

   /*** Find ending ')', error if unable. ***/

   if (NULL == strchr(p, ')')) return 0;

   /*** Loop over fields ***/

   while (')' != *p) {

      letter = *p++;
      if (NULL == strchr("+-FUVWXY", letter)) return 0;

      digit  = *p++;
      if (!isdigit(digit)) return 0;
      value  = digit - '0';
      switch (letter) {
         case '+':  *ch  =  value;  break;
         case '-':  *ch  = -value;  break;
         case 'V':  *val =  value;  break;
         case 'F':  if (0 == value) *fish = 0;  break;
         case 'U':  break;   /* don't deal with unpaired electrons */
         case 'W':
            *wt = value;
            if (isdigit(*p)) *wt = *wt * 10 + *p++ - '0';
            if (isdigit(*p)) *wt = *wt * 10 + *p++ - '0';
            break;
         case 'X':
            *ix = value;
            if (isdigit(*p)) *ix = *ix * 10 + *p++ - '0';
            if (isdigit(*p)) *ix = *ix * 10 + *p++ - '0';
            break;
         case 'Y':
            *iy = value;
            if (isdigit(*p)) *iy = *iy * 10 + *p++ - '0';
            if (isdigit(*p)) *iy = *iy * 10 + *p++ - '0';
            break;
         default:
	    return 0;
      }

      if (',' == *p) p++;
   }

   /*** Advance to matching ')' and return happy. ***/

   if (')' != *p) return 0;

   *pp = p;
   return 1;
}

/*============================================================================
 *  bond_modifiers()
 *
 *  bond modifier mapping: (U:0) (A+:+1) (A-:-1) (B+:-2) (B-:+2) (S:3)  xx
 *
 *  +1 if "to" is in front of "from", -1 if vise versa
 *
 *  possible: +[1-4] -[1-9] A A+ A- B B+ B- F F0 F6 S U+ U- V[0-8] W[6-131]
 *  responds to: [ABU][+-]? S
 */

static int bond_modifiers(char **pp, int *ori)
{
   char *p = *pp;
   int   i, letter;

   *ori = DX_NA;                  /* initialize*/

   /*** Require starting '(', point to letter. ***/

   if ('(' != *p++) return 1;

   /*** Find ending ')', error if unable. ***/

   if (NULL == strchr(p, ')')) return 0;

   /*** Loop until ')' and switch. ***/

   while (')' != *p) {
      switch (*p++) {
         case 'A':  *ori =  1;  if ('+' == *p || '-' == *p) p++;  break;
         case 'B':  *ori = -1;  if ('+' == *p || '-' == *p) p++;  break;
         case 'U':  *ori =  0;  if ('+' == *p || '-' == *p) p++;  break;
         case 'S':  *ori =  3;  break;
      }
      if (')' != *p && ',' != *p) return 0;
      if (',' == *p) p++;
   }

   /*** REturn happy with matching ')' if able. ***/

   if (')' != *p) return 0;

   *pp = p;
   return 1;
}

/*============================================================================
 *  setdbo() -- set double bond orientation about bond number db
 *
 *  The method used to determine cis-trans bond relationships differs from
 *  that used in Grins because "extra" H-connections aren't present.
 */

#define PI     3.14159265358979323846
#define dpr(x) (x * 180. / PI)

static int setdbo(ROSMOL *rm, int db)
{
   int     i, ia, ja, iat, jat, ib, jb, cistrans;
   double  angle, x0, y0, x[4], y[4];

   /*** Double bonded atom indexes are iat, jat. ***/

   iat = rm->b[db].fr;
   jat = rm->b[db].to;

   /*** Find atoms ia, ja connected to iat, jat. ***/

   for (ib = -1, i = 0; i < rm->nb; i++)
      if (i != db) {
         if (iat == rm->b[i].fr) { ia = rm->b[i].to; ib = i; break; }
         if (iat == rm->b[i].to) { ia = rm->b[i].fr; ib = i; break; }
      }

   for (jb = -1,  i = 0; i < rm->nb; i++)
      if (i != db) {
         if (jat == rm->b[i].fr) { ja = rm->b[i].to; jb = i; break; }
         if (jat == rm->b[i].to) { ja = rm->b[i].fr; jb = i; break; }
      }

   if (-1 == ib || -1 == jb) {                  /* require both ends */
      dt_errorsave("setdbo", DX_ERR_ERROR, "didn't find opposite connections");
      return 0;
   }

   /*** Create four (x,y) points, translated to origin iat. ***/

   x0   = rm->a[iat].x;         y0   = rm->a[iat].y;
   x[0] = 0.0;                  y[0] = 0.0;
   x[1] = rm->a[jat].x - x0;    y[1] = rm->a[jat].y - y0;
   x[2] = rm->a[ia ].x - x0;    y[2] = rm->a[ia ].y - y0;
   x[3] = rm->a[ja ].x - x0;    y[3] = rm->a[ja ].y - y0;

   /*** Rotate jat to (x,0) direction. ***/

   angle = atan2(y[1], x[1]);
   rotate2d(-angle, 4, x, y);

   /*** ia, ja are cis if y[2] and y[3] are of the same sign. ***/

   cistrans = ((0 < y[2]) == (0 < y[3])) ? DX_CHI_CIS : DX_CHI_TRANS;

   dt_setdbo(rm->b[db].bond, rm->b[ib].bond, rm->b[jb].bond, cistrans);

   return 1;
}

/*============================================================================
 *  setchi() -- set chirality for atom number ix
 *
 *  find three or four connections
 */

static int setchi(ROSMOL *rm, int ix)
{
   ROSATOM   *a, *ax = &rm->a[ix];
   ROSBOND   *b;
   int        ib[4], i, j, k, ncon, chi, isori, endon, kori;
   int        flag, nori, nup, ndown;
   double     x[4], y[4], z[4], ang[4], angle, zdir, zori, tmp;
   dt_Handle  atom, bonds;

   double     xx[4], yy[4], zz[4];
  /*
   *  Find connections to central atom ix, load in arrays.
   *    x,y,z are ax-centered coordinates
   *    ib    is index of corresponding ROSBOND
   *    ncon  is number of connections
   *    nori,nup,ndown are number of total-, up-, and down-oriented bonds
   */

   for (ncon = nup = ndown = i = 0; i < rm->nb; i++) {
      if (ix == rm->b[i].fr | ix == rm->b[i].to) {
         if (4 <= ncon) {
            dt_errorsave("setchi", DX_ERR_ERROR,
                          "more than 4 bonds to chiral atom");
            return 0;
         }
         k         = (ix == rm->b[i].fr) ? rm->b[i].to : rm->b[i].fr;
         a         = &rm->a[k];
         b         = &rm->b[i];
         ib[ncon]  = i;
         x[ncon]   = a->x - ax->x;
         y[ncon]   = a->y - ax->y;
         kori      = (ix == rm->b[i].fr) ? b->ori : -b->ori;
         switch (kori) {
            case -1:  z[ncon] = zori = -1.0;  isori = ncon;  ndown++;  break;
            case  1:  z[ncon] = zori =  1.0;  isori = ncon;  nup++;    break;
            default:  z[ncon] =  0.0;  break;
         }
         ncon++;
      }
   }
   for (i = 0; i < ncon; i++) { xx[i] = x[i]; yy[i] = y[i]; zz[i] = z[i]; }

   /*** nori is total oriented bonds; flag is decimal <U><D> ***/

   nori = nup + ndown;
   flag = 10 * nup + ndown;

   /*** Require at least 3 connections and one orientation. ***/

   if (0 == nori) {
      dt_errorsave("setchi", DX_ERR_ERROR, "didn't find any oriented bonds");
      return 0;
   }

   if (3 > ncon) {
      dt_errorsave("setchi", DX_ERR_ERROR,
                    "fewer than three bonds to chiral center");
      return 0;
   }
   
   /*** A single oriented bond among 4 connections is end-on. ***/

   endon = -1;
   if (4 == ncon) {
      if (1 == nup + ndown) {
         endon    = isori;
         x[endon] = 0.0;
         y[endon] = 0.0;
      } else if (2 == flag || 20 == flag) {
         for (i = 0; i < 4; i++)
            if (0.0 == z[i]) z[i] = -z[isori];
      }

  /*
   *  Add hydrogen to 3-connected chiral atom:
   *     If 1/3 up or down, add an implicit H at it's (x,y,-z).
   *     If 2/3 or 3/3 all up or down, add end-on H at (0,0,-z).
   *  Implicit hydrogen is indicated by ib[] == -1.
   */

   } else if (3 == ncon) {
      switch (flag) {
      case  1: case 10:                                /* 1 up or down */
         x[3] = x[isori]; y[3] = y[isori];
         break;
      case  2: case 20: case  3: case 30:              /* 2,3 up or down */
         x[3] = 0.0; y[3] = 0.0; endon = 3;
         break;
      default:                                         /* can't do rest */
         dt_errorsave("setchi", DX_ERR_ERROR, "uninterpretable orientation");
         return 0;
      }
      z[3]   = -z[isori];
      ib[3]  = -1;
   }

   /*** Find angles from central atom, or -10.0 if superimposed. ***/

   for (i = 0; i < 4; i++)
      if      (0.0 != x[i]) ang[i] = atan2(y[i], x[i]);
      else if (0.0 != y[i]) ang[i] = atan2(y[i], x[i]); /* 3.14159 / 2.0; */
      else                  ang[i] = -10.0;

   /* if (-1 != endon) ang[endon] = -10.0;  /* superfluous ? */

   /*** Sort connections by angle; superimposed atoms will be first. ***/
 
   for (i = 0; i < 3; i++)
      for (j = i + 1; j < 4; j++)
         if (ang[i] > ang[j]) {
            SWAP3( ib[i],  ib[j], k  )
            SWAP3(  x[i],   x[j], tmp)
            SWAP3(  y[i],   y[j], tmp)
            SWAP3(  z[i],   z[j], tmp)
            SWAP3(ang[i], ang[j], tmp)
         }
 
   /*** If no atoms are end-on, make a Fischer projection. ***/

   if (-1 == endon) {

      /*** Set (x,y) locations to form a "+". ***/

      x[0] =  0.87;  y[0] =  0.00;
      x[1] =  0.00;  y[1] =  0.87;
      x[2] = -0.87;  y[2] =  0.00;
      x[3] =  0.00;  y[3] = -0.87;

      /*** Set tetrahedral Z-locations. ***/

      for (i = 0; i < 4; i++) {
         if      ( 0.1 < z[i]  ) z[i] =  0.5;
         else if (-0.1 > z[i]  ) z[i] = -0.5;
         else                    z[i] =  0.0;
      }

   /*** Else check for consistency and set end-on geometry. ***/

   } else if (-1 != endon && 0.0 == x[0] && 0.0 == y[0]) {

      /*** Set (x,y) locations to form a sideways "Y". ***/

      zori = copysign(1.0, z[0]);
      x[0] =  0.00;  y[0] =  0.00;  z[0] =  zori;
      x[1] =  0.87;  y[1] =  0.00;  z[1] =  0.5 * zori;
      x[2] = -0.43;  y[2] =  0.75;  z[2] =  0.5 * zori;
      x[3] = -0.43;  y[3] = -0.75;  z[3] =  0.5 * zori;

   }  else {
      dt_errorsave("setchi", DX_ERR_ERROR, "OUCH uninterpretable orientation");
      fprintf(stderr, "OUCH endon %d, nup %d, ndown %d, flag %d\n",
              endon, nup, ndown, flag);
      fprintf(stderr, "saxy: ");
      for (i = 0; i < 4; i++) {
         fprintf(stderr, " %d(%.2f:%.2f,%.2f)",
                 i, ang[i], x[i], y[i]);
      }
      fprintf(stderr, "\n");
   }

   /*** Rotate 3X ***/

   if (0.0 != y[0]) { angle = atan2(y[0], x[0]); rotate2d(-angle, 4, x, y); }
   if (0.0 != z[0]) { angle = atan2(z[0], x[0]); rotate2d(-angle, 4, x, z); }
   if (0.0 != z[1]) { angle = atan2(z[1], y[1]); rotate2d(-angle, 4, y, z); }

   /*** Check for inconsistencies. ***/

   if (z[3] == z[2]) {
      dt_errorsave("setchi", DX_ERR_ERROR, "chiral inconsistency, SP?");
      fprintf(stderr, "(setori) chiral inconsistency -- square planar?\n");
fprintf(stderr, "OUCH endon %d, nup %d, ndown %d, flag %d\n",
 endon, nup, ndown, flag);
fprintf(stderr, "ros:");
for (i = 0; i < ncon; i++) {
 fprintf(stderr, " [%.2f:%.2f,%.2f]", xx[i], yy[i], zz[i]);
}
fprintf(stderr, "\n");
fprintf(stderr, "xyz:");
for (i = 0; i < 4; i++) {
 fprintf(stderr, " (%.2f:%.2f,%.2f)", x[i], y[i], z[i]);
}
fprintf(stderr, "\n");
      return 0;
   } else if (0.1 > DU_ABS(z[3] - z[2])) {
      dt_errorsave("setchi", DX_ERR_ERROR, "chiral close call, SP?");
      fprintf(stderr, "(setori) chiral close call -- square planar?\n");
fprintf(stderr, "OUCH endon %d, nup %d, ndown %d, flag %d, ncon %d, dif %.5f\n",
 endon, nup, ndown, flag, ncon, DU_ABS(z[3] - z[2]));
fprintf(stderr, "ros:");
for (i = 0; i < ncon; i++) {
 fprintf(stderr, " [%.2f:%.2f,%.2f]", xx[i], yy[i], zz[i]);
}
fprintf(stderr, "\n");
fprintf(stderr, "xyz:");
for (i = 0; i < 4; i++) {
 fprintf(stderr, " (%.2f:%.2f,%.2f)", x[i], y[i], z[i]);
}
fprintf(stderr, "\n");
      return 0;
   }

   /*** Set chiral value. ***/

   atom  = ax->atom;
   bonds = dt_alloc_seq();
   for (i = 0; i < 4; i++)
      dt_append(bonds, (-1 == ib[i]) ? dt_isohydro() : rm->b[ib[i]].bond);
   dt_reset(bonds);

   /*** Looking down the X-axis, low z's are to the *right*. ***/

   chi  = (z[2] > z[3]) ? DX_CHI_THCCW : DX_CHI_THCW;
   dt_setchival(atom, bonds, chi);

   dt_dealloc(bonds);

   return 1;
}

/*============================================================================
 *  rotate2d() -- rotate n points (x,y) through angle a
 */
 
static void rotate2d(double ang, int n, double *x, double *y)
{
   int   i;
   double tx, ty;
   double sina = sin(ang);
   double cosa = cos(ang);
 
   for (i = 0; i < n; i++) {
      tx   = x[i];
      ty   = y[i];
      x[i] = tx * cosa - ty * sina;
      y[i] = ty * cosa + tx * sina;
   }
}

/*============================================================================
 *  reset() --
 */

static void reset(ROSMOL *rm)
{
   rm->na = rm->nb = 0;
}

/*============================================================================
 *  new_rosbond() -- return a freshly-initialized ROSBOND
 */

static ROSBOND *new_rosbond(ROSMOL *rm)
{
   int      i, nb;
   ROSBOND *b;

   if (rm->mb <= rm->nb) {
      rm->mb = rm->nb + DM_BINC;
      rm->b  = (ROSBOND *) realloc(rm->b, rm->mb * sizeof(ROSBOND));
   }

   b      = &rm->b[rm->nb];
   b->fr  = 0;
   b->to  = 0;
   b->bo  = 0;
   b->ori = DX_NA;

   rm->nb++;

   return b;
}

/*============================================================================
 *  get_rosatom() -- return a ROSATOM from a ROSMOL, creating it if needed
 *
 *  Note: input argument iatom is 1-origin
 */

static ROSATOM *get_rosatom(ROSMOL *rm, int iatom)
{
   ROSATOM *a;
   int      i;

   /*** Realloc rosatom array as needed. ***/

   if (rm->ma < iatom) {
      rm->ma = iatom + DM_AINC;
      rm->a  = (ROSATOM *) realloc(rm->a, rm->ma * sizeof(ROSATOM));
   }

   /*** Initialize new atoms as needed. ***/

   if (rm->na < iatom) {
      for (i = rm->na; i < iatom; i++) {
         a       = &rm->a[i];
         a->an   = DX_NA;
         a->ch   = DX_NA;
         a->wt   = DX_NA;
         a->val  = DX_NA;
         a->fish = DX_NA;
         a->hc   = DX_HUNK;
         a->bc   = 0;
         a->atom = NULL_OB;
      }
      rm->na = iatom;
   }

   /*** Return appropriate atom. ***/

   a = rm->a + iatom - 1;
   return a;
}

/*============================================================================
 *  rosval() --
 */

static int not_rosval(int an)
{
   switch (an) {
      case  1:  return 1;   /* H  */
      case  5:  return 3;   /* B  */
      case  6:  return 4;   /* C  */
      case  7:  return 3;   /* N  */
      case  8:  return 2;   /* O  */
      case 15:  return 3;   /* P  */
      case 16:  return 2;   /* S  */
      case  9:  return 1;   /* F  */
      case 17:  return 1;   /* Cl */
      case 35:  return 1;   /* Br */
      case 53:  return 1;   /* I  */
   }

   return 0;
}

/*============================================================================
 *  rosval() --  This is from the InfoChem documentation
 */

static int rosval(int an)
{
   static int stdval[106] = { 0,
			      1, 0, 1, 2, 3, 4, 3, 2, 1, 0,
                              1, 2, 3, 4, 3, 2, 1, 0, 1, 2,
                              3, 4, 5, 6, 7, 2, 2, 2, 2, 2,
                              3, 4, 3, 2, 1, 0, 1, 2, 3, 4,
                              5, 6, 7, 2, 2, 2, 1, 2, 3, 4,
                              3, 2, 1, 0, 1, 2, 3, 3, 3, 3,
                              3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
                              3, 4, 5, 6, 7, 2, 2, 2, 1, 2,
                              1, 4, 3, 2, 1, 0, 1, 2, 3, 4,
                              5, 6, 5, 4, 3, 3, 3, 3, 3, 3,
                              3, 3, 3, 0, 0 };

   return stdval[an];
}

/*============================================================================
 *  numbof() --
 */

static int numbof(char **s)
{
   int  rv = 0;
   char c  = **s;
   char cc = *(*s + 1);

   /*** Single letter symbols. ***/

   if (0 == islower(cc)) {
      switch (c) {
         case 'B':  rv =   5;  break;
         case 'C':  rv =   6;  break;
         case 'D':  rv =   1;  break;
         case 'F':  rv =   9;  break;
         case 'H':  rv =   1;  break;
         case 'I':  rv =  53;  break;
         case 'K':  rv =  19;  break;
         case 'N':  rv =   7;  break;
         case 'O':  rv =   8;  break;
         case 'P':  rv =  15;  break;
         case 'S':  rv =  16;  break;
         case 'T':  rv =   1;  break;
         case 'U':  rv =  92;  break;
         case 'V':  rv =  23;  break;
         case 'W':  rv =  74;  break;
         case 'Y':  rv =  39;  break;
      }

   /*** Two letter symbols. ***/

   } else {

      switch (c) {
      case 'A':
         switch (cc) {
            case 'c':  rv = 89;  break;
            case 'g':  rv = 47;  break;
            case 'l':  rv = 13;  break;
            case 'm':  rv = 95;  break;
            case 'r':  rv = 18;  break;
            case 's':  rv = 33;  break;
            case 't':  rv = 85;  break;
            case 'u':  rv = 79;  break;
         }
         break;
      case 'B':
         switch (cc) {
            case 'a': rv =  56;  break;
            case 'e': rv =   4;  break;
            case 'i': rv =  83;  break;
            case 'k': rv =  97;  break;
            case 'r': rv =  35;  break;
         }
         break;
      case 'C':
         switch (cc) {
            case 'a': rv =  20;  break;
            case 'd': rv =  48;  break;
            case 'e': rv =  58;  break;
            case 'f': rv =  98;  break;
            case 'l': rv =  17;  break;
            case 'm': rv =  96;  break;
            case 'o': rv =  27;  break;
            case 'r': rv =  24;  break;
            case 's': rv =  55;  break;
            case 'u': rv =  29;  break;
         }
         break;
      case 'D':
         switch (cc) {
            case 'y': rv =  66;  break;
         }
         break;
      case 'E':
         switch (cc) {
            case 'r': rv =  68;  break;
            case 's': rv =  99;  break;
            case 'u': rv =  63;  break;
         }
         break;
      case 'F':
         switch (cc) {
            case 'e': rv =  26;  break;
            case 'm': rv = 100;  break;
            case 'r': rv =  87;  break;
         }
         break;
      case 'G':
         switch (cc) {
            case 'a': rv =  31;  break;
            case 'd': rv =  64;  break;
            case 'e': rv =  32;  break;
         }
         break;
      case 'H':
         switch (cc) {
            case 'a': rv = 105;  break;
            case 'e': rv =   2;  break;
            case 'f': rv =  72;  break;
            case 'g': rv =  80;  break;
            case 'o': rv =  67;  break;
         }
         break;
      case 'I':
         switch (cc) {
            case 'n': rv =  49;  break;
            case 'r': rv =  77;  break;
         }
         break;
      case 'K':
         switch (cc) {
            case 'r': rv =  36;  break;
         }
         break;
      case 'L':
         switch (cc) {
            case 'a': rv =  57;  break;
            case 'i': rv =   3;  break;
            case 'r': rv = 103;  break;
            case 'u': rv =  71;  break;
         }
         break;
      case 'M':
         switch (cc) {
            case 'd': rv = 101;  break;
            case 'g': rv =  12;  break;
            case 'n': rv =  25;  break;
            case 'o': rv =  42;  break;
         }
         break;
      case 'N':
         switch (cc) {
            case 'a': rv =  11;  break;
            case 'b': rv =  41;  break;
            case 'd': rv =  60;  break;
            case 'e': rv =  10;  break;
            case 'i': rv =  28;  break;
            case 'o': rv = 102;  break;
            case 'p': rv =  93;  break;
         }
         break;
      case 'O':
         switch (cc) {
            case 's': rv =  76;  break;
         }
         break;
      case 'P':
         switch (cc) {
            case 'a': rv =  91;  break;
            case 'b': rv =  82;  break;
            case 'd': rv =  46;  break;
            case 'm': rv =  61;  break;
            case 'o': rv =  84;  break;
            case 'r': rv =  59;  break;
            case 't': rv =  78;  break;
            case 'u': rv =  94;  break;
         }
         break;
      case 'R':
         switch (cc) {
            case 'a': rv =  88;  break;
            case 'b': rv =  37;  break;
            case 'e': rv =  75;  break;
            case 'f': rv = 104;  break;
            case 'h': rv =  45;  break;
            case 'n': rv =  86;  break;
            case 'u': rv =  44;  break;
         }
         break;
      case 'S':
         switch (cc) {
            case 'b': rv =  51;  break;
            case 'c': rv =  21;  break;
            case 'e': rv =  34;  break;
            case 'i': rv =  14;  break;
            case 'm': rv =  62;  break;
            case 'n': rv =  50;  break;
            case 'r': rv =  38;  break;
         }
         break;
      case 'T':
         switch (cc) {
            case 'a': rv =  73;  break;
            case 'b': rv =  65;  break;
            case 'c': rv =  43;  break;
            case 'e': rv =  52;  break;
            case 'h': rv =  90;  break;
            case 'i': rv =  22;  break;
            case 'l': rv =  81;  break;
            case 'm': rv =  69;  break;
         }
         break;
      case 'X':
         switch (cc) {
            case 'e': rv =  54;  break;
         }
         break;
      case 'Y':
         switch (cc) {
            case 'b': rv =  70;  break;
         }
         break;
      case 'Z':
         switch (cc) {
            case 'n': rv =  30;  break;
            case 'r': rv =  40;  break;
         }
         break;
      }
      if (0 != rv) (*s)++;
   }

   return rv;
}
