    /**********************************************************************
    * This source code is public domain, and may be freely distributed,	  *
    * modified, and used for any purpose.  IT COMES WITH ABSOLUTELY NO    *
    * WARRANTY OF ANY KIND.						  *
    **********************************************************************/

/*======================================================================
| FILE:		ros2tdt.c
| DESCRIPTION:
|	Converts ROSDAL to SMILES.  Output is in TDT format.
|	
| Copyright (c) 1993, Daylight Chemical Information Systems, Inc.
+======================================================================*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dt_smiles.h>
#include <dt_depict.h>
#include <du_utils.h>
#include <du_smiles.h>

/*** Buffer size.  The longest Rosdal in 1991 "Current Facts" is 4323. ***/

#define DM_BUF   8000

static char *     fgets_sentence(char *str, int mx, FILE *fp);
static int        brnscan(char *str);
static int        snoscan(char *str);
extern dt_Handle  du_rosin(char *str, int reqtype, int booiso);

/*============================================================================
 *  ros2tdt main program
 */

main(int ac, char **av)
{
   FILE      *fp;
   char      str[DM_BUF], *p, *smi, *ism, *usm;
   int       sno, brn, lens, lenu, i;
   float     x, y, z;
   int       smifmt = 0, booiso = 1;
   dt_Handle mol, atom, atoms, dd;

/*** work around SGI "feature": can vastly improve performance ***/

#ifdef sgi
#include <malloc.h>
   mallopt(M_MXCHK, 100);
#endif

   /*** Read command line params; open input file or die trying. ***/

   if (2 > ac || 4 < ac) {
      fprintf(stderr, "Syntax: $ ros2tdt [-s] [-u] file\n");
      exit(1);
   }
   for (i = 1; i < ac; i++) {
      if (0 == strcmp("-s", av[i])) {
         smifmt = 1;
      } else if (0 == strcmp("-u", av[i])) {
         booiso = 0;
      } else if (NULL == (fp = fopen(av[i], "r"))) {
         fprintf(stderr, "Can't open file %s (ros2tdt)\n", av[1]);
         exit(2);
      }
   }

   /*** Read one rosdal per input line. ***/

   while (NULL != fgets_sentence(str, DM_BUF, fp)) {

      /*** Require SNO specified as a Rosdal comment. ***/

      if (0 == (sno = snoscan(str))) continue;

      /*** Skip to end of preface; p points to start of Rosdal string. ***/

      p = str;
      if ('(' != *p) continue;
      while (*p)
         if (')' == *p++) break;
      if (!*p) continue;

      /*** Interpret Rosdal string into molecule, reject if unable. ***/

      if (NULL_OB == (mol = du_rosin(p, TYP_MOLECULE, booiso))) {
         dt_errorsave("ros2tdt", DX_ERR_ERROR, "can't interpret Rosdal:");
         dt_errorsave("ros2tdt", DX_ERR_ERROR, str);
         du_printerrors(stderr, DX_ERR_NOTE);

      /*** If "-s", write $SMI<SMILES>\n$SNO<SNO>\n| to standard output. ***/

      } else {
         du_printerrors(stderr, DX_ERR_NOTE);
         smi = dt_cansmiles(&lens, mol, booiso);
         /* printf("$SMI<%.*s>\n$SNO<%d>\n|\n", lens, smi, sno); */
	 if (smifmt) {
            printf("%.*s %d-%d\n", lens, smi, sno/1000, sno%1000);
	 } else if (0 == booiso) {
            printf("$SMI<%.*s>\n$SNO<%d-%d>\n|\n",
		    lens, smi, sno/1000, sno%1000);
	 } else {
	    ism = du_strndup(smi, lens);
            usm = dt_cansmiles(&lenu, mol, 0);
            printf("$SMI<%.*s>\n$SNO<%d-%d>\n",
		   lenu, usm, sno/1000, sno%1000);
            if (lens != lenu || 0 != strncmp(usm, ism, lens))
               printf("$ISM<%.*s>\nSNO<%d-%d>\n",
		      lens, smi, sno/1000, sno%1000);
            printf("|\n");
	    du_free(ism);
	 }
      }

      /*** Clear errors and molecule (and all dependent objects). ***/

      dt_errorclear();
      dt_dealloc(mol);
   }
}


/*============================================================================
 *  brnscan() -- interpret number in form "BRN=number")
 */

static int brnscan(char *str)
{
   int   brn = 0;
   char *p   = strstr(str, "\"BRN=");
   char *pp  = strstr(str, "\")");

   /*** Require expression in form "BRN=.*") ***/

   if (NULL == p || NULL == pp | pp < p + 5) return 0;

   /*** Return BRN interpreted as integer ***/

   return (int) atoi(p + 5);
}

/*============================================================================
 *  snoscan() -- interpret number in form "SNO=number")
 */

static int snoscan(char *str)
{
   int   nin, s1 = 0, s2 = 0, sno = 0;
   char *p   = strstr(str, "\"SNO=");
   char *pp  = strstr(str, ")");

   /*** Require expression in form "SNO=dddddddd-ddd.*") ***/

   if (NULL == p || NULL == pp | pp < p + 5) return 0;

   /*** Return SNO interpreted as integer ***/

   nin = sscanf(p + 5, "%d-%d", &s1, &s2);
   if (2 != nin) return 0;

   /*** Return SNO as integer ***/

   sno = 1000 * s1 + s2;
   return sno;
}

/*============================================================================
 *  fgets_sentence() -- like fgets, but read '.'-terminated input line
 */

static char *fgets_sentence(char *str, int mx, FILE *fp)
{
   char *rv, *pp, *p = str;
   int   lens, len;

   while (NULL != (rv = fgets(p, mx, fp))) {
      len = lens = strlen(p) - 1;
      for (pp = p + lens - 1; pp > p && ' ' == *pp; pp--)
	 len--;
      if ('.' == *pp) lens = len;
      p   += lens;
      mx  -= lens;
      if ('.' == *pp) break;
   }
   return rv;
}
