/*************************************************************************
*** syb2smi.c -- Convert Sybyl MOL2 format to SMILES/TDT
***
***   The following Sybyl data is converted:
***       @<TRIPOS>MOLECULE
***       @<TRIPOS>ATOM       including 3d or 2d and atom-types
***       @<TRIPOS>BOND
***       
*** Author: Jeremy Yang
*** Contributions and chemistry by: John Bradshaw
*** Rev:  11 Dec 2001
**************************************************************************
*** Ref: (1) Tripos MOL2 File Format, 10/00.
*************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include "dt_smiles.h"
#include "du_utils.h"
#include "du_3d.h"

#define MAXLINE 100      /*** max mol2 line length ***/
#define D3D_SOURCE "n/a"
#define MAJOR_REPORT 1000
#define MINOR_REPORT 25

int sameword(char *p,char *q)
{while (*p||*q) {if (tolower(*p++)!=tolower(*q++)) return 0;} return 1;}

void help()
{
  fprintf(stderr,
  "\t************************************************************\n"
  "\t| syb2smi - Convert Sybyl/MOL2 file to SMILES/TDTs         |\n"
  "\t************************************************************\n"
  "\t|                                                          |\n"
  "\t| syb2smi [options] <infile.syb [>outfile.smi|tdt]         |\n"
  "\t|                                                          |\n"
  "\t| options:                                                 |\n"
  "\t|  -output_format SMI        SMILES output                 |\n"
  "\t|  -output_format ISM        Isomeric SMILES output        |\n"
  "\t|  -output_format TDT        TDT output (w/ ISM+D3D)       |\n"
  "\t|  -guess_charges NEVER      default                       |\n"
  "\t|  -guess_charges ALWAYS     fill valences                 |\n"
  "\t|  -guess_charges ONFAIL     only to overcome failure      |\n"
  "\t|  -v                        report warnings               |\n"
  "\t|  -vv                       report notes and warnings     |\n"
  "\t|  -help                     This help                     |\n"
  "\t|                                                          |\n"
  "\t************************************************************\n"
  "\t|                   Daylight CIS, Inc.                     |\n"
  "\t|             Toolkit Demonstration Program                |\n"
  "\t************************************************************\n");
  exit(1);
}

/*******************************************************
*** Struct unifying atom-ntuple information
*** (Note: not so sure about the string lengths!)
*******************************************************/
typedef struct {
  dt_Handle h;                    /*** Handle of atom object ***/
  float x;                        /*** x coordinate          ***/
  float y;                        /*** y coordinate          ***/
  float z;                        /*** z coordinate          ***/
  float pcharge;                  /*** Sybyl partial charge  ***/
  char type[6];                   /*** Sybyl atom type       ***/
  char name[10];                  /*** Sybyl atom name       ***/
  char subname[10];               /*** Sybyl substr name     ***/
  char stat_bits[81];             /*** Sybyl status flags    ***/
  int subst_id;                   /*** Sybyl substr id       ***/
  int visited;                    /*** internal utility flag ***/
} ATOMINFO;

/*******************************************************
*** Function prototypes
*******************************************************/
int       atom_typer   (char *atom_type);
int       bond_typer   (char *bond_type);
char     *d3d          (dt_Handle mol, int num_atoms);
char     *d2d          (dt_Handle mol, int num_atoms);
char     *atypes       (dt_Handle mol, int num_atoms);
char     *isosmi       (dt_Handle mol, int *len, int wegot_3d);
int       setchiral3   (dt_Handle mol);
int       setcistrans  (dt_Handle mol);
float     signed_volume(dt_Handle atom, dt_Handle bondstream);
dt_Handle stream2seq   (dt_Handle stream);
dt_Handle infer_charges   (dt_Handle mol,float pcharge_total);
int du_vbond(dt_Handle atom,int *nbond);

/*******************************************************
*** Global variables
*******************************************************/
char *name=NULL;            /*** Name of molecule               ***/
int ct_in = 0;              /*** N connection tables read       ***/
ATOMINFO *atoms = NULL;     /*** Atom info array (dynamic)      ***/
char errtxt[500];           /*** error buffer                   ***/

main(int argc, char *argv[])
{
  dt_Handle atom;             /*** atom object                    ***/
  dt_Handle atom2;            /*** 2nd atom object                ***/
  dt_Handle atomstream;       /*** atom stream                    ***/
  dt_Handle bond;             /*** bond object                    ***/
  dt_Handle bondstream;       /*** bond stream                    ***/
  dt_Handle mol,mol2;         /*** molecule                       ***/
  char *line=NULL;            /*** ASCII line from MOL-file       ***/
  char *cansmi=NULL;          /*** canonical smiles               ***/
  char *canismi=NULL;         /*** canonical isomeric smiles      ***/
  char *p;                    /*** generic (char *)               ***/
  char *mol_type=NULL;        /*** Sybyl molecule type            ***/
  char *charge_type=NULL;     /*** Sybyl mol charge type          ***/
  char *status_bits=NULL;     /*** Sybyl mol flags                ***/
  char *mol_comment=NULL;     /*** Sybyl mol comment              ***/
  char dict_type[MAXLINE];    /*** Sybyl dictionary type          ***/
  char dict_name[MAXLINE];    /*** Sybyl dictionary name          ***/
  char subst_name[MAXLINE];   /*** Sybyl substr name              ***/
  char subst_chain[MAXLINE];  /*** chain of substructure          ***/
  char subst_type[MAXLINE];   /*** substructure type              ***/
  char sub_type[MAXLINE];     /*** subtype of subst chain         ***/
  char subst_status[MAXLINE]; /*** subst status flags             ***/
  char subst_comment[MAXLINE]; /*** substructure comment          ***/
  char bond_type[5];          /*** bond type                      ***/
  int subst_id;               /*** substructure id                ***/
  int root_atom;              /*** substructure root atom id      ***/
  int inter_bonds;            /*** substructure # inter bonds     ***/
  int subst_dict_type;        /*** substructure dict type         ***/
  int atom_from;              /*** bond-from atom-index           ***/
  int atom_to;                /*** bond-to atom-index             ***/
  int i,j,k,l;                /*** generic indices                ***/
  int charge;                 /*** formal charge                  ***/
  int lensmi;                 /*** cansmi len                     ***/
  int lenismi;                /*** isosmiles len                  ***/
  int smi_out=0;              /*** n smiles written               ***/
  int tdts_out=0;             /*** n TDTs written                 ***/
  int write_ism=0;            /*** boolean: Iso-SMILES output ?   ***/
  int write_smi=1;            /*** boolean: SMILES output ?       ***/
  int write_tdt=0;            /*** boolean: TDT output ?          ***/
  int num_bonds;              /*** # bonds                        ***/
  int num_subst;              /*** # substructures                ***/
  int num_feat;               /*** # features                     ***/
  int num_sets;               /*** # sets                         ***/
  int num_atoms;              /*** # atoms                        ***/
  int atom_id;                /*** atom id #                      ***/
  int bond_id;                /*** bond id #                      ***/
  int err_count=0;            /*** error count                    ***/
  int wegot_2d;               /*** 2D flag                        ***/
  int wegot_3d;               /*** 3D flag                        ***/
  int atnum;                  /*** atomic number                  ***/
  int btype;                  /*** bond type number               ***/
  int max_num_atoms=0;        /*** max number of atoms            ***/
  int max_smilen=0;           /*** max cansmi len                 ***/
  int max_ismilen=0;          /*** max icansmi len                ***/
  int errlev=DX_ERR_ERROR;    /*** error level shown              ***/
  int guess_charges=0;        /*** 0=never,1=always,2=onfail      ***/
  float pcharge_total=0.0;    /*** total of partial charges       ***/

  /*****************************************
  ***  Parse command line.
  *****************************************/
  for (++argv; --argc; ++argv) {
    if (sameword("-help",*argv)) { help(); }
    else if (sameword("-t",*argv)) {
      write_smi=0; write_ism=0; write_tdt=1;
    } else if (sameword("-i",*argv)) {
      write_smi=0; write_ism=1; write_tdt=0;
    } else if (sameword(*argv,"-output_format")) {
      --argc; ++argv;
      if (sameword(*argv,"SMI")) {
        write_smi=1; write_ism=0; write_tdt=0;
      } else if (sameword(*argv,"ISM")) {
        write_smi=0; write_ism=1; write_tdt=0;
      } else if (sameword(*argv,"TDT")) {
        write_smi=0; write_ism=0; write_tdt=1;
      } else {
        fprintf(stderr, "ERROR (SYB2SMI): Unknown format: \"%s\"\n", *argv);
        help();
      }
    } else if (sameword(*argv,"-guess_charges")) {
      --argc; ++argv;
      if (sameword(*argv,"NEVER")) {
        guess_charges = 0;
      } else if (sameword(*argv,"ALWAYS")) {
        guess_charges = 1;
      } else if (sameword(*argv,"ONFAIL")) {
        guess_charges = 2;
      } else {
        fprintf(stderr, "ERROR (SYB2SMI): Unknown arg: \"%s\"\n", *argv);
        help();
      }
    } else if (sameword(*argv,"-v")) {
      errlev=DX_ERR_WARN;
    } else if (sameword(*argv,"-vv")) {
      errlev=DX_ERR_NOTE;
    } else {
      printf("ERROR (SYB2SMI): Unknown option \"%s\"\n", *argv); 
      help();
    }
  }

  if (!(line=du_fgetline(&l,stdin))) {
    fprintf(stderr, "ERROR (SYB2SMI): Unexpected EOF... ");
    ++err_count;
    goto byebye;
  }

  /********************************************************
  *** MOLECULE LOOP
  ********************************************************/
  while (1) {
    num_atoms = num_bonds = num_subst = num_feat = num_sets = 0;
    wegot_3d = wegot_2d = 0;
    status_bits = mol_comment = NULL;
    pcharge_total = 0.0;

    if (!line) {
      fprintf(stderr, "NOTE (SYB2SMI): EOF; normal exit...\n");
      goto byebye;
    }

    /************************************************************
    *** First comes the MOLECULE RTI (Record Type Indicator)
    ************************************************************/
    while (strncmp(line,"@<TRIPOS>MOLECULE",17)) {
      if (!(line=du_fgetline(&l,stdin))) {
        fprintf(stderr,"ERROR (SYB2SMI): can't find MOLECULE RTI...\n");
        goto byebye;
      }
    }

    name = du_strdup(du_fgetline(&l,stdin));
    line = du_fgetline(&l,stdin);

    sscanf(line,"%d%d%d%d%d",&num_atoms,&num_bonds,&num_subst,
                               &num_feat,&num_sets);

    /***********************************************************
    *** (Re)allocate ATOMINFO array of structs
    ***********************************************************/
    if (!atoms) {
      atoms = (ATOMINFO *)malloc((num_atoms+1)*sizeof(ATOMINFO));
      max_num_atoms = num_atoms;
    } else if (num_atoms>max_num_atoms) {
      atoms = (ATOMINFO *)realloc(atoms,(num_atoms+1)*sizeof(ATOMINFO));
      max_num_atoms = num_atoms;
    }

    /***********************************************************
    *** Allocate mol.
    ***********************************************************/
    mol = dt_alloc_mol();
    dt_mod_on(mol);

    mol_type = du_strdup(du_fgetline(&l,stdin));
    charge_type = du_strdup(du_fgetline(&l,stdin));

    /***********************************************************
    *** STATUS_BITS and MOL_COMMENT may, apparently, be absent
    *** in Insight files.
    ***********************************************************/
    line = du_fgetline(&l,stdin);
    if ('@' != line[0]) {
      status_bits = du_strdup(line);
      line = du_fgetline(&l,stdin);
      if ('@' != line[0]) {
        mol_comment = du_strdup(line);
        if (!(line=du_fgetline(&l,stdin))) {
          ++err_count;
          fprintf(stderr, "ERROR (SYB2SMI): Unexpected EOF... ");
          goto byebye;
        }
      }
    }

    /***********************************************************
    *** Now comes other RTIs which may occur IN ANY ORDER.
    ***********************************************************/
    while (line != strstr(line,"@<TRIPOS>MOLECULE")) {
      if (line == strstr(line,"@<TRIPOS>ALT_TYPE")) {
        do {if (!(line=du_fgetline(&l,stdin))) goto mkmol;} while ('@'!=*line);
      } else if (line==strstr(line,"@<TRIPOS>ANCHOR_ATOM")) {
        do {if (!(line=du_fgetline(&l,stdin))) goto mkmol;} while ('@'!=*line);
      } else if (line==strstr(line,"@<TRIPOS>ASSOCIATED_ANNOTATION")) {
        do {if (!(line=du_fgetline(&l,stdin))) goto mkmol;} while ('@'!=*line);

      /*****************************************
      *** Atoms!
      *****************************************/
      } else if (line == strstr(line,"@<TRIPOS>ATOM")) {
        for (i = 1; i <= num_atoms; ++i) {
          atoms[i].subst_id = 0;
          *(atoms[i].subname) = '\0';
          atoms[i].pcharge = 0.00;
          *(atoms[i].stat_bits) = '\0';
          line=du_fgetline(&l,stdin);
 
          sscanf(line, "%d%s%f%f%f%s%d%s%f%s",
                 &atom_id, atoms[i].name,
                 &atoms[i].x, &atoms[i].y, &atoms[i].z,
                 atoms[i].type,
                 &atoms[i].subst_id, atoms[i].subname, &atoms[i].pcharge,
                 atoms[i].stat_bits);

          if (0 > (atnum = atom_typer(atoms[i].type))) {
            sprintf(errtxt,"Unknown atom type \"%s\"; name=\"%s\",mol#=%d",
              atoms[i].type,name,ct_in+1);
            dt_errorsave("syb2smi",DX_ERR_ERROR,errtxt);
            ++err_count;
            dt_dealloc(mol);
            goto nextct;
          } else if (999 == atnum) {    /*** Don't add LPs ***/
            atoms[i].h = NULL_OB;  ;
          } else {
            atoms[i].h = dt_addatom(mol, atnum, 0);
            pcharge_total += atoms[i].pcharge;
            dt_setadjunct(atoms[i].h, (dt_Handle) i);
            wegot_3d += (int) fabs(atoms[i].z);
            wegot_2d += (int) (fabs(atoms[i].x) + fabs(atoms[i].y));
          }
        }
        if (wegot_3d) wegot_2d = 0;
        if (!(line=du_fgetline(&l,stdin))) goto mkmol;

      /*****************************************
      *** Bonds!
      *****************************************/
      } else if (line == strstr(line,"@<TRIPOS>BOND")) {
        for (i = 1; i <= num_bonds; ++i) {
          line=du_fgetline(&l,stdin);
          sscanf(line, "%d%d%d%s",
                 &bond_id, &atom_from, &atom_to, bond_type);
          if (0 > (btype = bond_typer(bond_type))) {
            sprintf(errtxt,"Unknown bond type \"%s\"; name=\"%s\",mol#=%d",
              bond_type,name,ct_in+1);
            dt_errorsave("syb2smi",DX_ERR_ERROR,errtxt);
            ++err_count;
            dt_dealloc(mol);
            goto nextct;

          /*** Do nothing for "Not-connected" ***/
          } else if (0 == btype) {
            sprintf(errtxt,"\"Not-connected\" bond ignored.");
            dt_errorsave("syb2smi",DX_ERR_NOTE,errtxt);
          }
          /*** Check that atoms exist (e.g., not LPs) ***/
          else if ((NULL_OB!=atoms[atom_from].h) &&
                   (NULL_OB!=atoms[atom_to].h))
            dt_addbond(atoms[atom_from].h, atoms[atom_to].h, btype);
        }
        if (!(line=du_fgetline(&l,stdin))) goto mkmol;
      } else if (line == strstr(line,"@<TRIPOS>CENTER_OF_MASS")) {
        do {if (!(line=du_fgetline(&l,stdin))) goto mkmol;} while ('@'!=*line);
      } else if (line == strstr(line,"@<TRIPOS>CENTROID")) {
        do {if (!(line=du_fgetline(&l,stdin))) goto mkmol;} while ('@'!=*line);
      } else if (line == strstr(line,"@<TRIPOS>COMMENT")) {
        do {if (!(line=du_fgetline(&l,stdin))) goto mkmol;} while ('@'!=*line);
      } else if (line == strstr(line,"@<TRIPOS>CRYSIN")) {
        do {if (!(line=du_fgetline(&l,stdin))) goto mkmol;} while ('@'!=*line);
      } else if (line == strstr(line,"@<TRIPOS>CURR_POS")) {
        do {if (!(line=du_fgetline(&l,stdin))) goto mkmol;} while ('@'!=*line);
      } else if (line == strstr(line,"@<TRIPOS>DATA_FILE")) {
        do {if (!(line=du_fgetline(&l,stdin))) goto mkmol;} while ('@'!=*line);

      /*** Dictionary Type/Name ***/
      } else if (line == strstr(line,"@<TRIPOS>DICT")) {
        if (!(line=du_fgetline(&l,stdin))) goto mkmol;
        sscanf(line, "%s%s", dict_type, dict_name); 
        while ('@'!=*line) if (!(line=du_fgetline(&l,stdin))) goto mkmol;
      } else if (line == strstr(line,"@<TRIPOS>EXTENSION_POINT")) {
        do {if (!(line=du_fgetline(&l,stdin))) goto mkmol;} while ('@'!=*line);

      /*** Periodic Boundary Conditions ***/
      } else if (line == strstr(line,"@<TRIPOS>FF_PBC")) {
        do {if (!(line=du_fgetline(&l,stdin))) goto mkmol;} while ('@'!=*line);
      } else if (line == strstr(line,"@<TRIPOS>FFCON_ANGLE")) {
        do {if (!(line=du_fgetline(&l,stdin))) goto mkmol;} while ('@'!=*line);
      } else if (line == strstr(line,"@<TRIPOS>FFCON_DIST")) {
        do {if (!(line=du_fgetline(&l,stdin))) goto mkmol;} while ('@'!=*line);
      } else if (line == strstr(line,"@<TRIPOS>FFCON_RANGE")) {
        do {if (!(line=du_fgetline(&l,stdin))) goto mkmol;} while ('@'!=*line);
      } else if (line == strstr(line,"@<TRIPOS>FFCON_TORSION")) {
        do {if (!(line=du_fgetline(&l,stdin))) goto mkmol;} while ('@'!=*line);
      } else if (line == strstr(line,"@<TRIPOS>LINE")) {
        do {if (!(line=du_fgetline(&l,stdin))) goto mkmol;} while ('@'!=*line);
      } else if (line == strstr(line,"@<TRIPOS>LSPLANE")) {
        do {if (!(line=du_fgetline(&l,stdin))) goto mkmol;} while ('@'!=*line);
      } else if (line == strstr(line,"@<TRIPOS>NORMAL")) {
        do {if (!(line=du_fgetline(&l,stdin))) goto mkmol;} while ('@'!=*line);
      } else if (line == strstr(line,"@<TRIPOS>POLYBUILD_HIST")) {
        do {if (!(line=du_fgetline(&l,stdin))) goto mkmol;} while ('@'!=*line);
      } else if (line == strstr(line,"@<TRIPOS>QSAR_ALIGN_RULEQSAR")) {
        do {if (!(line=du_fgetline(&l,stdin))) goto mkmol;} while ('@'!=*line);
      } else if (line == strstr(line,"@<TRIPOS>RING_CLOSURE")) {
        do {if (!(line=du_fgetline(&l,stdin))) goto mkmol;} while ('@'!=*line);
      } else if (line == strstr(line,"@<TRIPOS>ROTATABLE_BOND")) {
        do {if (!(line=du_fgetline(&l,stdin))) goto mkmol;} while ('@'!=*line);
      } else if (line == strstr(line,"@<TRIPOS>SEARCH_DIST")) {
        do {if (!(line=du_fgetline(&l,stdin))) goto mkmol;} while ('@'!=*line);
      } else if (line == strstr(line,"@<TRIPOS>SEARCH_OPTIONS")) {
        do {if (!(line=du_fgetline(&l,stdin))) goto mkmol;} while ('@'!=*line);
      } else if (line == strstr(line,"@<TRIPOS>SET")) {
        do {if (!(line=du_fgetline(&l,stdin))) goto mkmol;} while ('@'!=*line);

      /*** Substructure ***/
      } else if (line == strstr(line,"@<TRIPOS>SUBSTRUCTURE")) {
        if (!(line=du_fgetline(&l,stdin))) goto mkmol;
        while ('@'!=*line) {
          sscanf(line, "%d%s%d%s%d%s%s%d%s%s",
               &subst_id, subst_name, &root_atom, subst_type, &subst_dict_type,
               subst_chain, sub_type, &inter_bonds, subst_status,
               subst_comment); 
          if (!(line=du_fgetline(&l,stdin))) goto mkmol;
        }
      } else if (line == strstr(line,"@<TRIPOS>U_FEAT")) {
        do {if (!(line=du_fgetline(&l,stdin))) goto mkmol; } while ('@'!=*line);
      } else if (NULL == strtok(line," \t")) { /*** whitespace line ***/
        do {if (!(line=du_fgetline(&l,stdin))) goto mkmol; } while ('@'!=*line);
      } else if (*line == '#') {
        sprintf(errtxt,"COMMENT LINE: \"%s\"",line);
        dt_errorsave("syb2smi",DX_ERR_NOTE,errtxt);
        while ('@'!=*line) if (!(line=du_fgetline(&l,stdin))) goto mkmol;
      } else if (line == strstr(line,"@<TRIPOS>")) {
        sprintf(errtxt,"Unknown RTI: \"%s\"",line);
        dt_errorsave("syb2smi",DX_ERR_WARN,errtxt);
        do {if (!(line=du_fgetline(&l,stdin))) goto mkmol; } while ('@'!=*line);
      } else {
        sprintf(errtxt,"Unknown format: \"%s\"",line);
        dt_errorsave("syb2smi",DX_ERR_WARN,errtxt);
        if (!(line=du_fgetline(&l,stdin))) goto mkmol;
      }
    }

mkmol:
    /********************************************************************
    ***  Adjust charges if requested.
    ********************************************************************/
    switch (guess_charges) {
      case 0:
        break;
      case 1:
        mol = infer_charges(mol,pcharge_total);
        break;
      case 2:
        mol2 = dt_copy(mol);
        if (!dt_mod_off(mol2)) infer_charges(mol,pcharge_total);
        else dt_dealloc(mol2);
        break;
      default: /*** error ***/
        break;
    }

    /********************************************************************
    ***  Interpret molecule.
    ********************************************************************/
    if (!dt_mod_off(mol)) {
      sprintf(errtxt,"dt_mod_off failure; name=\"%s\",mol#=%d",name,ct_in+1);
      dt_errorsave("syb2smi",DX_ERR_ERROR,errtxt);
      ++err_count;
      goto nextct;
    }
    p = dt_cansmiles(&lensmi,mol,0);
    if (lensmi <= 0) {
      sprintf(errtxt,"dt_cansmiles failure; name=\"%s\",mol#=%d",name,ct_in+1);
      dt_errorsave("syb2smi",DX_ERR_ERROR,errtxt);
      ++err_count;
      goto nextct;
    }
    if (!cansmi) {
      cansmi = (char *)malloc((lensmi+1)*sizeof(char));
      max_smilen = lensmi;
    }
    else if (lensmi>max_smilen) {
      cansmi = (char *)realloc(cansmi,(lensmi+1)*sizeof(char));
      max_smilen = lensmi;
    }
    strncpy(cansmi,p,lensmi); cansmi[lensmi] = '\0';
    ++smi_out;
    if (write_smi) {
      printf("%s %s\n",cansmi,name);
    }
    if (write_ism || write_tdt) {
      p = isosmi(mol, &lenismi, wegot_3d);
      if (lenismi <= 0) {
        sprintf(errtxt,"dt_cansmiles (isomeric) failure; name=\"%s\",mol#=%d",
          name,ct_in+1);
        dt_errorsave("syb2smi",DX_ERR_ERROR,errtxt);
        ++err_count;
      } else {
        if (!canismi) {
          canismi = (char *)malloc((lenismi+1)*sizeof(char));
          max_ismilen = lenismi;
        } else if (lenismi>max_ismilen) {
          canismi = (char *)realloc(canismi,(lenismi+1)*sizeof(char));
          max_ismilen = lenismi;
        }
        strncpy(canismi,p,lenismi); canismi[lenismi] = '\0';
      }
    }
    if (write_ism) {
      printf("%s %s\n",lenismi>0?canismi:cansmi,name);
    }
    if (write_tdt) {
      printf("$SMI<%s>\n",cansmi);
      printf("ATYP<%s>\n",atypes(mol, num_atoms));
      printf("NSUB<%d>\n",num_subst);
      printf("NFEA<%d>\n",num_feat);
      printf("NSET<%d>\n",num_sets);
      if (mol_type&&*mol_type) printf("MOL_TYPE<%s>\n", mol_type);
      if (charge_type&&*charge_type) printf("CHG_TYPE<%s>\n", charge_type);
      if (status_bits&&*status_bits) printf("STATUS_BITS<%s>\n", status_bits);
      if (mol_comment&&*mol_comment) printf("COM<%s>\n", mol_comment);
      if (name&&*name) printf("$NAM<%s>\n", name);
      if (wegot_3d) 
        printf("$D3D<;%s;%s;%s;>\n",name,d3d(mol,num_atoms), D3D_SOURCE);
      else if (wegot_2d) 
        printf("2D<%s;>\n",d2d(mol,num_atoms));
      if (lenismi>0 && strcmp(cansmi,canismi)) printf("ISM<%s>\n",canismi);
      printf("|\n");
      ++tdts_out;
    }
    dt_dealloc(mol);

nextct:
    du_printerrors(stderr,errlev);
    dt_errorclear();

   /******************************************************
    *** Progress reports...
    ******************************************************/
    if (MINOR_REPORT&&!((++ct_in)%MINOR_REPORT))
      { fprintf(stderr, "."); fflush(stderr); }
    if (MAJOR_REPORT&&!((ct_in)%MAJOR_REPORT))
      fprintf(stderr," %d MOL2s, %d errors\n",ct_in,err_count);

    free(name); free(mol_type); free(charge_type); free(status_bits);
    free(mol_comment);

  } /*** End of molecule loop ***/

byebye:
  du_printerrors(stderr,errlev);
  dt_errorclear();
  fprintf(stderr,"RESULTS (SYB2SMI):  SYB's in:   %8d\n", ct_in);
  if (write_tdt)
    fprintf(stderr,"RESULTS (SYB2SMI):  TDTs out:   %8d\n", tdts_out);
  fprintf(stderr,"RESULTS (SYB2SMI):  SMILES out: %8d\n", smi_out);
  fprintf(stderr,"RESULTS (SYB2SMI):  Errors:     %8d\n", err_count);

  free(cansmi); free(canismi); free(atoms);

  fprintf(stderr, "So long, baby!\n");
  exit (0);
}

/********************************************************************
*** atom_typer - return atomic number
********************************************************************/
int atom_typer(char *atom_type)
{
  char ch1, ch2;
  int atnum;

  static char *symbol_table[] = {
    "*",
    "H",  "He", "Li", "Be", "B",  "C",  "N",  "O",  "F",  "Ne", /*   1-10 */
    "Na", "Mg", "Al", "Si", "P",  "S",  "Cl", "Ar", "K",  "Ca", /*  11-20 */
    "Sc", "Ti", "V",  "Cr", "Mn", "Fe", "Co", "Ni", "Cu", "Zn", /*  21-30 */
    "Ga", "Ge", "As", "Se", "Br", "Kr", "Rb", "Sr", "Y",  "Zr", /*  31-40 */
    "Nb", "Mo", "Tc", "Ru", "Rh", "Pd", "Ag", "Cd", "In", "Sn", /*  41-50 */
    "Sb", "Te", "I",  "Xe", "Cs", "Ba", "La", "Ce", "Pr", "Nd", /*  51-60 */
    "Pm", "Sm", "Eu", "Gd", "Tb", "Dy", "Ho", "Er", "Tm", "Yb", /*  61-70 */
    "Lu", "Hf", "Ta", "W",  "Re", "Os", "Ir", "Pt", "Au", "Hg", /*  71-80 */
    "Tl", "Pb", "Bi", "Po", "At", "Rn", "Fr", "Ra", "Ac", "Th", /*  81-90 */
    "Pa", "U",  "Np", "Pu", "Am", "Cm", "Bk", "Cf", "Es", "Fm", /*  91-100 */
    "Md", "No", "Lr", "Rf", "Ha", "*" };                        /* 101-105 */

   ch1 = atom_type[0];
   if( ch1!='\0' ) {
       ch2 = tolower(atom_type[1]);
       if( ch2 == '.' ) ch2 = '\0';

       for( atnum=0; atnum<106; atnum++ )
           if( (symbol_table[atnum][0]==ch1) &&
               (symbol_table[atnum][1]==ch2) )
               return atnum;

       if( (ch1=='L') && (ch2=='p') ) return 999;  /* Lone Pair */
       if( (ch1=='D') && (ch2=='u') ) return 0;    /* Dummy */
   }
   return -1;  /* Not Found */
}

/********************************************************************
*** bond_typer - return bond type
********************************************************************/
int bond_typer(char *bond_type)
{
  int btype;

       if (!strcmp(bond_type,"1"))  btype = 1;
  else if (!strcmp(bond_type,"2"))  btype = 2;
  else if (!strcmp(bond_type,"3"))  btype = 3;
  else if (!strcmp(bond_type,"ar")) btype = 4;  /*** Aromatic      ***/
  else if (!strcmp(bond_type,"am")) btype = 1;  /*** Amide         ***/
  else if (!strcmp(bond_type,"du")) btype = 1;  /*** Dummy         ***/
  else if (!strcmp(bond_type,"un")) btype = 1;  /*** Unknown       ***/
  else if (!strcmp(bond_type,"nc")) btype = 0;  /*** Not connected ***/
  else                              btype = -1;
  return(btype);
}

/**********************************************************************
*** d3d - returns string containing XYZ's for use in $D3D
***         datatype.  It is assumed that ALL H atoms are explicitly
***         defined and the x/y/z arrays contain their coordinates.
***
*** ALSO, I'm still not sure... are xyz's for explicit H's
*** in sequence with other explicit heavy atom xyz's? or at the
*** end with the implicit H xyz's???  The latter is implemented.
**********************************************************************/
char *d3d(dt_Handle mol, int num_atoms)
{
  dt_Handle atomstream, atom, bondstream, bond, atom2;
  int       acount, hcount_total, i, bufsize, bytes;
  char      *p;
  static char *buf3d = NULL;
  static int max_num_atoms = 0;

  for(i = 1; i < num_atoms+1; ++i) atoms[i].visited = 0;

  atomstream = dt_canatom_stream(mol, 0, 0);
  if (NULL_OB == atomstream) return ("");
 
  bufsize = 24 * num_atoms + 1;

  if (NULL == buf3d) buf3d = (char *) malloc(bufsize);
  else               buf3d = (char *) realloc(buf3d, bufsize);

  /***********************************************************
  ***  1st write xyz's for explict atoms in cansmi order
  ***********************************************************/
  for (p = buf3d; NULL_OB!=(atom = dt_next(atomstream)); p += bytes) {
    i = (int) dt_adjunct(atom);
    bytes = sprintf(p, "%.2f,%.2f,%.2f,", atoms[i].x, atoms[i].y, atoms[i].z);
    atoms[i].visited = 1;
  }

  /***********************************************************
  ***  Now for the H's -- since we have ALL coordinates
  ***********************************************************/
  dt_reset(atomstream);
  while (NULL_OB!=(atom = dt_next(atomstream))) {
    bondstream = dt_stream(atom, TYP_BOND);
    while (NULL_OB!=(bond = dt_next(bondstream))) {
      atom2 = dt_xatom(atom, bond);
      if (DX_ATN_H == dt_number(atom2)) {
        i = (int) dt_adjunct(atom2);

        /**********************************************************
        ***  If (visited) then H is explicit, hence xyz's already
        ***  written.
        **********************************************************/
        if (!atoms[i].visited) {
          bytes = sprintf(p, "%.2f,%.2f,%.2f,",
                          atoms[i].x, atoms[i].y, atoms[i].z);
          atoms[i].visited = 1;
          p += bytes;
        }
      }
    }
    dt_dealloc(bondstream);
  }
  dt_dealloc(atomstream);
 
  *(--p) = '\0'; /*** Terminate at trailing comma ***/
 
  return (buf3d);
}

/**********************************************************************
*** d2d - returns string containing XYZ's for use in 2D
***       datatype.  It is assumed that ALL H atoms are explicitly
***       defined and the x/y arrays contain their coordinates.
**********************************************************************/
char *d2d(dt_Handle mol, int num_atoms)
{
  dt_Handle atomstream, atom, bondstream, bond, atom2;
  int       acount, hcount_total, i, bufsize, bytes;
  char      *p;
  static char *buf2d = NULL;
  static int max_num_atoms = 0;

  for(i = 1; i < num_atoms+1; ++i) atoms[i].visited = 0;

  atomstream = dt_canatom_stream(mol, 0, 0);
  if (NULL_OB == atomstream) return ("");
 
  bufsize = 16 * num_atoms + 1;

  if (NULL == buf2d) buf2d = (char *) malloc(bufsize);
  else               buf2d = (char *) realloc(buf2d, bufsize);

  /***********************************************************
  ***  1st write xy's for explict atoms in cansmi order
  ***********************************************************/
  for (p = buf2d; NULL_OB!=(atom = dt_next(atomstream)); p += bytes) {
    i = (int) dt_adjunct(atom);
    bytes = sprintf(p, "%.2f,%.2f,", atoms[i].x, atoms[i].y);
    atoms[i].visited = 1;
  }

  /***********************************************************
  ***  Now for the H's -- since we have ALL coordinates
  ***********************************************************/
  dt_reset(atomstream);
  while (NULL_OB!=(atom = dt_next(atomstream))) {
    bondstream = dt_stream(atom, TYP_BOND);
    while (NULL_OB!=(bond = dt_next(bondstream))) {
      atom2 = dt_xatom(atom, bond);
      if (DX_ATN_H == dt_number(atom2)) {
        i = (int) dt_adjunct(atom2);

        /**********************************************************
        ***  If (visited) then H is explicit, hence xyz's already
        ***  written.
        **********************************************************/
        if (!atoms[i].visited) {
          bytes = sprintf(p, "%.2f,%.2f,", atoms[i].x, atoms[i].y);
          atoms[i].visited = 1;
          p += bytes;
        }
      }
    }
    dt_dealloc(bondstream);
  }
  dt_dealloc(atomstream);
 
  *(--p) = '\0'; /*** Terminate at trailing comma ***/
 
  return (buf2d);
}

/**********************************************************************
*** atypes - returns string containing Sybyl atom types for use
***            in the ATYPES datatype.  The atom types may take
*** up to 5 bytes (e.g., "N.pl3").
*********************************************************************
*** Atom types:
***  "C.3"    ... >C< sp3 ALKANE      
***  "C.2"    ... >C=C< ALKENE       
***  "C.1"    ... -C=C- ALKINE      
***  "C.ar"   ... -C=C- AROMATIC 
***  "C.CAT"  ... -N=C<N2 GUANIDINIUM 
***  "N.3"    ... >N< sp3             
***  "N.2"    ... -N< sp2            
***  "N.1"    ... -N=C sp CYANID    
***  "N.ar"   ... -N= AROMATIC    
***  "N.am"   ... H-N-C=O AMIDE  
***  "N.pl3"  ... -N< sp3 PLANAR
***  "N.4"    ... >N+< AMMONIUM       
***  "O.3"    ... -O- sp3            
***  "O.2"    ... O=C< sp2          
***  "O.CO2"  ... O=C-O(-) CARBOXYLATE 
***  "S.3"    ... -S- sp3             
***  "S.2"    ... S=C< sp2           
***  "S.O"    ... >S=O SULFOXID     
***  "S.O2"   ... O=S=O SULFATE   
***  "S.o2"
***  "P.3"    ... >P< sp3         
***  "LP"     ... Lone Pair       
***  "Du"     ... Dummy        
**********************************************************************/
char *atypes(dt_Handle mol, int num_atoms)
{
  dt_Handle atomstream, atom, bondstream, bond, atom2;
  int       i, bufsize, bytes;
  char      *p, *q;
  static char *buf = NULL;
  static int max_num_atoms = 0;

  for(i = 1; i < num_atoms+1; ++i) atoms[i].visited = 0;

  atomstream = dt_canatom_stream(mol, 0, 0);
  if (NULL_OB == atomstream) return ("");
 
  bufsize = 6 * num_atoms + 1;

  if (NULL == buf) buf = (char *) malloc(bufsize);
  else             buf = (char *) realloc(buf, bufsize);

  /***********************************************************
  ***  1st write atom_types for explict atoms in cansmi order
  ***********************************************************/
  for (p = buf; NULL_OB!=(atom = dt_next(atomstream)); p += bytes) {
    i = (int) dt_adjunct(atom);
    atoms[i].visited = 1;
    bytes = sprintf(p, "%s,", atoms[i].type);
  }

  /***********************************************************
  ***  Now for the H's
  ***********************************************************/
  dt_reset(atomstream);
  while (NULL_OB!=(atom = dt_next(atomstream))) {
    bondstream = dt_stream(atom, TYP_BOND);
    while (NULL_OB!=(bond = dt_next(bondstream))) {
      atom2 = dt_xatom(atom, bond);
      if (DX_ATN_H == dt_number(atom2)) {
        i = (int) dt_adjunct(atom2);

        /**********************************************************
        ***  If (visited) then H is explicit, hence atom_type already
        ***  written.
        **********************************************************/
        if (!atoms[i].visited) {
          bytes = sprintf(p, "%s,", atoms[i].type);
          atoms[i].visited = 1;
          p += bytes;
        }
      }
    }
    dt_dealloc(bondstream);
  }
  dt_dealloc(atomstream);
 
  *(--p) = '\0'; /*** Terminate at trailing comma ***/
 
  return (buf);
}


/********************************************************************
***  setchiral3 -
***  3D -> tetrahedral chirality (all possible chiral centers)
*********************************************************************
***  Note that setchiral3 will set chirality in some cases with
***  one implicit H.  The signed volume is calculated with the
***  implicit H located at the central atom.
********************************************************************/
int setchiral3(dt_Handle mol)
{
  dt_Handle atom;             /*** Atom object                    ***/
  dt_Handle atomseq;          /*** Chiral Atom-sequence           ***/
  dt_Handle atomstream;       /*** Atom stream                    ***/
  dt_Handle bond;             /*** Bond object                    ***/
  dt_Handle bondseq;          /*** Chiral & CIS/TRANS bond-seq    ***/
  dt_Handle bondstream;       /*** Bond stream                    ***/
  dt_Handle hyd;              /*** Chiral hydrogen atom           ***/
  float vol;                  /*** (Chirality) signed volume      ***/
  int i, j;                   /*** Generic indices                ***/
  int imp_hcount = 0;         /*** (Chiral) implicit H-count      ***/

  atomstream = dt_stream(mol, TYP_ATOM);
  atomseq = stream2seq(atomstream);
  dt_dealloc(atomstream);
  while (NULL_OB!=(atom = dt_next(atomseq))) {
    bondstream = dt_stream(atom, TYP_BOND);
    bondseq = stream2seq(bondstream);
    dt_dealloc(bondstream);

    /*********************************************************
    *** Check: No more than one imp H; four single bonds.
    *** (There should be no implicit H's since none were
    *** set!)
    *********************************************************/
    imp_hcount = dt_imp_hcount(atom);
    for (j=0; NULL_OB!=(bond=dt_next(bondseq)); ) j+=dt_bondtype(bond);
    if (4!=(j+imp_hcount)   ||
        1<imp_hcount        ||
        4!=(imp_hcount+dt_count(bondseq, TYP_BOND))) {
      dt_dealloc(bondseq);
      continue;
    }

    vol = signed_volume(atom, bondseq);
    if (fabs(vol) > FLT_EPSILON) {
      /***************************************************
      *** If there's an imp H, we add it to the seq.
      ***************************************************/
      if (imp_hcount) {
        dt_append(bondseq, dt_isohydro());
      }
      if (0.0>vol) dt_setchival(atom, bondseq, DX_CHI_THCCW);
      else dt_setchival(atom, bondseq, DX_CHI_THCW);
    }
    dt_dealloc(bondseq);
  }
  dt_dealloc(atomseq);
  return (1);
}

/********************************************************************
***  setcistrans -  CIS/TRANS adjustments from 2D/3D for all
***                 possible sites.
*********************************************************************
***                              We look for double bonded
***      2       4               atoms that are 3-connected.
***       \     /                Replace imp H's with exp H's,
***        0 = 1    ez_atoms[]   and check coordinates for
***       /     \                CIS/TRANS.   
***      3       5               
***                              
***   Note also: we don't allow CIS/TRANS for ring double bonds.
********************************************************************/
int setcistrans(dt_Handle mol) {

  dt_Handle atom;             /*** Atom object                    ***/
  dt_Handle atomseq;          /*** Chiral Atom-sequence           ***/
  dt_Handle atomstream;       /*** Atom stream                    ***/
  dt_Handle bond;             /*** Bond object                    ***/
  dt_Handle bondseq;          /*** Chiral & CIS/TRANS bond-seq    ***/
  dt_Handle bondstream;       /*** Bond stream                    ***/
  dt_Handle dbond;            /*** CIS/TRANS double bond          ***/
  dt_Handle ez_atoms[6];      /*** CIS/TRANS atoms                ***/
  dt_Handle hyd;              /*** Chiral hydrogen atom           ***/
  float a[3];                 /*** Vector for dihedral calc       ***/
  float b[3];                 /*** Vector for dihedral calc       ***/
  float c[3];                 /*** Vector for dihedral calc       ***/
  int bond_stereo;            /*** Bond stereo designator         ***/
  int dboval;                 /*** (CIS/TRANS) bond designator    ***/
  int i, j, k, jj;            /*** Generic indices                ***/
  int i_j, i_k;               /*** Indices for chiral atoms j, k  ***/
  int imp_hcount = 0;         /*** (Chiral) implicit H-count      ***/
  int dihedral;               /*** Dihedral angle in degrees      ***/
  int ii[6];                  /*** Indices for CIS/TRANS atoms    ***/

  bondstream = dt_stream(mol, TYP_BOND);
  bondseq = stream2seq(bondstream);
  dt_dealloc(bondstream);
  while (NULL_OB!=(dbond = dt_next(bondseq))) {
    if (2 == dt_bondtype(dbond)) {

      /*******************************************************
      *** We don't allow CIS/TRANS for ring bonds.
      *** (Must temporarily mod_off to check for cycles.)
      *******************************************************/
      if (!dt_mod_off(mol)) {
        sprintf(errtxt,"setcistrans/dt_mod_off failure. (mol#=%d)",ct_in+1);
        dt_errorsave("syb2smi",DX_ERR_ERROR,errtxt);
        return (0);
      }
      if (0!=dt_count(dbond, TYP_CYCLE)) continue;
      dt_mod_on(mol);

      atomstream = dt_stream(dbond, TYP_ATOM);
      ez_atoms[0] = dt_next(atomstream);
      ez_atoms[1] = dt_next(atomstream);
      dt_dealloc(atomstream);

      /*** Check for 3-connectedness. ***/
      imp_hcount = dt_imp_hcount(ez_atoms[0]);
      if (3!=(imp_hcount + dt_count(ez_atoms[0], TYP_BOND)))
        continue;
      imp_hcount = dt_imp_hcount(ez_atoms[1]);
      if (3!=(imp_hcount + dt_count(ez_atoms[1], TYP_BOND)))
        continue;

      /********************************************************************
      *** Replace implicit H's with explicit.  Set their adjuncts
      *** (indices) to zero for identification later.
      ********************************************************************/
      for (j = 0; j <= 1; ++j) {
        if (1 == dt_imp_hcount(ez_atoms[j])) {
          dt_setimp_hcount(ez_atoms[j], 0);
          hyd = dt_addatom(mol, DX_ATN_H, 0);
          bond = dt_addbond(ez_atoms[j], hyd, 1);
          dt_setadjunct(hyd, (dt_Handle) 0);
        }
      }

      /*** Assign atoms #2 and #3 ***/
      bondstream = dt_stream(ez_atoms[0], TYP_BOND);
      for (j = 2; NULL_OB!=(bond = dt_next(bondstream)); ++j) {
        if (2 == dt_bondtype(bond)) { --j; continue; }
        ez_atoms[j] = dt_xatom(ez_atoms[0], bond);
      }
      dt_dealloc(bondstream);
      if (j!=4) continue; /*** Check for 2 single bonds. ***/

      /*** Assign atoms #4 and #5 ***/
      bondstream = dt_stream(ez_atoms[1], TYP_BOND);
      for (j = 4; NULL_OB!=(bond = dt_next(bondstream)); ++j) {
        if (2 == dt_bondtype(bond)) { --j; continue; }
        ez_atoms[j] = dt_xatom(ez_atoms[1], bond);
      }
      dt_dealloc(bondstream);
      if (j!=6) continue; /*** Check for 2 single bonds. ***/

      /***********************************************************
      ***  If ez_atoms[j=2] was an implicit H, use j=3.
      ***  If ez_atoms[k=4] was an implicit H, use k=5.
      ***********************************************************/
      j = (0!=((int) dt_adjunct(ez_atoms[2])) ? 2 : 3);
      k = (0!=((int) dt_adjunct(ez_atoms[4])) ? 4 : 5);

      /*** Atom indices ***/
      for (jj = 0; jj < 6; ++jj) ii[jj] = (int) dt_adjunct(ez_atoms[jj]);

      a[0] = atoms[ii[0]].x - atoms[ii[j]].x;
      a[1] = atoms[ii[0]].y - atoms[ii[j]].y;
      a[2] = atoms[ii[0]].z - atoms[ii[j]].z;

      b[0] = atoms[ii[1]].x - atoms[ii[0]].x;
      b[1] = atoms[ii[1]].y - atoms[ii[0]].y;
      b[2] = atoms[ii[1]].z - atoms[ii[0]].z;

      c[0] = atoms[ii[k]].x - atoms[ii[1]].x;
      c[1] = atoms[ii[k]].y - atoms[ii[1]].y;
      c[2] = atoms[ii[k]].z - atoms[ii[1]].z;

      dihedral = (int) fabs(rad2deg(du_3d_dangle(a, b, c)));
      dboval = (((90 - dihedral) * pow(-1, j + k)) > 0 ?
                DX_CHI_CIS : DX_CHI_TRANS);
      /*** Assign CIS/TRANS stereochemistry. ***/
      dt_setdbo(dt_bond(ez_atoms[0], ez_atoms[1]),
                dt_bond(ez_atoms[0], ez_atoms[j]),
                dt_bond(ez_atoms[1], ez_atoms[k]),
                dboval);
    }
  }
  dt_dealloc(bondseq);
  return (1);
}

/********************************************************************
***  signed_volume - compute volume defined by the xyzs
***  of the four atoms single-bonded to the chiral atom. Sign
***  of the volume determines R/S chirality.
***  (Implicit H's still not quite handled right.)
********************************************************************/
float signed_volume(dt_Handle atom, dt_Handle bondseq)
{
  dt_Handle atom2, bond;
  float   a[3], b[3], c[3], d[3];
  int     i, imp_hcount;

  switch (dt_count(bondseq, TYP_BOND)) {
    case 3: imp_hcount = 1; break;
    case 4: imp_hcount = 0; break;
    default: return (0.0);
  }

  /**********************************************************
  *** Extract xyz's of attached atoms in bondseq order.
  **********************************************************/
  dt_reset(bondseq);
  atom2 = dt_xatom(atom, bond = dt_next(bondseq));
  i = (int) dt_adjunct(atom2);
  a[0] = atoms[i].x; a[1] = atoms[i].y; a[2] = atoms[i].z;

  atom2 = dt_xatom(atom, bond = dt_next(bondseq));
  i = (int) dt_adjunct(atom2);
  b[0] = atoms[i].x; b[1] = atoms[i].y; b[2] = atoms[i].z;

  atom2 = dt_xatom(atom, bond = dt_next(bondseq));
  i = (int) dt_adjunct(atom2);
  c[0] = atoms[i].x; c[1] = atoms[i].y; c[2] = atoms[i].z;

  /**********************************************************
  *** If we have an implicit H, use central atom xyz's.
  *** This will work for all 3D cases, but not all conceivable
  *** "2.5D" pictures (i.e., _!_ ).
  **********************************************************/
  if (imp_hcount) i = (int) dt_adjunct(atom);
  else {
    atom2 = dt_xatom(atom, bond = dt_next(bondseq));
    i = (int) dt_adjunct(atom2);
  }
  d[0] = atoms[i].x; d[1] = atoms[i].y; d[2] = atoms[i].z;

  return (du_3d_signed_volume(a, b, c, d));
}

/********************************************************************
*** isosmi - set stereochemistry of molecule
********************************************************************/
char *isosmi(dt_Handle mol, int *plen, int wegot_3d)
{
  dt_mod_on(mol);

  if ((wegot_3d && !setchiral3(mol)) ||
      !setcistrans(mol) ||
      !dt_mod_off(mol))
  {
    sprintf(errtxt,"Stereo error; name=\"%s\",mol#=%d",name,ct_in+1);
    dt_errorsave("syb2smi",DX_ERR_ERROR,errtxt);
    *plen = -1;
    return ("");
  }
  return (dt_cansmiles(plen, mol, 1));
}

/********************************************************************
*** infer_charges - guess charges based on valence and connectivity
*** for selected elements (possible aromatics).  This function was
*** motivated by dt_mod_off failure due to aromatic rings with
*** unrecognized charge (e.g., pyridinium).
********************************************************************/
dt_Handle infer_charges(dt_Handle mol,float pcharge_total)
{
  dt_Handle atom,atoms;
  int charge,vbond,nbond,fcharge_total=0;

  atoms = dt_stream(mol,TYP_ATOM);
  
  while (NULL_OB!=(atom=dt_next(atoms))) {
    charge = 0;
    vbond = du_vbond(atom,&nbond);
    switch (dt_number(atom)) {
      case DX_ATN_C:
        if (3==vbond && 1==nbond) charge = -1;
        else if (3==vbond) charge = 1;
        break;
      case DX_ATN_N:
        if (2==vbond && 1==nbond) charge = -1;
        else if (4==vbond) charge = 1;
        break;
      case DX_ATN_O:
        if (1==vbond) charge = -1;
        else if (3==vbond) charge = 1;
        break;
      case DX_ATN_S:
        if (1==vbond) charge = -1;
        else if (3==vbond) charge = 1;
        break;
      default:
        break;
    }
    if (charge) dt_setcharge(atom,charge);
    fcharge_total += charge;
  }
  dt_dealloc(atoms);
  if (pcharge_total - (float)fcharge_total > 0.05)  /*** crude check ***/
    dt_errorsave("syb2smi",DX_ERR_WARN,"Charges not fully understood.");

  return(mol);
}

/*****************************************************************************
*** du_vbond - Count bond equivalents.  Also return number of
*** connections in nbond.
*****************************************************************************/
int du_vbond(dt_Handle atom,int *nbond)
{
  dt_Handle bond,bonds;
  int abond=0,vbond=0;

  bonds = dt_stream(atom,TYP_BOND);
  while (NULL_OB!=(bond=dt_next(bonds))) {
    switch (dt_bondtype(bond)) {
      case DX_BTY_SINGLE: vbond += 1; break;
      case DX_BTY_DOUBLE: vbond += 2; break;
      case DX_BTY_TRIPLE: vbond += 3; break;
      case DX_BTY_AROMAT: vbond += 1; abond += 1; break;
    }
  }
  if (0<abond) vbond++;
  *nbond = dt_count(bonds,TYP_BOND);
  dt_dealloc(bonds);
  return (vbond);
}

/********************************************************************
*** stream2seq 
********************************************************************/
dt_Handle stream2seq(dt_Handle stream) {
  dt_Handle ob,seq=dt_alloc_seq();
  while (NULL_OB!=(ob=dt_next(stream))) dt_append(seq,ob);
  return seq;
}

