#!/usr/local/bin/perl
#############################################################################
### dtypefix.pl - the idea here is to convert the TDTs produced
### by mol2smi from Derwent's WDI raw files to our standard format.
### That includes renaming datatypes and rearranging the TDT.
###
### In general, we want $SMI first, then $DXRN, then the data belonging
### to that $DXRN, then other id's.  When these TDTs are merged under
### one SMILES, the separate $DXRN's will thereby keep their associated
### data.
#############################################################################
### Note: (AUTOGEN $NAM) is not used for NAM,DPN because this won't
### work with non-SMILES TDTs.
#############################################################################
### Author: Jeremy Yang
### 28 Jan 1999
#############################################################################
$| = 1;
$* = 1;

require DU_LIB;

%tree = ( );

while (<STDIN>) {
  chop();

  next if (/^ALAB</); ### ALAB causes problems due to H/ntuple bug.

got_newitem:

  $combine = 0;

  if (/^\|$/) {
    &PRINT_TDT();
    undef %tree;
    next;
  }

  ($tag = $_) =~ s/^([^<]+)<(.*)$/$1/;
  ($data = $2) =~ s/>$//;

  #######################################################
  ### Dataitems needing no action:
  ###
  ###  $NAM, $SMI, $SMIG, 2D, ISM, CAS, etc.
  ### (any others not specified)
  #######################################################
  if ($tag eq '$SMI') { 
    if (exists($tree{'$SMI'})) {
      print STDERR "ERROR: >1 \$SMI data encountered...\n";
      &PRINT_TDT();
      undef %tree;
      goto got_newitem;
    }
  }

  #######################################################
  ### Dataitems for deletion.
  ###  "RC" = "Ring code", Derwent-obsolete in 1994.
  #######################################################
  elsif ($tag eq "RC") { next; }

  #######################################################
  ### Dataitems which stay separate instances.
  ###
  ###  MF -> DMF (Derwent molecular formula)
  #######################################################
  elsif ($tag eq "MOL_WEIGHT") { $tag = 'AMW'; }
  elsif ($tag eq "MOLNAME")    { $tag = 'DPN'; }
  elsif ($tag eq "MF")         { $tag = 'DMF'; }
  elsif ($tag eq "OTHER")      { $tag = 'NAM'; }
  elsif ($tag eq "RN")         { $tag = 'DRN'; }
  elsif ($tag eq "RNEXTREG")   { $tag = '$DXRN'; }
  elsif ($tag eq "TP")         { $tag = '$NAM'; }
  elsif ($tag eq "UPD_CODE")   { $tag = 'DYQ'; }

  #######################################################
  ### Dataitems which combine into one.
  #######################################################
  elsif ($tag eq "AE")                   { $combine = 1; }
  elsif ($tag eq "CI")                   { $combine = 1; }
  elsif ($tag eq "COMMENT") { $tag = 'SD'; $combine = 1; }
  elsif ($tag eq "IA")                   { $combine = 1; }
  elsif ($tag eq "IU")                   { $combine = 1; }
  elsif ($tag eq "MA")                   { $combine = 1; }
  elsif ($tag eq "PT")                   { $combine = 1; }
  elsif ($tag eq "PW")                   { $combine = 1; }
  elsif ($tag eq "SS")     { $tag = 'SSK'; $combine = 1; }

  #######################################################
  ### Dataitems requiring slight reformatting.
  ###
  ### e.g.,
  ### APP<IMPROSULFAN-TOSILATE                   JAN>
  ###       39 chars, * chars
  ### -->  APN<IMPROSULFAN-TOSILATE;JAN>
  ###  
  ### USAN<ACEBUTOLOL                             USAN>
  ###       39 chars, * chars
  ###  -->  USAN<ACEBUTOLOL;USAN>
  ###
  ### INN<ACEBUTOLOL                             Rec.>
  ###       39 chars, * chars
  ###  -->  INN<ACEBUTOLOL;Rec.>
  ###  
  ### TRADE<FLOMOXEF-SODIUM                        SHIONOGI            JP>
  ###       39 chars, 20 chars, 2 chars
  ###  -->  TN<FLOMOXEF-SODIUM;SHIONOGI;JP>
  #######################################################
  elsif ($tag eq "INN" || $tag eq "USAN" || $tag eq "APP") {
    if ($tag eq "APP") { $tag = "APN"; }
    $data =~ s/^"(.*)"$/$1/;
    ($f1 = substr($data, 0, 39)) =~ s/\s*$//;
    ($f2 = substr($data, 39)) =~ s/\s*$//;
    $f2 =~ s/^\s*//;
    $data = "\"$f1\";\"$f2\"";
  }
  elsif    ($tag eq "TRADE") {
    $tag = "TN";
    $data =~ s/^"(.*)"$/$1/;
    ($f1 = substr($data, 0, 39)) =~ s/\s*$//;
    ($f2 = substr($data, 39, 20)) =~ s/\s*$//;
    $f2 =~ s/^\s*//;
    ($f3 = substr($data, 59)) =~ s/\s*$//;
    $f3 =~ s/^\s*//;
    $data = "\"$f1\";\"$f2\";\"$f3\"";
  }

  #######################################################
  ### Dataitems requiring special handling.
  #######################################################

  #######################################################
  ###  REFERENCE<AN  - 95-09342>
  ###  REFERENCE<JN  - Arch.Pharmacol. (350, Suppl., R21, 1994)>
  ###  REFERENCE<TI  - The chemical synthesis of 6-hydroxybenzbromarone...>
  ###  REFERENCE<benzbromarone, and its biotransformation by human liver.>
  ###    ... more lines possible
  ###  
  ###   ->  JOUR<ti;jn;an>
  #######################################################
  elsif (($tag eq "REFERENCE") && ($data =~ /^"*AN/)) {
    $an = $jn = $ti = "";
    if ($data =~ /^".*"$/) { $data =~ s/^"(.*)"$/$1/; }    ### de-quote
    while (1) {
      if ($data =~ /^AN\s+-/) {
        ($an = $data) =~ s/^AN\s+-\s+//;
      } elsif ($data =~ /^JN\s+-/) {
        ($jn = $data) =~ s/^JN\s+-\s+//;
      } elsif ($data =~ /^TI\s+-/) {
        ($ti = $data) =~ s/^TI\s+-\s+//;
        $ti =~ s/"/""/g;                      ### double quotes
      } else {
        $data =~ s/"/""/g;                    ### double quotes
        $ti .= " ".$data;
      }
      chop($_ = <STDIN>);
      ($tag = $_) =~ s/^([^<]+)<(.*)$/$1/;
      ($data = $2) =~ s/>$//;
      if ($data =~ /^".*"$/) { $data =~ s/^"(.*)"$/$1/; }    ### de-quote
      last if  (!/^REFERENCE</ || /^REFERENCE<"*Conference/);
    }

    $tag = "JOUR";
    $data = "\"".$ti."\";\"".$jn."\";".$an;       ### re-quote
    if (exists($tree{$tag})) { $tree{$tag} .= "\0".$data; }
    else                     { $tree{$tag} = $data; }
    
    goto got_newitem;
  }

  #######################################################
  ### REFERENCE<Conference Fast-Track Reference Number: NCE-97-1455>
  ### REFERENCE<Conference: 37th Interscience Conference on Antimicrobial Agents and C hemotherapy, Toronto, Ontario, Canada,  28 September - 1 October 1997.>
  ###    ... more lines possible
  ### REFERENCE<Presentation Details: Poster, F267>
  ### REFERENCE<WDA Reference Number: WD-97-013049>
  ###  
  ###   ->  CONF<name;ftrn;pd;wdarn>
  #######################################################
  elsif (($tag eq "REFERENCE") && ($data =~ /^"*Conference/)) {
    $name = $ftrn = $pd = $wdarn = "";
    if ($data =~ /^".*"$/) { $data =~ s/^"(.*)"$/$1/; }    ### de-quote
    while (1) {
      if ($data =~ /^Conference Fast-Track Reference Number:/) {
        ($ftrn = $data) =~ s/^Conference Fast-Track Reference Number:\s+//;
      } elsif ($data =~ /^Presentation Details:/) {
        ($pd = $data) =~ s/^Presentation Details:\s+//;
        $pd =~ s/;/,/g;
      } elsif ($data =~ /^WDA Reference Number:/) {
        ($wdarn = $data) =~ s/^WDA Reference Number:\s+//;
      } elsif ($data =~ /^Conference:/) {
        ($name = $data) =~ s/^Conference:\s+//;
        $name =~ s/"/""/g;                      ### double quotes
      } else {
        $data =~ s/"/""/g;                      ### double quotes
        $name .= " ".$data;
      }
      chop($_ = <STDIN>);
      ($tag = $_) =~ s/^([^<]+)<(.*)$/$1/;
      ($data = $2) =~ s/>$//;
      if ($data =~ /^".*"$/) { $data =~ s/^"(.*)"$/$1/; }    ### de-quote
      last if  (!/^REFERENCE</ || /^REFERENCE<"*AN/);
    }

    $tag = "CONF";
    $data = "\"".$name."\";".$ftrn.";".$pd.";".$wdarn;       ### re-quote
    if (exists($tree{$tag})) { $tree{$tag} .= "\0".$data; }
    else                     { $tree{$tag} = $data; }
    
    goto got_newitem;
  }

  #######################################################
  ### Miscellaneous references
  #######################################################
  elsif ($tag eq "REFERENCE") {
    $ref = "";
    if ($data =~ /^".*"$/) { $data =~ s/^"(.*)"$/$1/; }    ### de-quote
    while (1) {
      $data =~ s/"/""/g;                    ### double quotes
      $ref .= " ".$data;
      chop($_ = <STDIN>);
      ($tag = $_) =~ s/^([^<]+)<(.*)$/$1/;
      ($data = $2) =~ s/>$//;
      if ($data =~ /^".*"$/) { $data =~ s/^"(.*)"$/$1/; }    ### de-quote
      last if  (!/^REFERENCE</ || /^REFERENCE<"*Conference/
                               || /^REFERENCE<"*AN/);
    }

    $tag = "REF";
    $data = "\"".$ref."\"";       ### re-quote
    if (exists($tree{$tag})) { $tree{$tag} .= "\0".$data; }
    else                     { $tree{$tag} = $data; }
    
    goto got_newitem;
  }


  #######################################################
  ### COMB_PREP<F-99-SULGAN             DIVAPHARMA INC.          DE>
  ### COMB_PREP<( + aluminum-chlorohydrate + propylene-glycol + camphor>
  ### COMB_PREP<+ linoleate-ethyl + ethyl-linoleate + hamamelis + menthol>
  ### COMB_PREP<+ zinc-oxide )>
  ###  -->  COMB_PREP<F-99-SULGAN;DIVAPHARMA INC.;DE;( + aluminum-chlorohydrate + propylene-glycol + camphor + linoleate-ethyl + ethyl-linoleate + hamamelis + menthol + zinc-oxide )>
  #######################################################
  elsif ($tag eq "COMB_PREP") {
    while (1) {
      $data =~ s/^"(.*)"$/$1/;
      if ($data =~ /^\(/) {
        $desc = $data;
      } elsif ($data =~ /^\+/) {
        $desc .= " ".$data;
      } else {
        ($a = $data) =~ s/^(\S+).*$/$1/;
        ($b = $data) =~ s/^(\S+)\s+(.*)..$/$2/;
        $b =~ s/\s*$//;
        ($c = $data) =~ s/^.*(..)$/$1/;
      }
      last if (/\)>$/);  ### Normal end of COMB_PREP

      chop($_ = <STDIN>);
      ($tag = $_) =~ s/^([^<]+)<(.*)$/$1/;
      ($data = $2) =~ s/>$//;

      last if (!/^COMB_PREP</);  ### Error situation
    }

    $tag = "COMB_PREP";
    $data = "\"$a\";\"$b\";\"$c\";\"$desc\"";
    $a = $b = $c = "";
    if (exists($tree{$tag})) { $tree{$tag} .= "\0".$data; }
    else                     { $tree{$tag} = $data; }
    
    if (!/^COMB_PREP</) { goto got_newitem; }
    else { next; }
  }

  #######################################################
  ### Add new data to (tree).
  #######################################################
  if ($combine) {
    if (exists($tree{$tag})) {
      $tree{$tag} =~ s/^"(.*)"$/$1/;                  ### de-quote...
      $data =~ s/^"(.*)"$/$1/;
      $tree{$tag} = "\"".$tree{$tag}." ".$data."\"";  ### re-quote.
    } else                   { $tree{$tag} = $data; }
  } else {
    if (exists($tree{$tag})) { $tree{$tag} .= "\0".$data; }
    else                     { $tree{$tag} = $data; }
  }

  ### Final touches

  if ($tag eq 'SD') { $tree{$tag} =~ s/- ([A-Z][a-z][a-z])/-$1/g; }
  $tree{$tag} =~ s/\s+/ /g;
}

#############################################################################
##  PRINT_TDT - print standardized WDI TDT
#############################################################################
sub PRINT_TDT
{
  local($wegot_smi, $wegot_dxrn, $end_data, $end_ids, $names) =
    (0, 0, "", "", "");

  ################################################
  ### $SMI and/or $DXRN first.  ISM next.
  ################################################
  if (exists($tree{'$SMI'})) { $wegot_smi = 1; }
  if (exists($tree{'$DXRN'})) { $wegot_dxrn = 1; }
  if (!$wegot_smi && !$wegot_dxrn) {
    printf STDERR "Error (PRINT_TDT): no legal root: ";
    print STDERR %tree, "\n";
    return;
  }
  print &SPRINT_DTYPE('$SMI'); delete $tree{'$SMI'};
  print &SPRINT_DTYPE('$DXRN'); delete $tree{'$DXRN'};
  print &SPRINT_DTYPE('ISM'); delete $tree{'ISM'};

  ################################################
  ### Dataitems moved to end of TDT: $NAM
  ################################################
  $_ = &SPRINT_DTYPE('$NAM');
  if (!$wegot_smi) { s/\$NAM/NAM/g; }
  $names .= $_;
  delete $tree{'$NAM'};

  ################################################
  ### NAM,DPN: add $NAM's to $SMI-rooted TDTs.
  ### CAS: add $CAS's to $SMI-rooted TDTs.
  ### (Like AUTOGEN but avoids problem if no SMILES.)
  ################################################
  if ($wegot_smi) {
    foreach $tag ('NAM', 'DPN') {
      $_ = &SPRINT_DTYPE($tag);
      s/$tag/\$NAM/g;
      $names .= $_;
    }
    foreach $tag ('INN', 'APN', 'TN', 'USAN') {
      if (exists($tree{$tag})) {
        foreach $datum (split(/\0/, $tree{$tag})) {

#          $datum =~ s/;.*$//;

          $datum = substr($datum, 0, &DU_INDEX($datum, ";", 0));
          $names .= "\$NAM<$datum>\n";
        }
      }
    }
    $_ = &SPRINT_DTYPE('CAS');
    s/CAS/\$CAS/g;
    $end_ids .= $_;
  }

  ################################################
  ### Dataitems moved to end of TDT (but before end-ids):
  ###   TN (Trade Names)
  ################################################
  $end_data .= &SPRINT_DTYPE('TN');
  delete $tree{'TN'};

  ################################################
  ### If we got SMILES, move id's to end; else
  ### delete other id's.  Note: non-id's which
  ### AUTOGEN id's are not ok.
  ################################################
  foreach $tag (keys(tree)) {
    if ($tag =~ /^\$/) {
      if ($wegot_smi) { $end_ids .= &SPRINT_DTYPE($tag); }
      delete $tree{$tag};
    }
  }

  ################################################
  ### Print: first normal data, then end-data,
  ### then end-ids.
  ################################################
  foreach $tag (keys(tree)) { print &SPRINT_DTYPE($tag); }
  print $end_data;
  print $end_ids if ($wegot_smi);
  print $names;
  print "|\n";
}


#############################################################################
### SPRINT_DTYPE - Return printable dataitem(s) for a given tag.
#############################################################################
sub SPRINT_DTYPE
{
  local($tag) = @_;
  local($buff);

  if (exists($tree{$tag})) {
    foreach $datum (split(/\0/, $tree{$tag})) { $buff .= "$tag<$datum>\n"; }
  }
  return $buff;
}
